# -*- python -*-
# $RCSfile: outputdestination.py,v $
# $Revision: 1.9 $
# $Author: langer $
# $Date: 2010-09-16 19:24:45 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@nist.gov. 

from ooflib.SWIG.common import switchboard
from ooflib.common import debug
from ooflib.common import enum
from ooflib.common import utils
from ooflib.common import registeredclass
from ooflib.common.IO import datafile
from ooflib.common.IO import formatchars
from ooflib.common.IO import filenameparam
from ooflib.common.IO import parameter
from ooflib.common.IO import reporter
import weakref

class OutputDestination(registeredclass.RegisteredClass):
    registry = []

##########################

class GfxWindowDestination(OutputDestination):
    def shortrepr(self):
        return "<Graphics Window>"
    def flush(self):
        pass
    def rewind(self):
        pass
    def close(self):
        pass

registeredclass.Registration(
    'Graphics Window',
    OutputDestination,
    GfxWindowDestination,
    rewindable=False,
    ordering=0)

##############################################

# TextOutputDestination is an intermediate baseclass for
# OutputDestinations that produce some sort of human readable text
# (ie, not OOF2 data files which might be ascii).  This includes the
# results of Analysis operations.  There are three subclasses:
# OutputStream, which writes to a file; ExistingStream, which writes
# to another OutputStream, and MessageWindowStream, which writes to
# the OOF2 Message Window.

class TextOutputDestination(OutputDestination):
    _lastoutput = weakref.WeakValueDictionary()
    def __init__(self):
        self.seplast = True
    def rewind(self):           # for files, clears the file.
        self.seplast = True  # was the last thing written a separator?
        # Clear the _lastoutput value so that headers will be printed
        # the next time this destination is used.  Check for KeyError
        # in case the destination hasn't been used yet.
        try:
            del self._lastoutput[self.identifier()]
        except KeyError:
            pass
    def printHeadersIfNeeded(self, output, *args, **kwargs):
        lastout = self._lastoutput.get(self.identifier(), None)
        if lastout is not output:
            if lastout is not None:
                self.file().write("\n") # extra new line before header
            output.printHeaders(self, *args, **kwargs)
            self._lastoutput[self.identifier()] = output
    def write(self, text):
        # When an object with a "write" method is used as the argument
        # of "print >>", write is called once for each printed string,
        # once for each space between printed strings, and once for
        # the newline at the end.
        if text == " " and not self.seplast:
            self.file().write(formatchars.getSeparator())
            self.seplast = True
        elif text == "\n":
            self.file().write(text)
            self.seplast = True
        else:
            self.file().write(text)
            self.seplast = False
    def comment(self, *args):
        self.file().write(" ".join([formatchars.getCommentChar()] + 
                                   [x for x in args] ))
        self.file().write("\n")
        self.seplast = False

allTextOutputStreams = utils.OrderedDict()

def forgetTextOutputStreams():
    allTextOutputStreams.clear()
    switchboard.notify("output destinations changed")

# OutputStream directs output to a file, specified by a file name and
# mode.  More than one OutputStream may point to a given file. (See
# comments for ExistingStream, below, for the reason.)  If two
# OutputStreams have the same filename but different modes, the first
# mode specified is used.

class OutputStream(TextOutputDestination):
    def __init__(self, filename, mode):
        self.filename = filename
        self.mode = mode
        TextOutputDestination.__init__(self)
        # Because more than one OutputStream can refer to the same
        # file, each OutputStream object has a referent, and all
        # actual work is always done by the referent.  All streams
        # with the same file have the same referent, which is the
        # first stream created for that file.
        try:
            self._referent = allTextOutputStreams[filename]
        except KeyError:
            # There is not an OutputStream using this file already.
            allTextOutputStreams[filename] = self
            self._referent = self
            self._file = None
            self.rewind()
            self.references = 1
            switchboard.notify("output destinations changed")
        else:
            # Reusing an existing stream.
            self._referent.references += 1
    def getFileName(self):
        return self._referent.filename
    def identifier(self):
        # This is used to determine if different outputs are writing
        # to the same place.
        return self.getFileName()
    def rewind(self):
        if self._referent._file is not None:
            # TODO: what if mode == 'append'?  The next call to
            # dfile() will reopen at the end of the file.
            self._referent._file.close()
            self._referent._file = None
        TextOutputDestination.rewind(self)
    def flush(self):
        if self._referent._file is not None:
            self._referent._file.flush()
    def file(self):
        if self._referent._file is None:
            self._referent._file = file(self.filename, self.mode.string())
        return self._referent._file
    def close(self):
        self._referent.references -= 1
        if self._referent.references == 0:
            if self._referent._file is not None:
                self._referent._file.close()
                self._referent._file = None
            del allTextOutputStreams[self.filename]
    def shortrepr(self):
        return self._referent.filename

newreg = registeredclass.Registration(
    'Output Stream',
    OutputDestination,
    OutputStream,
    ordering=1,
    rewindable=True,
    params=[
        filenameparam.WriteFileNameParameter('filename',
                                             tip=parameter.emptyTipString),
        filenameparam.WriteModeParameter('mode')
        ]
)

###############

# ExistingStream allows GUI users to direct more output to an existing
# OutputStream. Only the file name is needed to identify the existing
# OutputStream; the file mode is acquired from the OutputStream. In
# scripts it's a bad idea to use ExistingStream, because the script
# may have been edited and the referent OutputStream may have been
# removed.  Therefore, ExistingStream.__repr__ uses the OutputStream
# repr, and ExistingStream objects don't appear in script at all.

class ExistingStream(TextOutputDestination):
    def __init__(self, filename):
        self.filename = filename
        TextOutputDestination.__init__(self)
        self.outstream = allTextOutputStreams[filename]
        self.outstream.references += 1
    def getFileName(self):
        return self.outstream.filename
    def identifier(self):
        # Used to know if different outputs are writing to the same place.
        return self.getFileName()
    def rewind(self):
        self.outstream.rewind()
    def flush(self):
        self.outstream.flush()
    def file(self):
        return self.outstream.file()
    def close(self):
        self.outstream.close()
    def shortrepr(self):
        return self.outstream.shortrepr()
    def __repr__(self):
        return repr(self.outstream)

class ExistingStreamParam(parameter.StringParameter):
    pass

registeredclass.Registration(
    'Existing Stream',
    OutputDestination,
    ExistingStream,
    ordering=2,
    rewindable=True,
    params=[
        ExistingStreamParam("filename",
                            tip="Name of an already opened output stream.")
        ],
    tip="Send output to a stream already in use by another Output operation."
)
    
###############
    
class MessageWindowStream(TextOutputDestination):
    def __init__(self):
        TextOutputDestination.__init__(self)
        self._file = reporter.fileobj
    def file(self):
        return self._file
    def identifier(self):
        # Used to know if different outputs are writing to the same place.
        return self.__class__  
    def open(self):
        pass
    def close(self):
        pass
    def flush(self):
        self._file.flush()
    def shortrepr(self):
        return "<Message Window>"

registeredclass.Registration(
    'Message Window',
    OutputDestination,
    MessageWindowStream,
    ordering=0,
    rewindable=False,
    tip="Send output to the Message Window."
)

msgWindowOutputDestination = MessageWindowStream()

##############################################

class DataFileOutput(OutputDestination):
    def __init__(self, filename, mode, format):
        self.filename = filename
        self.mode = mode
        self.format = format
        self._dfile = None
    def dfile(self):
        if self._dfile is None:
            self._dfile = datafile.writeDataFile(
                self.filename, self.mode.string(), self.format)
        return self._dfile
    def flush(self):
        if self._dfile is not None:
            self._dfile.flush()
    def close(self):
        if self._dfile is not None:
            self._dfile.close()
            self._dfile = None
    def rewind(self):
        # TODO: what if mode == 'append'?  The next call to dfile()
        # will reopen at the end of the file.
        self.close() 
    def shortrepr(self):
        return "%s (%s)" % (self.filename, self.format.string())

registeredclass.Registration(
    'Data File',
    OutputDestination,
    DataFileOutput,
    rewindable=True,
    params=[
        filenameparam.WriteFileNameParameter(
            "filename", tip=parameter.emptyTipString),
        filenameparam.WriteModeParameter("mode", tip=parameter.emptyTipString),
        enum.EnumParameter('format', datafile.DataFileFormat, datafile.ASCII,
                           tip="Format of the file.")
        ],
    ordering=3)
        
        
##############

class OutputDestinationParameter(parameter.RegisteredParameter):
    def __init__(self, name, value=None, default=None, tip=None):
        parameter.RegisteredParameter.__init__(
            self, name=name, reg=OutputDestination,
            value=value, default=default, tip=tip)
    def clone(self):
        return self.__class__(self.name, self.value, self.default, self.tip)
