// -*- C++ -*-
// $RCSfile: cskeleton.C,v $
// $Revision: 1.126 $
// $Author: langer $
// $Date: 2010-12-06 04:11:47 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@nist.gov. 
 */


#include <oofconfig.h>

#include "common/cmicrostructure.h"
#include "common/geometry.h"
#include "common/lock.h"
#include "common/pixelsetboundary.h"
#include "common/random.h"
#include "engine/cskeleton.h"
#include "engine/doublevec.h"
#include "engine/mastercoord.h"
#include "engine/ooferror.h"
#include <algorithm>
#include <iostream>
#include <math.h>

#ifdef DEBUG
#include <assert.h>
#endif

#include <time.h>

#if DIM==3
#include "vtk-5.0/vtkPolyData.h"
#include "vtk-5.0/vtkGeometryFilter.h"
#include "vtk-5.0/vtkPointsProjectedHull.h"
#include "vtk-5.0/vtkCellData.h"
#include "vtk-5.0/vtkCell.h"
#include "vtk-5.0/vtkGenericCell.h"
#include "vtk-5.0/vtkPlane.h"
#include "vtk-5.0/vtkTriangle.h"
#include "vtk-5.0/vtkPolygon.h"
#include "vtk-5.0/vtkMath.h"
#include "vtk-5.0/vtkDoubleArray.h"
#include "vtk-5.0/vtkQuad.h"
#include "vtk-5.0/vtkLine.h"
#include "vtk-5.0/vtkFacetWriter.h"
#include "vtk-5.0/vtkCellArray.h"
#endif


#define LEGAL_ELEMENT_AREA_TOLERANCE 1.0e-8
#define COS_SQUARED_TOLERANCE 1.0e-13

const unsigned char CSkeletonNode::xmovable_ = 1;
const unsigned char CSkeletonNode::ymovable_ = 2;
#if DIM==3
const unsigned char CSkeletonNode::zmovable_ = 4;
const unsigned char CSkeletonNode::unpinned_ = 8;
#else
const unsigned char CSkeletonNode::unpinned_ = 4;
#endif

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Set of perturbation directions for moving nodes.  This is a set of
// nonhorizontal, nonvertical unit vectors which are candidate
// directions for perturbing nodes when searching for intersections
// with pixel-boundary segments.  The exclusion of the i=0 case is
// deliberate.  Nodes are perturbed so that they won't lie directly on
// top of pixel boundaries, which complicates the intersection finding
// code.

// This should be prime, and greater than one more than the number of
// sides of the most-sided possible element.
#if DIM==2
#define NODE_PERTURB_DIRECTIONS 7


// node_perturbations() returns a Coord* instead of a
// std::vector<Coord> because it's just a little bit faster.  It's
// used in a context (in categoryAreas()) where speed is important.

static const Coord* node_perturbations() {
  static Coord *perturbation_list = 0;
  if(!perturbation_list) {
    // This shows up as a memory leak in valgrind.  It's not, really.
    perturbation_list = new Coord[NODE_PERTURB_DIRECTIONS-1];
    double delta = 2.0*M_PI/((double) NODE_PERTURB_DIRECTIONS);
    for(int i=1; i<NODE_PERTURB_DIRECTIONS; i++) {
      double angle = i*delta;
      perturbation_list[i-1] = Coord(cos(angle), sin(angle));
    }
  }
  return perturbation_list;
}
#endif  
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

long CSkeletonNode::globalNodeCount = 0; // used for code testing
static SLock globalNodeCountLock;

#if DIM==2
CSkeletonNode::CSkeletonNode(double x, double y)
  : mobility(xmovable_|ymovable_|unpinned_),
    position_(x, y)
{
  globalNodeCountLock.acquire();
  ++globalNodeCount;
  globalNodeCountLock.release();
}
#elif DIM==3
CSkeletonNode::CSkeletonNode(double x, double y, double z, vtkPoints *p, int idx)
    : mobility(xmovable_|ymovable_|zmovable_|unpinned_),
      index_(idx)
{
  points=p; points->InsertPoint(idx,x,y,z);
  globalNodeCountLock.acquire();
  ++globalNodeCount;
  globalNodeCountLock.release();
}
#endif

CSkeletonNode::~CSkeletonNode() {
  globalNodeCountLock.acquire();
  --globalNodeCount;
  globalNodeCountLock.release();
}

void CSkeletonNode::setMobilityX(bool mob) {
  if(mob)
    mobility |= xmovable_;
  else
    mobility &= ~xmovable_;
}

void CSkeletonNode::setMobilityY(bool mob) {
  if(mob)
    mobility |= ymovable_;
  else
    mobility &= ~ymovable_;
}

#if DIM==3
void CSkeletonNode::setMobilityZ(bool mob) {
  if(mob)
    mobility |= zmovable_;
  else
    mobility &= ~zmovable_;
}
#endif

void CSkeletonNode::setPinned(bool pin) {
  if(pin)
    mobility &= ~unpinned_;
  else
    mobility |= unpinned_;
}

// moveTo and moveBy should not be called directly, except by a
// Skeleton or DeputySkeleton object.  Everyone else should use
// Skeleton.moveNodeTo, etc.

bool CSkeletonNode::moveTo(const Coord *pos) {
  lastmoved = nodemoved;
#if DIM==2
  lastposition_ = position_;
  position_ = *pos;
  if(!movable_x()) 
    position_.x = lastposition_.x;
  if(!movable_y())
    position_.y = lastposition_.y;
  if(position_ != lastposition_) {
    ++nodemoved;
    return true;
  }
#elif DIM==3
  double x,y,z,currentpos[3];
  points->GetPoint(index_,currentpos);
  points->GetPoint(index_,lastposition_);
  if(movable_x())
    x = pos->x;
  else
    x = currentpos[0];
  if(movable_y())
    y = pos->y;
  else
    y = currentpos[1];
  if(movable_z())
    z = pos->z;
  else
    z = currentpos[2];
  points->SetPoint(index_,x,y,z);
  if( !(x==currentpos[0] && y==currentpos[1] && z==currentpos[2]) ) {
    ++nodemoved;
    return true;
  }
#endif
  return false;
};


bool CSkeletonNode::canMoveTo(const Coord *pos) const {
#if DIM==2
  return ((movable_x() || position_.x == pos->x) &&
	  (movable_y() || position_.y == pos->y));
#elif DIM==3
  double currentpos[3];
  points->GetPoint(index_,currentpos);
  return ((movable_x() || currentpos[0] == pos->x) &&
	  (movable_y() || currentpos[1] == pos->y) &&
	  (movable_z() || currentpos[2] == pos->z));
#endif
}


void CSkeletonNode::unconstrainedMoveTo(const Coord *pos) {
  lastmoved = nodemoved;
#if DIM==2
  lastposition_ = position_;
  position_ = *pos;
#elif DIM==3
  points->GetPoint(index_,lastposition_);
  double x,y,z;
  x = pos->x;
  y = pos->y;
  z = pos->z;
  points->SetPoint(index_,x,y,z);
#endif
  ++nodemoved;
}

bool CSkeletonNode::moveBy(const Coord *shift) {
  lastmoved = nodemoved;
#if DIM==2
  lastposition_ = position_;
  Coord target = position_ + *shift;
  position_ = target;
  if(!movable_x())
    position_.x = lastposition_.x;
  if(!movable_y())
    position_.y = lastposition_.y;
  if(position_ != lastposition_) {
    ++nodemoved;
    return true;
  }
#elif DIM==3
  double x,y,z,currentpos[3];
  points->GetPoint(index_,currentpos);
  points->GetPoint(index_,lastposition_);
  if(movable_x())
    x = currentpos[0]+shift->x;
  else
    x = currentpos[0];
  if(movable_y())
    y = currentpos[1]+shift->y;
  else
    y = currentpos[1];
  if(movable_z())
    z = currentpos[2]+shift->z;
  else
    z = currentpos[2];
  points->SetPoint(index_,x,y,z);
  if( !(x==currentpos[0] && y==currentpos[1] && z==currentpos[2]) ) {
    ++nodemoved;
    return true;
  }
#endif
  return false;
}

void CSkeletonNode::moveBack() {
  nodemoved = lastmoved;
#if DIM==2
  position_ = lastposition_;
#elif DIM==3
  points->SetPoint(index_,lastposition_);
#endif
}

// Can this node be moved to and merge with another node?
bool CSkeletonNode::canMergeWith(const CSkeletonNode *other) const {
#if DIM==2
  bool xmovable = movable_x();
  bool ymovable = movable_y();
  if(!xmovable && !ymovable)
    return false;
  else if(!xmovable && ymovable)
    return position_.x == other->position_.x;
  else if(xmovable && !ymovable)
    return position_.y == other->position_.y;
  return true;
#elif DIM==3
  bool movables[3] = {movable_x(), movable_y(), movable_z()};
  bool canmerge = 1;
  double pos[DIM], otherpos[DIM];
  points->GetPoint(index_, pos);
  points->GetPoint(other->index_, otherpos);
  for(int i=0; i<DIM; i++) 
    canmerge &= ( movables[i] || pos[i] == otherpos[i]); 
  return canmerge;
#endif
}

long get_globalNodeCount() {
  return CSkeletonNode::globalNodeCount;
}

// Coord CSkeletonNode::random_position(double vx, double vy) const
// {
//   return position_ + Coord(gasdev()*vx, gasdev()*vy);
// }

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Is pnt inside the triangle (n0, n1, n2)?

#if DIM==2

static bool insideTriangle(const Coord &n0, const Coord &n1,
			   const Coord &n2, const Coord &pnt)
{
  Coord r0(n0 - pnt);
  Coord r1(n1 - pnt);
  Coord r2(n2 - pnt);
  return r0%r1 >= 0.0 && r1%r2 >= 0.0 && r2%r0 >= 0.0; // cross products
}

static double triangleArea(const Coord &v0, const Coord &v1, const Coord &v2) {
  return 0.5*(v1 - v0)%(v2 - v0);
}

#elif DIM==3

static double triangleArea(const Coord &v0, const Coord &v1, const Coord &v2) {
  return 0.5*sqrt(norm2((v1 - v0)%(v2 - v0)));
}

static double triangleArea2(const Coord &v0, const Coord &v1, const Coord &v2) {
  return 0.25*norm2((v1 - v0)%(v2 - v0));
}

#endif

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

#if DIM == 3
vtkCutter* CSkeletonElement::edgeExtractor = vtkCutter::New();
vtkMergePoints* CSkeletonElement::locator = vtkMergePoints::New();
vtkCellLocator* CSkeletonElement::cell_locator = vtkCellLocator::New();
#endif

long CSkeletonElement::globalElementCount = 0;
static SLock globalElementCountLock;

long get_globalElementCount() {
  return CSkeletonElement::globalElementCount;
}

CSkeletonElement::~CSkeletonElement() {
  globalElementCountLock.acquire();
  --globalElementCount;
  globalElementCountLock.release();
#if DIM == 3
  cell->Delete();
#endif
}


CSkeletonElement::CSkeletonElement(int n)
  :nodes(n)
{
  globalElementCountLock.acquire();
  ++globalElementCount;
  globalElementCountLock.release();

#if DIM == 3
//   edgeExtractor = vtkCutter::New();
  if(!edgeExtractor->GetLocator()) {
    edgeExtractor->CreateDefaultLocator();
//   // The number of divisions makes allocating and deallocating the
//   // locator very slow.  The default is 50, 50, 50
    edgeExtractor->GetLocator()->SetDivisions(8,8,8);
  }
 
//   // we use MergePoints instead of the parent class PointLocator.
//   // We only want to merge points if they are exactly coincident.
//   // PointLocator merges them based on a tolerance in the distance
//   // between their global coordinates.  This is not good since we
//   // want the user to be able to use arbitrary units.  We set the 
//   // number of buckets to be very small since we are dealing with
//   // a small number of points.
//   locator = vtkMergePoints::New();
  locator->SetDivisions(8,8,8);

#endif
}

void CSkeletonElement::replaceNode(int which, CSkeletonNode *replacement) {
  nodes[which] = replacement;
}

std::vector<Coord> *CSkeletonElement::perimeter() const {
  int n = nnodes();
  std::vector<Coord> *pvec = new std::vector<Coord>(n);
  for(int i=0; i<n; i++)
    (*pvec)[i] = nodes[i]->position();
  return pvec;
}

double CSkeletonElement::perimeterLength() const {
  int n = nnodes();
  double p = 0.0;
  for(int i=0; i<n; i++) {
    int j = (i+1)%n;
    Coord r(nodes[i]->position() - nodes[j]->position());
    p += sqrt(norm2(r));
  }
  return p;
}

double CSkeletonElement::perimeterSquared() const {
  int n = nnodes();
  double p = 0.0;
  for(int i=0; i<n; i++) {
    int j = (i+1)%n;
    Coord r(nodes[i]->position() - nodes[j]->position());
    p += norm2(r);
  }
  return p;
}

double CSkeletonElement::edgeLength(int which) const {
  Coord r(nodes[which]->position() - nodes[(which+1)%nnodes()]->position());
  return sqrt(norm2(r));
}

double CSkeletonElement::edgeLengthSquared(int which) const {
  Coord r(nodes[which]->position() - nodes[(which+1)%nnodes()]->position());
  return norm2(r);
}

#if DIM==2
double CSkeletonElement::cosCornerAngleSquared(int which) const {
  int n = nnodes();
  Coord p0 = nodes[which]->position();
  Coord side0 = nodes[(which+n-1)%n]->position() - p0;
  Coord side1 = nodes[(which+1)%n]->position() - p0;
  double sdots = dot(side0, side1);
  double cos2 = sdots*sdots/(norm2(side0)*norm2(side1));
  if(cos2 > 1.0)
    return 1.0;
  return cos2;
}

double CSkeletonElement::cosCornerAngle(int which) const {
  int n = nnodes();
  Coord p0 = nodes[which]->position();
  Coord side0 = nodes[(which+n-1)%n]->position() - p0;
  Coord side1 = nodes[(which+1)%n]->position() - p0;
  double cosAngle = dot(side0, side1)/sqrt(norm2(side0)*norm2(side1));
  // Due to a round-off error, this step is needed.
  // (Sometimes cosAngle that should be -1 turns out to be -1.000000001)
  if (cosAngle > 1.0)
    return 1.0;
  else if (cosAngle < -1.0)
    return -1.0;
  else
    return cosAngle;
}

double CSkeletonElement::getRealAngle(int which) const {
  int n = nnodes();
  Coord p0 = nodes[which]->position();
  Coord side0 = nodes[(which+n-1)%n]->position() - p0;
  Coord side1 = nodes[(which+1)%n]->position() - p0;
  return atan2(side1%side0, dot(side1,side0))*180./M_PI;
}
#elif DIM==3

double CSkeletonTetra::cosCornerAngle(int face, int corner) const {
  double p0[3], p1[3], p2[3], s0[3], s1[3];
  vtkPoints *cellPoints = cell->GetPoints();
  int *ptIds;
  cell->GetFacePoints(face, ptIds);
  cellPoints->GetPoint(ptIds[corner], p0);
  cellPoints->GetPoint(ptIds[(corner+3-1)%3], p1);
  cellPoints->GetPoint(ptIds[(corner+1)%3], p2);
  for(int i=0; i<3; i++) {
    s0[i] = p1[i]-p0[i];
    s1[i] = p2[i]-p0[i];
  }
  double cosAngle = vtkMath::Dot(s0,s1)/(vtkMath::Norm(s0)*vtkMath::Norm(s1));
  if (cosAngle > 1.0)
    return 1.0;
  else if (cosAngle < -1.0)
    return -1.0;
  else
    return cosAngle;
}

double CSkeletonTetra::cosCornerAngleSquared(int face, int corner) const {
  double p0[3], p1[3], p2[3], s0[3], s1[3], dist02=0, dist12=0;
  vtkPoints *cellPoints = cell->GetPoints();
  int *ptIds;
  cell->GetFacePoints(face, ptIds);
  cellPoints->GetPoint(ptIds[corner], p0);
  cellPoints->GetPoint(ptIds[(corner+3-1)%3], p1);
  cellPoints->GetPoint(ptIds[(corner+1)%3], p2);
  for(int i=0; i<3; i++) {
    s0[i] = p1[i]-p0[i];
    dist02 += s0[i]*s0[i];
    s1[i] = p2[i]-p0[i];
    dist12 += s1[i]*s1[i];
  }
  double sdots = vtkMath::Dot(s0,s1);
  double cos2 = sdots*sdots/(dist02*dist12);
  if(cos2 > 1.0 || (1-cos2) < COS_SQUARED_TOLERANCE)
    return 1.0;
  return cos2;
}

// Useful topological maps.  These are the same for all tetrahedra.

// This maps each node to the index of the opposite face
const int CSkeletonTetra::nodeToOppFaceMap[4] = {1,2,0,3};

// Maps face id to edge ids in order
const int CSkeletonTetra::faceToEdgesMap[][3] = {
  {0, 4, 3},
  {1, 5, 4},
  {2, 3, 5},
  {2, 1, 0}
};

// Maps the direction each edge relative to it's face to the direction
// of the edge relative to the whole tet.  That is tet edge 2 is
// defined as node 2 to node 0 however for face 3, in order for the
// edges to be defined in a way such that the normal points out, this
// edge is directed from node 0 to node 2.
const int CSkeletonTetra::faceToEdgeDirMap[][3] = {
  {1, 1, -1},
  {1, 1, -1},
  {1, 1, -1},
  {-1, -1, -1}
};

#define PI 4.0*atan(1.0)

double CSkeletonTetra::solidCornerAngle(int corner) const {
  // We want to be able to use the sign to determine whether the solid
  // angle is above or below pi, so we must order the three other
  // points such that the norm of the face they define points out.
  // The formula is from Oosterom and Strackee (IEEE
  // Trans. Biom. Eng., Vol BME-30, No 2, 1983) and gives the tangent
  // of half of the solid angle.
  double x[3], s[3][3], c[3], n[3], d0, d1, d2, triple;
  cell->GetPoints()->GetPoint(corner, c);
  vtkPoints *facePoints = cell->GetFace(nodeToOppFaceMap[corner])->GetPoints();
  for(int i = 0; i<3; i++) {
    facePoints->GetPoint(i, x);
    for(int k = 0; k<3; k++)
      s[i][k] = x[k]-c[k];
    n[i] = vtkMath::Norm(s[i]);
  }
  vtkMath::Cross(s[1],s[2],x);
  triple = vtkMath::Dot(s[0],x);
  d0 = vtkMath::Dot(s[1],s[2]);
  d1 = vtkMath::Dot(s[0],s[2]);
  d2 = vtkMath::Dot(s[0],s[1]);
  // we take the arctangent here because tangents of angles are no
  // good for comparison.
  double tanhalf = triple/(n[0]*n[1]*n[2]+d0*n[0]+d1*n[1]+d2*n[2]);
  if(tanhalf >= 0)
    return 2*atan(tanhalf);
  else
    return 2*(PI-atan(tanhalf));
} 

double CSkeletonTetra::cosDihedralAngle(int face1, int face2) const {
  double n1[3], n2[3];
  int ptids[3] = {0,1,2};
  vtkPoints *face1Points = cell->GetFace(face1)->GetPoints();
  vtkTriangle::ComputeNormal(face1Points,3,ptids,n1);
  vtkPoints *face2Points = cell->GetFace(face2)->GetPoints();
  vtkTriangle::ComputeNormal(face2Points,3,ptids,n2);
  return - vtkMath::Dot(n1,n2);
}
  

#endif

Coord CSkeletonElement::center() const {
  int n = nnodes();
  Coord c = nodes[0]->position();
  for(int i=1; i<n; i++)
    c += nodes[i]->position();
  return (1./n)*c;
}

Coord CSkeletonElement::frommaster(MasterCoord *mc, int rotation) const {
  int n = nnodes();
  Coord result(0., 0.);
  for(int i=0; i<n; i++)
    result += shapefun(i, *mc) * nodes[(i+rotation)%n]->position();
  return result;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// CategoryArea calculation, and related infrastructure.

typedef std::multimap<double, PixelBdyIntersection> ElEdgeMap;
typedef std::vector<ElEdgeMap> ElEdge;
typedef std::pair<double, PixelBdyIntersection> ElEdgeDatum;

typedef std::multimap<Coord, PixelBdyIntersection> CoordIsec;
typedef std::pair<Coord, PixelBdyIntersection> CoordIsecDatum;

// Utility function for deleting an intersection from one of the
// ordered lists of intersections.  Intersections have sufficient data
// to locate themselves in these lists.
void delete_isec(const PixelBdyIntersection &pbi,
		 ElEdge &eledge) {
  int ni = pbi.get_element_node_index();
  double fr = pbi.get_element_fraction();
  
  int count = eledge[ni].count(fr);
  if (count==1) {
    eledge[ni].erase(fr);
  }
  else {
    ElEdgeMap::iterator emi = eledge[ni].lower_bound(fr);
    for(int i=0; i<count; ++emi, ++i) {
      if (((*emi).second.location==pbi.location) && 
	  ((*emi).second.entry==pbi.entry) ) {
	eledge[ni].erase(emi);
	break;
      }
    }
  }
}

//================

// Another utility function, this one for checking whether a point is
// interior or exterior to the element, using the perturbation-hint to
// resolve ambiguous cases.  If the cross-product of any element
// segment with the candidate point minus the start of the segment is
// negative, it means the point is exterior to that segment.  A point
// is exterior to the element if it is exterior to any segment.

// This function is used heavily during the homogeneity calculation.
// Rewriting it so that it uses doubles instead of Coords did not
// improve the speed.  (But that was before replacing
// std::vector<Coord> with a simple array...)
bool CSkeletonElement::interior(const Coord &pt, 
				const Coord &perturb) const
{
#if DIM==2
  int nn = nnodes();
  if(nn==3) {			// Loops unrolled
    Coord pos0 = nodes[0]->position();
    Coord pos1 = nodes[1]->position();
    Coord pos2 = nodes[2]->position();

    Coord edge = pos1 - pos0;
    double cross = edge % (pt - pos0);
    if(cross < 0.0)
      return false;
    if(cross == 0) {
      if(perturb % edge < 0.0)
	return false;
    }
    edge = pos2 - pos1;
    cross = edge % (pt - pos1);
    if(cross < 0.0)
      return false;
    if(cross == 0) {
      if(perturb % edge < 0.0)
	return false;
    }
    edge = pos0 - pos2;
    cross = edge % (pt - pos2);
    if(cross < 0.0)
      return false;
    if(cross == 0) {
      if(perturb % edge < 0.0)
	return false;
    }
    return true;
  }

  else if(nn == 4) {
    Coord pos0 = nodes[0]->position();
    Coord pos1 = nodes[1]->position();
    Coord pos2 = nodes[2]->position();
    Coord pos3 = nodes[3]->position();

    Coord edge = pos1 - pos0;
    double cross = edge % (pt - pos0);
    if(cross < 0.0)
      return false;
    if(cross == 0) {
      if(perturb % edge < 0.0)
	return false;
    }
    edge = pos2 - pos1;
    cross = edge % (pt - pos1);
    if(cross < 0.0)
      return false;
    if(cross == 0) {
      if(perturb % edge < 0.0)
	return false;
    }
    edge = pos3 - pos2;
    cross = edge % (pt - pos2);
    if(cross < 0.0)
      return false;
    if(cross == 0) {
      if(perturb % edge < 0.0)
	return false;
    }
    edge = pos0 - pos3;
    cross = edge % (pt - pos3);
    if(cross < 0.0)
      return false;
    if(cross == 0) {
      if(perturb % edge < 0.0)
	return false;
    }
    return true;
  }

  else {
    // This is the generic case, for nn!=3 and nn!=4, which will
    // probably never occur.  
    Coord *pos = new Coord[nn];
    bool result = true;
    for(int ni=0; ni<nn; ++ni) {
      pos[ni] = nodes[ni]->position();
    }

    for(int ni=0; ni<nn; ++ni) {
      const Coord &posni = pos[ni];
      Coord edge = pos[(ni+1)%nn] - posni;
      double cross = edge % (pt-posni);
      if (cross < 0) {
	result = false;
	break;
      }
      else if (cross==0) {
	if (perturb % edge < 0) {
	  result = false;
	  break;
	}
#ifdef DEBUG
	if (perturb % edge == 0) {
	  std::cerr << "Ambiguous interiority after perturbation!" 
		    << std::endl;
	}
#endif // DEBUG
      }
    }
    delete[] pos;
    return result;
  }
#endif
  return 0;
}

//==============

CRectangle CSkeletonElement::bbox() const {
  CRectangle bounds(nodes[0]->position(), nodes[1]->position());
  for(int i=2; i<nnodes(); i++)
    bounds.swallow(nodes[i]->position());
  return bounds;
}

//=============

// Returns a vector of doubles.  Each double in this vector is equal
// to the area of the corresponding microstructure category which is
// within this element.  This version uses category boundary
// intersections, but has the same API as other methods of computing
// the categoryArea.
#if DIM==2
const DoubleVec * CSkeletonElement::categoryAreas(const CMicrostructure &ms)
  const
{

  // std::cerr << std::endl << "CSkeletonElement::categoryAreas" << std::endl;

  // Redundant call to illegal().
  //   if (illegal()) {
  //     throw ErrProgrammingError("categoryAreas called on illegal element.",
  // 			      __FILE__, __LINE__);
  //   }

  int n = ms.nCategories();	// recomputes categories & boundaries if needed
  DoubleVec *result = new DoubleVec(n);
  bool areas_zero = true;
  int nn = nodes.size();

  //   std::cerr << "Nodes and diffs: " << std::endl;
  //   for(int di=0; di<nn; ++di) {
  //     std::cerr << nodes[di]->position();
  //     std::cerr << " " << nodes[(di+1)%nn]->position()-nodes[di]->position() 
  //  	      << std::endl;
  //   }


  // Build up element data for easy passage to intersection-finder.
  //   std::vector<Coord> element_points(nn);
  Coord *element_points = new Coord[nn];
  for(int nj=0; nj<nn; ++nj) {
    element_points[nj] = nodes[nj]->position();
  }

  // For perturbations, select the first one which is not parallel to
  // any element segment.  The set is already guaranteed to be
  // nonhorizontal and nonvertical, so it can't be parallel to the
  // pixel-set boundary.
  Coord element_perturb;
  const Coord *const ptbs = node_perturbations();
  for(int pt=0; pt<NODE_PERTURB_DIRECTIONS-1; ++pt) {
    bool parallel = false;
    for(int el=0;el<nn;++el) {
      Coord v = element_points[(el+1)%nn]-element_points[el];
      if ((ptbs[pt] % v) == 0 ) {
	parallel = true;
	break;
      }
    }
    if (parallel==false) {
      element_perturb = ptbs[pt];
      break;
    }
  }

#ifdef DEBUG
  assert(element_perturb!=Coord(0.0,0.0));
#endif

  // Get all the category boundaries.
  const std::vector<PixelSetBoundary> &bdys = ms.getCategoryBdys();

#ifdef DEBUG
  assert(int(bdys.size())==n);
#endif

  // Per-element-segment intersection data.
  PixelBdyIntersection pbi;

  //    std::cerr << std::endl << std::endl 
  //  	    << "Element bounded by: " << std::endl;
  //    for(int i=0;i<nn;i++) {
  //      std::cerr << element_points[i] << " -> " 
  //  	      << element_points[(i+1)%nn] << std::endl;
  //    }
  //    std::cerr << "Perturbation is " << element_perturb << std::endl;

  for(std::vector<PixelSetBoundary>::size_type i=0; i<bdys.size(); i++) {

    // std::cerr << "Examining a new category, " << i << std::endl;
    double accum_area = 0.0;
    const std::vector<PixelBdyLoop*> &loops = bdys[i].get_loops();
   

    // Store intersection data -- used to figure out what sections
    // of the element boundaries contribute to the area.  Stored a
    // couple of ways -- firstly by element edge, so it can be
    // easily traversed, and secondly indexed by coordinate, so
    // duplicates can be removed.
    ElEdge eledgedata(nn);  // vector of multimap<double, PixelBdyIntersection>
    CoordIsec coordisecs;   // multimap<Coord, PixelBdyIntersection>
    
    for(std::vector<PixelBdyLoop*>::const_iterator pbl = loops.begin();
	pbl!=loops.end(); ++pbl) {
      
      // std::cerr << "Starting a new loop of this category." << std::endl;

      // Skip this loop if its bounding box doesn't intersect the
      // element's bounding box.
      if(!bbox().intersects((*pbl)->bbox()))
	continue;

      std::vector<PixelBdySegment>::const_iterator k = 
	(*pbl)->segments().begin();
      std::vector<PixelBdySegment>::const_iterator kend =
	(*pbl)->segments().end();

      // Find the interiority of the start of the first pixel boundary
      // segment.  End-point interiorities are done on the fly in the
      // k loop.
      bool pbs_start_inside = interior((*k).start_pt(), element_perturb);
      bool pbs_end_inside;

      // std::cerr << "Start point is " << (pbs_start_inside ? "" : "not ") 
      // 	<< "interior." << std::endl;

      
      for( ; k!=kend; ++k) {	// Loop over PixelBdySegments
	pbs_end_inside = interior((*k).end_pt(), element_perturb);
	
	// std::cerr << "Checking pixel bdy segment " << *k << std::endl;


	if (pbs_start_inside && pbs_end_inside) {
	  // Pixel boundary segment is wholly interior, add the area.
	  // std::cerr << "Wholly interior bit, adding " << (*k).start_pt()  << " to " << (*k).end_pt() << ", value is " << (*k).start_pt() % (*k).end_pt() << std::endl;
	  accum_area += (*k).start_pt() % (*k).end_pt();
	}
	// If start and end are hetero-interior, so to speak, then
	// there's an intersection.  Find it.
	else if ( (pbs_start_inside && !pbs_end_inside) ||
		  (!pbs_start_inside && pbs_end_inside) ) {
	  int seg = (*k).find_one_intersection(element_points, nn,
					       pbi, pbs_end_inside);

	  // std::cerr << "One-intersection case, segment is " << *k << ", intersection is at " << pbi.location << std::endl;

	  double f = norm2(pbi.location - element_points[seg]);
	  pbi.set_element_data(seg,f);
	  eledgedata[seg].insert(ElEdgeDatum(f,pbi));

	  coordisecs.insert(CoordIsecDatum(pbi.location, pbi));

	  // Accumulate the area of the interior portion.
	  if (pbs_start_inside) {
	    // std::cerr << "Inside-to-outside, adding " << (*k).start_pt() << " to " << pbi.location << ", value is " << (*k).start_pt() % pbi.location << std::endl;
	    accum_area += (*k).start_pt() % pbi.location;
	  }
	  else {
	    // std::cerr << "Outside-to-inside, adding " << pbi.location << " to " << (*k).end_pt() << ", value is " << pbi.location % (*k).end_pt() << std::endl;
	    accum_area += pbi.location % (*k).end_pt();
	  }
	}
	else if ( !pbs_start_inside && !pbs_end_inside ) {
	  // The current segment has both endpoints outside the
	  // element.  It must intersect zero or two times.  Find out.

	  bool no_isecs = (*k).find_no_intersection(element_points, nn,
						    element_perturb);

	  if (!no_isecs) {
	    // We know there have to be two intersections, find them.
	    PixelBdyIntersection isec0, isec1;

	    // Entry first.
	    int s1 = (*k).find_one_intersection(element_points, nn,
						isec0, true);
	    double f =  norm2(isec0.location-element_points[s1]);
	    isec0.set_element_data(s1,f);
	    eledgedata[s1].insert(ElEdgeDatum(f,isec0));
	    
	    // Then find the exit.
	    int s2 = (*k).find_one_intersection(element_points, nn,
						isec1, false);
	    f = norm2(isec1.location-element_points[s2]);
	    isec1.set_element_data(s2,f);
	    eledgedata[s2].insert(ElEdgeDatum(f,isec1));


	    // Measure distance along the segment, for
	    // coincidence-checking and sorting.
	    double d1 = norm2(isec0.location - (*k).start_pt());
	    double d2 = norm2(isec1.location - (*k).start_pt());

	    // If two intersections coincide, they should annihilate.
	    if (d1==d2) {
	      if (isec0.entry != isec1.entry) {
		// std::cerr << "Annihilation." << std::endl;
		delete_isec(isec0, eledgedata);
		delete_isec(isec1, eledgedata);
	      }
	      else {
		throw ErrProgrammingError("CategoryAreas: Coincident intersections do not annihilate!", __FILE__, __LINE__);
	      }
	    }
	    // The intersections are not coincident.
	    else {
	      // Order the intersections.  Topologically, the first
	      // one has to be an entry, but if they're very close
	      // together, they could come out in the wrong order by
	      // roundoff.  If this happens, they're within roundoff
	      // of each other, and the area is zero -- delete them.
	      
	      PixelBdyIntersection &one = ( d1 < d2 ? isec0 : isec1 );
	      PixelBdyIntersection &two = ( d1 < d2 ? isec1 : isec0 );
	      
	      // Expected case, first intersection is an entry.  Add
	      // them to the coordisecs structure, and accumulate the
	      // area.
	      if (one.entry) { 
		// std::cerr << "Segment " << (*k).start_pt() << " to " << (*k).end_pt() << std::endl;
		coordisecs.insert(CoordIsecDatum(one.location, one));
		coordisecs.insert(CoordIsecDatum(two.location, two));
		// std::cerr << "Interior section, adding " << one.location << " to " << two.location << ", value is " << one.location % two.location << std::endl;
		accum_area += one.location % two.location;
	      }
	      // Pathological case -- if they're in the topologically
	      // disallowed order, it must be because of roundoff, and
	      // so they must be within roundoff of each other.
	      // Remove them.
	      else {
		delete_isec(one, eledgedata);
		delete_isec(two, eledgedata);
	      }
	    } // End of non-coincident, two-intersection case.

	  } // End of conditional based on find_no_intersections result.

	} // End of both-endpoints-exterior block.
  
	// Hand off status info in preparation for next pixel bdy segment.
	pbs_start_inside = pbs_end_inside;
      } // End of loop over segments.
    } // End of loop over pixel-boundary loops.  
      

    //     // Debugging -- list all intersections.
    //     std::cerr << std::endl << "Pre-annihilation, all intersections: " 
    // 	      << std::endl;
    //     for(int ni = 0; ni<nn; ++ni) {
    //       for(ElEdgeMap::iterator there = eledgedata[ni].begin();
    //  	  there!=eledgedata[ni].end(); ++there) {
    //  	std::cerr << "Intersection: " << (*there).second.location << 
    //  	  ((*there).second.entry ? ", Entry" : ", Exit" ) << std::endl;
    //       }
    //     }
    

    // At this point, we have accumulated all the area contributions
    // from the segments which are boundaries of the pixel set, and
    // are interior to the current element.  It remains to traverse
    // the element exterior and add up the contributions from those
    // portions of it which are between an exit intersection and an
    // entry intersection.

    // But first... Check for multiple coincident intersections.
    // These can arise when the corner of an element and the corner of
    // a pixel boundary coincide, and the perturbation resolution is
    // such that the corner point of the pixel boundary set is inside
    // the element, but only by epsilon.  Because each intersection is
    // on a different element-segment/pixel-boundary-segment pair,
    // they don't get picked up earlier.  Coincident intersections
    // cannot be ordered, and so are confusing to the traversal code,
    // so we rationalize them here.

    CoordIsec::iterator x = coordisecs.begin();
    while(x!=coordisecs.end()) {
      Coord key = (*x).first;
      int ct = coordisecs.count(key);
      if (ct>1) {

#ifdef DEBUG
	// std::cerr << "Nontrivial coordisec run." << std::endl;
	// std::cerr << "Coord is " << key << std::endl;
	// std::cerr << "Count is " << ct << std::endl;
#endif

	CoordIsec::iterator w = x;

	// Set x to the start of the next key, for the next
	// iteration of the outer loop.
	x = coordisecs.upper_bound(key);

	// Determine the sign of the resultant intersection.  It may
	// be entry, exit, or zero, if everything cancels out.
	int entries = 0;
	int exits = 0;
	for(CoordIsec::iterator t = w; t!=x; ++t) {
	  ( (*t).second.entry ? entries++ : exits++ );
	}

#ifdef DEBUG
	// std::cerr << "Entries: " << entries << ", exits: " << exits << std::endl;
#endif
	
	if (entries > exits) {
	  bool got_entry = false;
	  for(CoordIsec::iterator t2 = w; t2!=x; ++t2) {
	    if ((*t2).second.entry && !got_entry)
	      got_entry = true;
	    else 
	      delete_isec((*t2).second, eledgedata);
	  }
	}
	else if (entries < exits) {
	  bool got_exit = false;
	  for(CoordIsec::iterator t2 = w; t2!=x; ++t2) {
	    if (!(*t2).second.entry && !got_exit)
	      got_exit = true;
	    else 
	      delete_isec((*t2).second, eledgedata);
	  }
	}
	else {
	  for(CoordIsec::iterator t2 = w; t2!=x; ++t2)
	    delete_isec((*t2).second, eledgedata);
	}
	
      }
      else { // Key only occurs once, move on.
	++x;
      }
    }

    // Now traverse the boundary of the element, adding in the area
    // contribution from all the portions that are between an exit
    // and an entry.
      
    int icount = 0;
    for(n=0;n<nn;n++) {
      icount += eledgedata[n].size();
    }

    if (icount > 0) {
      Coord starting_point;
      Coord last_exit;  // to Brooklyn
      int current_segment = 0;
      bool done=false;
      bool started=false;
      bool outside=false;
      ElEdgeMap::iterator here = eledgedata[0].begin();
      
      // Debugging -- review all intersections.

      //       std::cerr << std::endl << "All intersections: " << std::endl;
      //       for(int ni = 0; ni<nn; ++ni) {
      // 	std::cerr << "Edge " << ni << std::endl;
      //  	for(ElEdgeMap::iterator there = eledgedata[ni].begin();
      //    	    there!=eledgedata[ni].end(); ++there) {
      //    	  std::cerr << "Intersection: " << (*there).second.location << 
      //   	    ((*there).second.entry ? ", Entry" : ", Exit" ) << std::endl;
      //  	}
      //       }

      
#ifdef DEBUG
      int et_count = 0;
#endif 

      while(!done) {
#ifdef DEBUG
	et_count++;
	assert(et_count < 10000);
#endif
	// If you've reached the end of a segment, switch to the
	// next segment.  If you're collecting areas, collect from
	// the last exit to the end of the segment, and set
	// last_exit to the beginning of the new segment.
	if (here == eledgedata[current_segment].end() ) {
	  current_segment = (current_segment+1)%nn;
	  here = eledgedata[current_segment].begin();
	  if (started && outside) {
	    // Segment already incremented, end_pt is end of
	    // previous segment, start of new current segment.
	    Coord end_pt = nodes[current_segment]->position();
	    // std::cerr << "(1) Adding area from " << last_exit << " to " << end_pt << ", value is " << last_exit % end_pt << std::endl;
	    accum_area += last_exit % end_pt;
	    last_exit = end_pt;
	  }
	}
	// If you've found an exit point, first check if you're
	// already outside.  If so, add the area from the last exit
	// point to this one.  Then, check if it's the initial exit
	// point -- if so, you're done.  Otherwise, update the
	// last_exit and set the "outside" marker to true.  If you
	// haven't started yet, then start, and set the starting
	// point.
	else if (!(*here).second.entry) {
	  if (started && outside) {
	    // std::cerr << "(2) Adding area from " << last_exit << " to " << (*here).second.location << ", value is " << last_exit % (*here).second.location << std::endl;
	    accum_area += last_exit % (*here).second.location;
	  }
	  if (started && (*here).second.location==starting_point ) {
	    done = true;
	  }
	  else {
	    if (!started) {
	      started = true;
	      starting_point = (*here).second.location;
	    }
	  }
	  outside = true;
	  last_exit = (*here).second.location;
	  ++here;
	}
	// If you've found an entrance point, and you're outside, then
	// add the area from the last exit point to here, and set the
	// outside flag to false.  If you find an entrance point and
	// you're not outside, do nothing.
	else if ((*here).second.entry) {
	  if (started && outside) {
	    // std::cerr << "(3) Adding area from " << last_exit << " to " << (*here).second.location << ", value is " << last_exit % (*here).second.location << std::endl;
	    accum_area += last_exit % (*here).second.location;
	    outside = false;
	  }
	  ++here;
	}
      }
    }
    else { 
      // If there are no intersections, we either need to add all of
      // the element boundary, or none of it.  If the area so far is
      // negative, we need all of it.  Actually, we need 2x it,
      // because the cross-products double-count the area, and this
      // needs to play the same game later on.
      if (accum_area < 0.0) {
	// std::cerr << "No intersections on this element." << std::endl;
	accum_area += 2.0*area();
      }
    }

    // "accum_area" now contains 2x the contribution of all loops of this
    // category to this element.
    (*result)[i] = 0.5*accum_area;
    if (accum_area != 0.0) {
      areas_zero = false;
    }

  } // End of loop over categories.

  // If all of the areas are zero, then the element is homogeneous.
  // Figure out which category is the right one by getting the
  // category of the central point.  The central point is used
  // because, for a homogeneous element, round-off will not put it
  // into a different category.  This may end up being slow.
  if (areas_zero) {
    int cat = ms.category(center());
    (*result)[cat]=area();
  }
    
#ifdef DEBUG
  // TODO: Leave this test in, even in non DEBUG mode.  If the test
  // fails, return a null pointer instead of the vector of areas.
  // This will tell the calling routine that the element is nonsense,
  // and should be treated as an illegal element.
#define CSKELETON_PIXELBDY_AREA_TOL 1.0e-6
  double sum = 0;
  for(unsigned int ct=0; ct<result->size(); ct++) {
    // std::cerr << "Fractional area: " << (*result)[ct] << std::endl;
    sum += (*result)[ct];
  }
  double frac = (sum-area())/area();
  // The areas should be exactly equal, but round-off may make them
  // slightly unequal.  What we're actually looking for here is
  // evidence of some kind of book-keeping error or logic problem in
  // the foregoing intersection computation.  If this occurs, the
  // fractional area is likely to deviate by a large amount.

  //   std::cerr << "Sum: " << sum << ", Area: " << area() << std::endl;
  //   std::cerr << "Fraction: " << fabs(frac) << std::endl;
  assert( fabs(frac) < CSKELETON_PIXELBDY_AREA_TOL );
#endif

  delete[] element_points;
  // Minor hack -- round-off can make some values less than zero.
  // Clamp these to zero.
  for(DoubleVec::size_type ct=0; ct<result->size(); ct++) {
    if ((*result)[ct] < 0)
      (*result)[ct]=0;
  }
  return result;  // Caller must de-allocate this vector.
} // end of CSkeletonElement::categoryAreas

#elif DIM==3


// const int CSkeletonBrick::edgeIdMap[][4] = { 
//   {8, 7, 10, 3},
//   {1, 11, 5, 9},
//   {0, 9, 4, 8},
//   {10, 6, 11, 2},
//   {3, 2, 1, 0},
//   {4, 5, 6, 7}
// };

// const int CSkeletonBrick::dirMap[][4] = { 
//   {1, 1, -1, -1},
//   {1, 1, -1, -1},
//   {1, 1, -1, -1},
//   {1, 1, -1, -1},
//   {1, 1, -1, -1},
//   {1, 1, -1, -1}
// };


bool pointOnFaceEdge(int numPoints, double *weights, double tol) {
  // Helper function to tell us if one point is on an edge of the
  // element face.  Weights should come from running EvaluatePostion
  // on the face in question.
  // Actually, using the weights is problematic for very skinny elements.
  int i, contributingPoints = 0;
  for(i=0; i<numPoints; i++) {
    cout << weights[i] << endl;
    if(weights[i] > tol) {
      contributingPoints++;
    }
  }
  return ( contributingPoints <= 2 );
}

bool pointOnFaceEdge(vtkPoints *facePoints, double x[3], double linetol2) {
  // Helper function to tell us if one point is on an edge of the
  // element face.
  int i, n=facePoints->GetNumberOfPoints();
  double p0[3], p1[3], c[3], t;
  for(i=0; i<n; ++i) {
    facePoints->GetPoint(i,p0);
    facePoints->GetPoint((i+1)%n,p1);
    // This version of DistanceToLine calculates the distance to a
    // *finite* line between p0 and p1.
    if(vtkLine::DistanceToLine(x,p0,p1,t,c)<linetol2)
      return true;
  }
  return false;
}

bool edgeOnFaceEdge(int numPoints, double *weights1, double *weights2, double tol) {
  // Helper function for determining if two points are on the edge of
  // a face.  The weights should come from calling the vtk routine
  // "EvaluatePosition" on the relevant cell.  If two points are on
  // the same edge, they will have weights within tol for the same
  // point(s).  There will be only two points in the cell for which
  // both test points do not have zero weights.
  int i, mutuallyNonContributingPoints = 0;
  //cout << "in edgeOnFaceEdge" << endl;
  for(i=0; i<numPoints; i++) {
    //cout << weights1[i] << " " << weights2[i] << endl;
    if(weights1[i] <= tol && weights2[i] <= tol)
      mutuallyNonContributingPoints++;
  }
  return ( (numPoints-mutuallyNonContributingPoints) == 2 );
}

bool edgeOnFaceEdge(vtkPoints *facePoints, double x[3], double y[3], double linetol2) {
  // Helper function for determining if two points are on the edge of
  // a face.
  int i, n=facePoints->GetNumberOfPoints();
  double p0[3], p1[3];
  for(i=0; i<n; ++i) {
    facePoints->GetPoint(i,p0);
    facePoints->GetPoint((i+1)%n,p1);
    // This version of DistanceToLine calculates the distance to a
    // *infinite* line defined by p0 and p1.
    if( vtkLine::DistanceToLine(x,p0,p1)<linetol2 && vtkLine::DistanceToLine(y,p0,p1)<linetol2 )
      return true;
  }
  return false;
}
  

bool boundsOverlap(double bounds1[6], double bounds2[6]) {
  // bounds are in form (xmin,xmax,ymin,ymax,zmin,zmax)
  if(bounds1[0] > bounds2[1] || bounds2[0] > bounds1[1])
    return false;
  if(bounds1[2] > bounds2[3] || bounds2[2] > bounds1[3])
    return false;
  if(bounds1[4] > bounds2[5] || bounds2[4] > bounds1[5])
    return false;
  return true;
}


bool CSkeletonTetra::faceOnElementFace(double normal[3], double pt[3], double maxLength, double tol) const{

  double dist, x[3];
  int i, pointsOnFace=0;
	
  for(i=0; i<4; i++) {
    this->getPoints()->GetPoint(i,x);
    dist = vtkPlane::DistanceToPlane(pt,normal,x);

    if(fabs(dist)/maxLength<=tol)
      pointsOnFace++;
  }

  return (pointsOnFace>=3);

}


// bool CSkeletonBrick::faceOnElementFace(double normal[3], double pt[3], double maxLength, double tol) const{
//   // TODO 3D: implement when we allow brick elements
//   return 0;
// }

// vtk makes a points object for each cell.  When we change the
// positions of the points in the master points object, the cellwise
// point objects are not updated.  So we do it manually when we need
// it.  Updating the points for each cell (tetra, etc) should update the
// sub-cells (faces and edges).
void CSkeletonElement::updateVtkCellPoints() const{
  for(int i=0; i<nnodes(); i++) 
    this->getPoints()->SetPoint(i, nodes[i]->getPoints()->GetPoint(nodes[i]->index()));
}

void CSkeletonElement::findOneIntersection(CellEdge &edgeData, int numEdges, int faceId, double p1[3], double p2[3], double x[3], int type) const {

  // This routine is called when we need to find exactly one intersection
  // in our edgeData map. We choose the intersection in the map that is closest. 

//   int edgeId, k;
//   double y[3], z[3], t, dist2, mindist2 = -1.0;
//   CellEdgeMap::iterator bestIntersection;
 
//   for(k=0; k<numEdges; k++) {
//     edgeId = faceEdgeToElementEdge(faceId, k);
//     for(CellEdgeMap::iterator here = edgeData[edgeId].begin();
// 	here!=edgeData[edgeId].end(); ++here) {
//       (*here).second.location.writePointer(y);
//       dist2 = vtkLine::DistanceToLine(y,p1,p2,t,z);
//       if(mindist2 < 0 || dist2 < mindist2) {
// 	mindist2 = dist2;
// 	bestIntersection = here;
//       }
//     }
//   }

//   //cout << "found intersection " << (*bestIntersection).second.location << " " << mindist2 << endl;
//   (*bestIntersection).second.location.writePointer(x);
//   (*bestIntersection).second.type += type;


// #ifdef DEBUG
//   if(mindist2 > 1e-3) {
//     cout << "nearest intersection is far " << mindist2 << endl;
//     cout << "p1 " << p1[0] << " " << p1[1] << " " << p1[2] << endl;
//     cout << "p2 " << p2[0] << " " << p2[1] << " " << p2[2] << endl;
//     cout << (*bestIntersection).second.location << endl;
//     //(*bestIntersection).second.type -= type;        
//     //throw ErrProgrammingError("CSkeletonElement: did not find required intersection.", __FILE__, __LINE__);  
//   }
//#endif 
}

bool CSkeletonElement::findTwoIntersections(CellEdge &edgeData, vtkCell *face, int numEdges, int faceId, double p1[3], double p2[3], double x[3], double y[3], double linetol2, double tol) const {

  // We look for 2 distinct intersections and ignore cases of 0, 1
  // (corner graze), or coincident intersections.
//   int edgeId, k;
//   double a[3], b[3], t, t1=-1, t2=-1, dist2;
//   bool found1 = false, found2 = false, onEdge;
//   CellEdgeMap::iterator int1, int2;

//   for(k=0; k<numEdges && !found2; k++) {
//     edgeId = faceEdgeToElementEdge(faceId, k);
//     for(CellEdgeMap::iterator here = edgeData[edgeId].begin();
// 	here!=edgeData[edgeId].end() && !found2; ++here) {
//       (*here).second.location.writePointer(a);
//       dist2 = vtkLine::DistanceToLine(a,p1,p2,t,b);
//       if(dist2 < linetol2) {
// 	if(!found1) {
// 	  int1 = here;
// 	  t1 = t;
// 	  (*here).second.location.writePointer(x);
// 	  found1 = true;
// 	}
// 	else if(!found2) {
// 	  int2 = here;
// 	  t2 = t;
// 	  (*here).second.location.writePointer(y);
// 	  // if x and y are the same, keep looking
// 	  if(vtkMath::Distance2BetweenPoints(x,y) > linetol2)
// 	    found2 = true;
// 	}
// 	break;
//       }
//     }
//   }
//   if(found2) {
//     //face->EvaluatePosition(x, a, nearestId, pcoords, dist2, weights1);
//     //face->EvaluatePosition(y, a, nearestId, pcoords, dist2, weights2);
//     //onEdge = edgeOnFaceEdge(numEdges, weights1, weights2, tol);
//     onEdge = edgeOnFaceEdge(face->GetPoints(), x, y, linetol2);
//     if(!onEdge) {
//       // cout << "found two intersections" << endl;
//       if(t1<t2) {
// 	// cout << (*int1).second.location << " " << (*int2).second.location << endl;
// 	(*int1).second.location.writePointer(x);
// 	(*int1).second.type += ENTRANCE_INTERSECTION;
// 	(*int2).second.location.writePointer(y);
// 	(*int2).second.type += EXIT_INTERSECTION;
//       }
//       else if(t1>t2) {
// 	// cout << (*int2).second.location << " " << (*int1).second.location << endl;
// 	(*int2).second.location.writePointer(x);
// 	(*int2).second.type += ENTRANCE_INTERSECTION;
// 	(*int1).second.location.writePointer(y);
// 	(*int1).second.type += EXIT_INTERSECTION;
//       }
//     }
//     else
//       found2 = false;
//   }
  
  return false; // found2;

}
  


void CSkeletonElement::setIntersectionType(CellEdge &edgeData, int faceId, int localEdgeId, double t, double tol, int type) const{

//   bool found = false;
//   double tolLimit = 100*tol;
//   // this is not meant as a permanent solution ....
//   while(!found and tol<tolLimit) {
//     int edgeId = faceEdgeToElementEdge(faceId, localEdgeId);
//     int edgeDir = faceEdgeToElementEdgeDir(faceId, localEdgeId);
//     int numFaceEdges = getFace(faceId)->GetNumberOfEdges();
//     double u;
//     if (edgeDir == -1) u = 1-t;
//     else u = t;
//     for(CellEdgeMap::iterator here = edgeData[edgeId].begin();
// 	here != edgeData[edgeId].end(); ++here) {
//       if(fabs((*here).first-u) < tol) {
// 	(*here).second.type += type;
// 	found = true;
//       }
//     }
//     // handle cases where u is within tolerance of 0 or
//     // 1 and the intersection info is potentially stored
//     // on the adjacent edge.
//     if(!found) {
//       double v = -1;
//       if(edgeDir==-1) u = 1-u;
//       if(fabs(u) < tol) { // at the beginning of the line, get previous line
// 	edgeId = faceEdgeToElementEdge(faceId,(localEdgeId-1+numFaceEdges)%numFaceEdges);
// 	  edgeDir = faceEdgeToElementEdgeDir(faceId,(localEdgeId-1+numFaceEdges)%numFaceEdges);
// 	if(edgeDir == -1) v = 0;
// 	else v = 1;
//       }
//       else if(fabs(1-u) < tol) { // at the end of the line, get next line
// 	edgeId = faceEdgeToElementEdge(faceId,(localEdgeId+1)%numFaceEdges);
// 	edgeDir = faceEdgeToElementEdgeDir(faceId,(localEdgeId+1)%numFaceEdges);
// 	if(edgeDir == -1) v = 1;
// 	else v = 0;
//       }
//       for(CellEdgeMap::iterator here = edgeData[edgeId].begin(); 
// 	  here!= edgeData[edgeId].end(); ++here) {
// 	if(fabs((*here).first-v) < tol) {
// 	  (*here).second.type += type;
// 	  found = true;
// 	}
//       }
//     }
//   tol *= 2;
//   }
  
//   if(!found) {
//     cout << "unable to assign intersection type " <<  faceEdgeToElementEdge(faceId,localEdgeId) << " " << t << endl;
//   //  throw ErrProgrammingError("CSkeletonElement: unable to assign intersection type", __FILE__, __LINE__);
//   }

}

vtkGenericCell* CSkeletonElement::generic_cell = vtkGenericCell::New();


// used by stl sort method for sorting vectors
bool operator<(const HullPoint hp1, const HullPoint hp2) {
  if(hp1.tan<hp2.tan) return true;
  // if the angles are the same, the closer point is first
  if(hp1.tan==hp2.tan && hp1.dist2<hp2.dist2) return true;
  return false;
}

// This maps the face edges to their tet-wide edge ids
int CSkeletonElement::faceEdges[4][3] = { {0,4,3}, {1,5,4}, {2,3,5}, {2,1,0} };
// Maps the direction each edge relative to it's face to the direction
// of the edge relative to the whole tet.  That is tet edge 2 is
// defined as node 2 to node 0 however for face 3, in order for the
// edges to be defined in a way such that the normal points out, this
// edge is directed from node 0 to node 2.
int CSkeletonElement::faceEdgeDirs[4][3] = { {1,1,-1}, {1,1,-1}, {1,1,-1}, {-1,-1,-1} };

// maps an edge index to the two faces it separates
int CSkeletonElement::edgeFaces[6][2] = { {0,3}, {1,3}, {2,3}, {0,2}, {0,1}, {1,2} };
//static int edgeFaceDirs[6][2] = { {1,-1}, {1,-1}, {1,-1}, {-1,1}, {1,-1}, {1,-1} };


// maps node index to the three faces it is on.
int CSkeletonElement::nodeFaces[4][3] = { {0,2,3}, {0,1,3}, {1,2,3}, {0,1,2} };

// maps node index to which edges it is on
int CSkeletonElement::nodeEdges[4][3] = { {0,2,3}, {0,1,4}, {1,2,5}, {3,4,5} };
// gives the parametric coordinate of the node for the edge given above
double CSkeletonElement::nodeEdgeParam[4][3] = { {0.0, 1.0, 0.0}, {1.0, 0.0, 0.0}, 
						 {1.0, 0.0, 0.0}, {1.0, 1.0, 1.0} };



void removeDuplicatesFromHullPoints(HullPoints &hullPoints, double linetol2) {
  //cout << "looking for duplicates" << endl;
  int l;
  for(HullPoints::iterator here=hullPoints.begin(); here!=hullPoints.end() && hullPoints.size()>=2; ++here) {
    for(HullPoints::iterator here2=here+1; here2!=hullPoints.end(); ++here2) {
      if(vtkMath::Distance2BetweenPoints((*here).x,(*here2).x) < linetol2 &&
	 ( ((*here).faces[0] && (*here2).faces[0]) || ((*here).faces[1] && (*here2).faces[1]) ||
	   ((*here).faces[2] && (*here2).faces[2]) || ((*here).faces[3] && (*here2).faces[3]) ) ) {
	//cout << "removing " << (*here2).x[0] << "," << (*here2).x[1] << "," << (*here2).x[2];
	//cout << " " << vtkMath::Distance2BetweenPoints((*here).x,(*here2).x) << " " << linetol2 << endl;
	for(l=0;l<4;++l) {
	  (*here).faces[l] |= (*here2).faces[l];
	  (*here).vgb_edges[l] |= (*here2).vgb_edges[l]; 
	}
	for(l=0;l<6;++l) {
	  (*here).edges[l] |= (*here2).edges[l];
	  if((*here2).edges[l]) (*here).t[l] = (*here2).t[l];
	}

	--here2;
	hullPoints.erase(here2+1);
      }
    }
  }
}
  

// k is the edge id.
void CSkeletonElement::addEdgeIntersection(double x[3], double t, int k, CellEdge &elementEdgeData, HullPoints &hullPoints, int *ptIds, double tol, double linetol2, double center[3], int planeidx) const
{
  bool found = false;
  int l, m;
  
#ifdef DEBUG
  cout << "intersection on element edge " << k << " " << x[0] << " " << x[1] << " " << x[2] << " " << t << " " << planeidx << endl;
#endif

  HullPoint hp = {0,0,{x[0],x[1],x[2]},
		  {false,false,false,false},
		  {false,false,false,false,false,false},
		  {false,false,false,false},
		  {0,0,0,0,0,0}};

  // if the intersection is at one of the ends of the edge, it is actually on three faces and three edges.
  if(t<tol || t>(1-tol)) {
    // get the node id that corresponds to this point
    l = (t<tol?ptIds[0]:ptIds[1]);
    hp.faces[nodeFaces[l][0]] = hp.faces[nodeFaces[l][1]] = hp.faces[nodeFaces[l][2]] = true;
    // in this case, the intersection is also on multiple edges
    CellEdgeMap::iterator here;
    for(m=0; m<3; ++m) {
      here = elementEdgeData[nodeEdges[l][m]].find(nodeEdgeParam[l][m]);
      if(here == elementEdgeData[nodeEdges[l][m]].end()) {
	elementEdgeData[nodeEdges[l][m]].insert(CellEdgeDatum(nodeEdgeParam[l][m],VoxelBdyIntersection(x[0],x[1],x[2],
							      edgeFaces[nodeEdges[l][m]][0],edgeFaces[nodeEdges[l][m]][1], planeidx)));
      }
      hp.edges[nodeEdges[l][m]] = true;
      hp.t[nodeEdges[l][m]] = nodeEdgeParam[l][m];
    }
  }

  // otherwise, just two faces and one edge, on which we check for an intersection within tolerance
  else {
    for(CellEdgeMap::iterator here = elementEdgeData[k].begin();
	here!=elementEdgeData[k].end(); ++here) {
      double y[3];
      (*here).second.location.writePointer(y);
      if( vtkMath::Distance2BetweenPoints(x,y) < linetol2) {
	found = true;
	t = (*here).first;
	(*here).second.addPlane(planeidx);
	break;
      }
    }
    if(!found) {
      // we find the intersection type later, if the intersection proves to be relevant
      elementEdgeData[k].insert(CellEdgeDatum(t,VoxelBdyIntersection(x[0],x[1],x[2],edgeFaces[k][0],edgeFaces[k][1],planeidx)));
    }
    hp.edges[k] = true;
    hp.t[k] = t;
    hp.faces[edgeFaces[k][0]] = hp.faces[edgeFaces[k][1]] = true;
  }
  // now check for duplicates in the hullPoints
  found = false;
  for(HullPoints::iterator here=hullPoints.begin(); here!=hullPoints.end() && hullPoints.size()>=2; ++here) {
    if(vtkMath::Distance2BetweenPoints((*here).x,hp.x) < linetol2 &&
       ( ((*here).faces[0] && hp.faces[0]) || ((*here).faces[1] && hp.faces[1]) ||
	 ((*here).faces[2] && hp.faces[2]) || ((*here).faces[3] && hp.faces[3]) ) ) {
      for(l=0;l<4;++l) {
	hp.faces[l] |= (*here).faces[l];
	hp.vgb_edges[l] |= (*here).vgb_edges[l]; 
      }
      for(l=0;l<6;++l) {
	hp.edges[l] |= (*here).edges[l];
	if((*here).edges[l]) hp.t[l] = (*here).t[l];
      }
      --here;
      hullPoints.erase(here+1);
    }
  }
  hullPoints.push_back(hp);    
}


const DoubleVec * CSkeletonElement::categoryVolumes(const CMicrostructure &ms)
  const
{
  // TODO 3D: cleanup!!!

  //#ifdef DEBUG

//   cerr << "\n\n\nPOINTS: " << endl;
//   double pt[3];
//   for(int i=0; i<4; i++){
//     this->getPoints()->GetPoint(i,pt);
//     cout << pt[0] << " " << pt[1] << " " << pt[2] << endl;
//   }

  // For now we are only using tetrahedra.
  int numElementFaces = 4; 
  int numElementEdges = 6; 
  int numFaceEdges = 3;
  unsigned int i, cat=0;
  int j, k, l, m;
  double *facearea = new double[numElementFaces];
  int numFaces, numBdyPoints, numPts, *ptIds;
  vtkIntArray *insideinfo = vtkIntArray::New();
  double x[3], y[3], weights[8], dist2, pcoords[3];
  bool ptinside, inside, outside, onFace, intersect;
  vtkPoints *polypoints;
  double normal[3], center[3], dot, area, p[4][3];
  double t, p1[3], p2[3], projfactor, tol;
  int subId, hullnum, proji[2];
  double basearea;
  double facebounds[6], boundedInputBounds[6];
  int edgeId, edgeDir, icount;
  vtkQuad *quad; 
  vtkTriangle *tri = vtkTriangle::New();

  double epts[4][3];
  for(i=0;i<4;++i) 
    this->getPoints()->GetPoint(i,epts[i]);
    //(*nodes)[i]->position(epts[i]);
    

#ifdef DEBUG
  cout.precision(10);
  cout << "\n\n\nPOINTS: " << endl;
  for(int i=0; i<4; i++){
    cout << epts[i][0] << " " << epts[i][1] << " " << epts[i][2] << endl;
  }
  cout << "begin categoryVolumes " << vtkTetra::ComputeVolume(epts[0],epts[1],epts[2],epts[3]) << " " << energyShape(epts) << endl;
#endif

  // The tolerance should only be compared to parametric coordinates,
  // which are in units of an element dimension (as calculated by vtk
  // routines).  If we need to compare it to anything else, make the
  // number dimensionful by multiplying by the maxLength which is
  // calculated below. 
  tol = 1e-10;

  double length, maxLength=0;
  double bounds[6] = {epts[0][0],epts[0][0],epts[0][1],epts[0][1],epts[0][2],epts[0][2]};
  for(j=1;j<4;++j) {
    for(k=0;k<3;++k) {
      if(epts[j][k] < bounds[2*k])
	bounds[2*k]=epts[j][k];
      if(epts[j][k] > bounds[2*k+1])
	bounds[2*k+1]=epts[j][k];
    }
  }

  for(j=0; j<3; j++) {
    length = bounds[2*j+1]-bounds[2*j];
    if(length>maxLength)
      maxLength=length;
  }

  double fnormal[4][3];
  int fpi[4][2] = {{-1,-1},{-1,-1},{-1,-1},{-1,-1}};
  double fprojf[4] = {0,0,0,0};
  vtkPlane *eplane[4];
  for(j=0; j<4; ++j) {
    ptIds = vtkTetra::GetFaceArray(j);
    vtkTriangle::ComputeNormal(epts[ptIds[0]], epts[ptIds[1]], epts[ptIds[2]], fnormal[j]);
    eplane[j] = vtkPlane::New();
    eplane[j]->SetNormal(fnormal[j]);
    eplane[j]->SetOrigin(epts[ptIds[0]]);
    // We find the axis direction which has the maximum contribution
    // to the normal.  We can save floating point operations later on
    // by projecting onto this direction, calculating the 2D cross
    // product, and then dividing by the cosine of the normal to find
    // the area contribution for each segment.  The sign of the
    // projfactor will correct for whether the bounding segments are
    // oriented CCW when projected onto the given axis oriented plane.
    if(fabs(fnormal[j][0]) >= fabs(fnormal[j][1]) && fabs(fnormal[j][0]) >= fabs(fnormal[j][2])) {
      fprojf[j] = fnormal[j][0];
      fpi[j][0]=1;
      fpi[j][1]=2;
    }
    else if(fabs(fnormal[j][1]) >= fabs(fnormal[j][0]) && fabs(fnormal[j][1]) >= fabs(fnormal[j][2])) {
      fprojf[j] = fnormal[j][1];
      fpi[j][0]=2;
      fpi[j][1]=0;
    }
    else if(fabs(fnormal[j][2]) >= fabs(fnormal[j][0]) && fabs(fnormal[j][2]) >= fabs(fnormal[j][1])) {
      fprojf[j] = fnormal[j][2];
      fpi[j][0]=0;
      fpi[j][1]=1;
    }
  }

  // We need a dimensionful tolerance for comparing distances to
  // planes, lines, and between points.
  double linetol = tol*maxLength; 
  double linetol2 = linetol*linetol;

#ifdef DEBUG
  cout << "linetol " << linetol << " linetol2 " << linetol2 << endl;
#endif

  unsigned int n = ms.nCategories();	// recomputes categories & boundaries if needed
  DoubleVec *result = new DoubleVec(n);

  // Get all the category boundaries.
  const std::vector<vtkPolyData*>  &bdys = ms.getCategoryBdys();

  // first loop over the categories and get portions of the boundaries 
  // that overlap the bounds of the element.  We also set up one shared
  // vtkPoints object and avoid using slow vtk objects such as CleanPolyData
  // and the Locator.
  std::vector<vtkPolyData*> boundedInput;
  // each entry in bdys has the same number of points
  unsigned int numPoints = bdys[0]->GetPoints()->GetNumberOfPoints();
  vtkPoints *bdypoints = vtkPoints::New();
  bdypoints->Allocate(numPoints, numPoints); 
  PointMap pointMap;
  PointMap::iterator pmi;
  int oldId, currId=0;
  vtkDoubleArray *scalarData, *normalData;
  int quadPtIds[4];
  for(i=0; i<bdys.size(); i++) {
    (*result)[i] = 0;
    numFaces = bdys[i]->GetNumberOfPolys();
    boundedInput.push_back(vtkPolyData::New());
    // This line leads to an Invalid Read in valgrind
    boundedInput[i]->Allocate(numFaces,numFaces);
    scalarData = vtkDoubleArray::New();
    scalarData->SetNumberOfComponents(2);
    boundedInput[i]->GetCellData()->SetScalars(scalarData);
    normalData = vtkDoubleArray::New();
    normalData->SetNumberOfComponents(3);
    boundedInput[i]->GetCellData()->SetNormals(normalData);		
    for(j=0; j<numFaces; j++) {
      // this is the most expensive line in this function
      bdys[i]->GetCellBounds(j,facebounds);
      if( boundsOverlap(facebounds,bounds) ) {
	// add the points to the point object, id map, and quad
	bdys[i]->GetCellPoints(j,numPts,ptIds);
	for(k=0; k<4; k++) {
	  oldId = ptIds[k];
	  pmi = pointMap.find(oldId);
	  if(pmi == pointMap.end()) {
	    pointMap[oldId] = currId;
	    bdys[i]->GetPoints()->GetPoint(oldId,x);
	    bdypoints->InsertNextPoint(x);
	    ++currId;
	  }
	  else
	    bdys[i]->GetPoints()->GetPoint(oldId, x);
	  quadPtIds[k] = pointMap[oldId];
	}
	// now add the cells
	boundedInput[i]->InsertNextCell(VTK_QUAD,4,quadPtIds);
	scalarData->InsertNextTuple(bdys[i]->GetCellData()->GetScalars()->GetTuple(j));
	normalData->InsertNextTuple(bdys[i]->GetCellData()->GetNormals()->GetTuple(j));
      }
    }
    boundedInput[i]->SetPoints(bdypoints);
    scalarData->Delete();
    normalData->Delete();
  }

  // Start by looping over the bdypoints in the input and storing
  // whether they are in or out.  The point is inside if
  // vtkPlane::EvaluateFunction gives a value less than -linetol for
  // all four planes.  This means the point is inside the tetrahedron
  // and not within tolerance of any face.  We classify points that
  // are within tolerance of the plane as outside so that
  // intersections are definitely found and so that we can keep track
  // of which face each point is on.  This method also works better
  // for badly shaped elements which can make built in vtk methods,
  // such as vtkTetra::EvaluatePosition, barf.
  numBdyPoints = bdypoints->GetNumberOfPoints();
  insideinfo->SetNumberOfValues(numBdyPoints);
  for(j=0; j<numBdyPoints; j++) {
    bdypoints->GetPoint(j,x);
    ptinside = true;
    for(k=0;k<4;++k) {
      ptinside &= (eplane[k]->EvaluateFunction(x) < -linetol);
    }
    insideinfo->SetValue(j,ptinside);
  }
  // this just deletes one reference to the shared points object
  bdypoints->Delete();



  // Loop over voxel categories.
  for(i=0; i<bdys.size(); i++) {
// #ifdef DEBUG
//     cout << "\n\nexamining category " << i << endl;
// #endif
    
    // facearea[j] stores the of element face j which corresponds to category i
    for(j=0; j<numElementFaces; ++j) facearea[j] = 0.0;

    CellEdge elementEdgeData(numElementEdges);
    numFaces = boundedInput[i]->GetNumberOfPolys();
    normalData = vtkDoubleArray::SafeDownCast(boundedInput[i]->GetCellData()->GetNormals());
    scalarData = vtkDoubleArray::SafeDownCast(boundedInput[i]->GetCellData()->GetScalars());

// #ifdef DEBUG
//     cout << "phase 1. numFaces: " << numFaces << " numBdyPoints: " << numBdyPoints << endl;
// #endif

    // Loop over the voxel boundary faces.
    for(j=0; j<numFaces; j++) {
      quad = vtkQuad::SafeDownCast(boundedInput[i]->GetCell(j));
      polypoints = quad->GetPoints();
      normalData->GetTuple(j,normal);


#ifdef DEBUG
      cout << "\nface " << j << endl;
      for(k=0;k<4;k++) {
	polypoints->GetPoint(k,x);
	cout << k << " " << quad->GetPointIds()->GetId(k) << " " << x[0] << " " << x[1] << " " << x[2] << " " << (int)insideinfo->GetValue(quad->GetPointIds()->GetId(k)) << endl;
      }
#endif

      polypoints->GetPoint(0,x);
      // We check if the face is actually coincident with one of the
      // element faces.  If that's the case, the volume contribution
      // will be added later, so we want to ignore it here.  (This
      // isn't true anymore, but just commenting this out for now)
      // onFace = faceOnElementFace(epts, fnormal, normal,x,linetol);
      // cout << "onFace " << onFace << endl;

      //if(true) {//!onFace) {

	// Since the faces are axis aligned, only one component of the
	// normal is non zero.
	int planeidx;
	if(fabs(normal[0])) {
	  planeidx = 0;
	  proji[0] = 1;
	  proji[1] = 2;
	}
	if(fabs(normal[1])) {
	  planeidx = 1;
	  proji[0] = 2;
	  proji[1] = 0;
	}
	if(fabs(normal[2])) {
	  planeidx = 2;
	  proji[0] = 0;
	  proji[1] = 1;
	}
	HullPoints hullPoints;

	center[0] = center[1] = center[2] = 0.0;
	dot = 0.0;
	inside = 0;
	outside = 0;
	for(k=0; k<4; k++ ){
	  polypoints->GetPoint(k,p[k]);

	  ptinside = (int)insideinfo->GetValue(quad->GetPointIds()->GetId(k));
	  inside |= ptinside;
	  outside |= !ptinside;
	  if(ptinside) {
	    HullPoint hp = {0,0,{p[k][0],p[k][1],p[k][2]},
			    {false,false,false,false},
			    {false,false,false,false,false,false}, 
			    {false,false,false,false},
			    {0,0,0,0,0,0}};
	    hullPoints.push_back(hp);
	  }
	}
	
	// If the voxel boundary face is totally inside, we just add a
	// volume contribution for the whole face.
	if(inside && !outside) { 
	  area = scalarData->GetComponent(j,0);
	  dot = scalarData->GetComponent(j,1);
	  (*result)[i] += 1.0/3.0 * dot * area;
	}
	// If some or none of the voxel group boundary face points are
	// outside, we must find the intersections of each element
	// edge with the voxel boundary faces and the voxel boundary
	// edge with the element faces.  This set of points, plus any
	// voxel group boundary points which are inside the element
	// will form a convex hull.  We remove duplicate points and
	// then use the Graham-Scan algorithm for finding a convex
	// hull to order them.  As we find the intersections, we store
	// which face or edge they are on.  We can use this
	// information to find both the volume contribution of the
	// voxel group boundary face *and* the contributions to the
	// category area for each element face.
	else { 

	  // Loop over the faces of the element and find the
	  // intersections with the edges of the voxel group boundary
	  // quad.  We test that the edge intersects the plane and
	  // that the intersection is within the triangle.  We don't
	  // need to test for the case the the edge is in or nearly in
	  // the same plane as the face because that was done above.
	  for(l=0; l<numElementFaces; ++l) {
	    ptIds = vtkTetra::GetFaceArray(l);
	    for(m=0; m<3; ++m) 
	      tri->GetPoints()->SetPoint(m,epts[ptIds[m]]);
	    for(k=0; k<4; k++) {
	      polypoints->GetPoint(k,p1);
	      polypoints->GetPoint((k+1)%4,p2);
	      intersect = vtkPlane::IntersectWithLine(p1,p2,fnormal[l],epts[ptIds[0]],t,x);
	      bool in = tri->EvaluatePosition(x,y,subId,pcoords,dist2,weights);
	      // Calculate a tolerance that represents the same length
	      // as tol.  Take advantage of the fact that all quads
	      // are axis aligned to simplify the length measurement.
	      double newtol = linetol / (fabs(p2[0]-p1[0])+fabs(p2[1]-p1[1])+fabs(p2[2]-p1[2]));
	      if( ((intersect || (t>-newtol && t<1+newtol)) && in )) { 
#ifdef DEBUG
		  cout << "intersection on element face " << l << " " << x[0] << " " << x[1] << " " << x[2] << endl;
#endif
		HullPoint hp = {0,0,{x[0],x[1],x[2]},
				{false,false,false,false},
				{false,false,false,false,false,false},
				{false,false,false,false},
				{0,0,0,0,0,0}};
		hp.faces[l] = true;
		hp.vgb_edges[k] = true;
		hullPoints.push_back(hp);
	      }
	    }
	  }


	  // Loop over element edges to find the intersections
	  // with the voxel boundary face.
	  quad->GetBounds(facebounds);
	  bool linegraze = false;
	  double z[3];
	  for(k=0; k<numElementEdges; k++) {
	    ptIds = vtkTetra::GetEdgeArray(k);
	    quad->GetPoints()->GetPoint(0,y);
	    // Find the intersection of the element edge with the
	    // plane of the voxel group boundary face.
	    intersect = vtkPlane::IntersectWithLine(epts[ptIds[0]],epts[ptIds[1]],normal,y,t,x);
	    // If there is an intersection, and it is within the
	    // simple bounds of the voxel group bounary quad, then add
	    // the intersection info to our storage objects.
#ifdef DEBUG
	    cout << intersect << " " << x[0] << "," << x[1] << "," << x[2] << " " << t << endl;
#endif
	    if(intersect && 
	       x[proji[0]]>=facebounds[2*proji[0]]-linetol && x[proji[0]]<=facebounds[2*proji[0]+1]+linetol 
	       && x[proji[1]]>=facebounds[2*proji[1]]-linetol && x[proji[1]]<=facebounds[2*proji[1]+1]+linetol) {
#ifdef DEBUG
	      cout << "regular intersection" << endl;
#endif
	      addEdgeIntersection(x,t,k,elementEdgeData,hullPoints,ptIds,tol,linetol2,center,planeidx);
	    }
	    // If we didn't find an intersection using the simple test
	    // above, we need to test for the possiblity that the
	    // element edge and the voxel group boundary face are all
	    // in the same plane or nearly in the same plane.  In this
	    // case, it is possible to have two intersections.
	    else {
	      double u, v;
	      for(l=0; l<4; ++l) {
		polypoints->GetPoint(l,p1);
		polypoints->GetPoint((l+1)%4,p2);
		linegraze = vtkLine::Intersection(epts[ptIds[0]],epts[ptIds[1]],p1,p2,u,v);
		if(linegraze) {
		  for(m=0; m<3; ++m) {
		    y[m] = p1[m]+v*(p2[m]-p1[m]);
		    z[m] = epts[ptIds[0]][m]+u*(epts[ptIds[1]][m]-epts[ptIds[0]][m]);
		  }
#ifdef DEBUG
		  cout << "linegraze x,y,z " << x[0] << "," << x[1] << "," << x[2] << " ";
		  cout << y[0] << "," << y[1] << "," << y[2] << " " << z[0] << "," << z[1] << "," << z[2] << endl;
#endif
		  if(vtkMath::Distance2BetweenPoints(y,z) < linetol2) {
 		    if(!intersect) {
#ifdef DEBUG
 		      cout << "simple graze" << endl;
#endif
		      addEdgeIntersection(z,u,k,elementEdgeData,hullPoints,ptIds,tol,linetol2,center,planeidx);
		    }
		    // In some cases, we will find a plane
		    // intersection that is barely outside the bounds
		    // of the voxel boundary face and then find a
		    // grazing intersection.  We need to check that
		    // these intersections are topologically sane --
		    // edges can only intersect planes in one place if
		    // they are not IN the plane.  We enforce this
		    // twice.  If the intersection found in the
		    // earlier plane test (x) is already in the list
		    // of edge intersections and represents an
		    // intersection of the same plane, we do not add z here.
   		    else {
		      int facexcount = 0;
		      for(HullPoints::iterator here=hullPoints.begin(); here!=hullPoints.end(); ++here) {
#ifdef DEBUG
			cout << (*here).x[0] << "," << (*here).x[1] << "," << (*here).x[2] <<  " ";
			cout << (*here).vgb_edges[l] << " " << vtkMath::Distance2BetweenPoints((*here).x,y) << endl;
#endif
			if((*here).vgb_edges[l] && vtkMath::Distance2BetweenPoints((*here).x,y) >= linetol2) ++facexcount;
		      }
  		      bool found = false;
		      for(CellEdgeMap::iterator here = elementEdgeData[k].begin();
			  here!=elementEdgeData[k].end(); ++here) {
			(*here).second.location.writePointer(y);
			if( (vtkMath::Distance2BetweenPoints(x,y) < linetol2 || vtkMath::Distance2BetweenPoints(z,y) < linetol2)  
			    && (*here).second.getPlane(planeidx)) {
#ifdef DEBUG
			  cout << "found " << y[0] << "," << y[1] << "," << y[2] << " " << planeidx << " ";
			  cout << (*here).second.getPlane(0) << (*here).second.getPlane(1) << (*here).second.getPlane(2) << endl;
#endif
			  found = true;
			  break;
			}
		      }
#ifdef DEBUG
		      cout << "near miss graze " << !found << (facexcount<2) << endl;
#endif
 		      if(!found && facexcount<2) {
			addEdgeIntersection(z,u,k,elementEdgeData,hullPoints,ptIds,tol,linetol2,center,planeidx);
		      }
 		    }

		  }
		}
	      }
	    }
	  }


	  // Since we are intersecting two convex shapes, the result
	  // must be convex.  We use the Graham-Scan algorithm to find
	  // the convex hull.

#ifdef DEBUG
	  cout << "\nfound points: " << endl;
	  for(HullPoints::iterator here=hullPoints.begin(); here!=hullPoints.end(); ++here) {
	    cout << (*here).x[0] << "," << (*here).x[1] << "," << (*here).x[2] << " ";
	    cout << (*here).faces[0] << (*here).faces[1] << (*here).faces[2] << (*here).faces[3] << " ";
	    cout << (*here).edges[0] << (*here).edges[1] << (*here).edges[2] << (*here).edges[3] << (*here).edges[4] << (*here).edges[5] << endl;
	  }
	  cout << endl;
#endif DEBUG

	  // First we remove points that are within tolerance and
	  // share at least one face.  The other topological
	  // information is combined.
	  std::reverse(hullPoints.begin(), hullPoints.end());
	  removeDuplicatesFromHullPoints(hullPoints,linetol2);

// #ifdef DEBUG
// 	  cout << "\ncontents of hull list: " << endl;
// 	  for(HullPoints::iterator here=hullPoints.begin(); here!=hullPoints.end(); ++here) {
// 	    cout << (*here).x[0] << "," << (*here).x[1] << "," << (*here).x[2] << " ";
// 	    cout << (*here).faces[0] << (*here).faces[1] << (*here).faces[2] << (*here).faces[3] << " ";
// 	    cout << (*here).edges[0] << (*here).edges[1] << (*here).edges[2] << (*here).edges[3] << (*here).edges[4] << (*here).edges[5] << endl;
// 	  }
// 	  cout << endl;
// #endif

	  CellEdgeMap::iterator cmi;
	  if( hullPoints.size() >2 ) { 

	    // We find the pivot, the point with the lowest horizontal
	    // coordinate (and lowest vertical coordinate in the case of a
	    // tie). 
	    HullPoints::iterator pivot = hullPoints.begin();
	    for(HullPoints::iterator here=hullPoints.begin(); here!=hullPoints.end(); ++here) {
	      if((*here).x[proji[0]] < (*pivot).x[proji[0]]) 
		pivot = here;
	      if((*here).x[proji[0]] == (*pivot).x[proji[0]] 
		 && (*here).x[proji[1]] < (*pivot).x[proji[1]])
		pivot = here;
	    }

	    // We set the value of the tangent of the angle from the
	    // horizontal axis and distance to the pivot.
	    for(HullPoints::iterator here=hullPoints.begin(); here!=hullPoints.end(); ++here) {
	      if(here==pivot) {
		// Set this so that the pivot is first when HullPoints is sorted.
		(*here).tan = -1*HUGE_VAL;
		(*here).dist2 = 0;
	      }
	      else {
		if(((*here).x[proji[0]]-(*pivot).x[proji[0]]) != 0)
		  (*here).tan = ((*here).x[proji[1]]-(*pivot).x[proji[1]])/((*here).x[proji[0]]-(*pivot).x[proji[0]]);
		else
		  (*here).tan = HUGE_VAL;
		(*here).dist2 = vtkMath::Distance2BetweenPoints((*here).x,(*pivot).x);
	      }
	    }
	    // The built in std::sort method uses operator< for
	    // HullPoints defined above.
	    std::sort(hullPoints.begin(), hullPoints.end());

	    hullnum = hullPoints.size();
	    HullPoints hullPoints2;
	    hullPoints2.push_back(hullPoints[0]);
	    hullPoints2.push_back(hullPoints[1]);
	    for(l=2; l<hullnum; ++l) {
	      while(hullPoints2.size() >=2 && 
		    (hullPoints2.back().x[proji[0]]-(*(hullPoints2.end()-2)).x[proji[0]]) *
		    (hullPoints[l].x[proji[1]]-hullPoints2.back().x[proji[1]]) -
		    (hullPoints[l].x[proji[0]]-hullPoints2.back().x[proji[0]]) *
		    (hullPoints2.back().x[proji[1]]-(*(hullPoints2.end()-2)).x[proji[1]]) < 0) {
		hullPoints2.pop_back();
	      }
	      hullPoints2.push_back(hullPoints[l]);
	    }

	    hullnum = hullPoints2.size();
	    for(l=0; l<hullnum; ++l) 
	      for(k=0; k<3; ++k) center[k] += hullPoints2[l].x[k] / hullnum;
	         
	    dot = normal[0]*center[0] + normal[1]*center[1] + normal[2]*center[2]; 
	    double signofdot = dot/fabs(dot);
	    basearea = 0.0;

// #ifdef DEBUG
// 	    cout << "\nhull: " << endl;
// 	    for(HullPoints::iterator here=hullPoints2.begin(); here!=hullPoints2.end(); ++here) {
// 	      cout << (*here).x[0] << "," << (*here).x[1] << "," << (*here).x[2] << " ";
// 	      cout << (*here).faces[0] << (*here).faces[1] << (*here).faces[2] << (*here).faces[3] << " ";
// 	      cout << (*here).edges[0] << (*here).edges[1] << (*here).edges[2] << (*here).edges[3] << (*here).edges[4] << (*here).edges[5] << endl;
// 	    }
// 	    cout << endl;
// #endif

	    for(k=0; k<hullnum; ++k) {
	      double *q1 = hullPoints2[k].x;
	      double *q2 = hullPoints2[(k+1)%hullnum].x;

	      basearea += (1.0/2.0)*(q1[proji[0]]*q2[proji[1]]-q1[proji[1]]*q2[proji[0]]);

#ifdef DEBUG
 	      cout << "points (" << q1[0] << ", " << q1[1] << ", " << q1[2] << ") (";
 	      cout << q2[0] << ", " << q2[1] << ", " << q2[2] << ")" << endl;
#endif

	      for(l=0;l<4;++l) {
		if(hullPoints2[k].faces[l] && hullPoints2[(k+1)%hullnum].faces[l] && 
		   // we don't want to consider points that are both on an edge
		   !((hullPoints2[k].edges[faceEdges[l][0]] && hullPoints2[(k+1)%hullnum].edges[faceEdges[l][0]]) ||
		     (hullPoints2[k].edges[faceEdges[l][1]] && hullPoints2[(k+1)%hullnum].edges[faceEdges[l][1]]) ||
		     (hullPoints2[k].edges[faceEdges[l][2]] && hullPoints2[(k+1)%hullnum].edges[faceEdges[l][2]])) ) {
		  // The points are sorted into a CCW hull regardless
		  // of which direction the voxel boundary face
		  // points, so we have to effectivley reverse the
		  // order of the points if the normal points in a
		  // postive direction (indicated by sign of dot).
		  facearea[l]+=(-signofdot)*(1.0/2.0)*(q1[fpi[l][0]]*q2[fpi[l][1]]-q1[fpi[l][1]]*q2[fpi[l][0]]);
#ifdef DEBUG
		  cout << "hull points on face " << l << " " << q1[fpi[l][0]] << " " << q1[fpi[l][1]];
		  cout << " to " << q2[fpi[l][0]] << " " << q2[fpi[l][1]] << " contrib ";
		  cout <<  (-signofdot)*(1.0/2.0)*(q1[fpi[l][0]]*q2[fpi[l][1]]-q1[fpi[l][1]]*q2[fpi[l][0]]);
		  cout << " " << dot << endl;
#endif

		  for(m=0; m<numFaceEdges; ++m) {
		    if(hullPoints2[k].edges[faceEdges[l][m]]) {
		      cmi = elementEdgeData[faceEdges[l][m]].find(hullPoints2[k].t[faceEdges[l][m]]);
		      if(cmi != elementEdgeData[faceEdges[l][m]].end()) {
			if(signofdot>0) (*cmi).second.addType(l,EXIT_INTERSECTION);
			if(signofdot<0) (*cmi).second.addType(l,ENTRANCE_INTERSECTION);
		      }
#ifdef DEBUG
		      cout << "setting point " << q1[0] << "," << q1[1] << "," << q1[2] << " to " << (signofdot>0?"exit":"entrance") << " for face " << l << " edge " << faceEdges[l][m] << endl;
#endif
		    }
		    if(hullPoints2[(k+1)%hullnum].edges[faceEdges[l][m]]) {
		      cmi = elementEdgeData[faceEdges[l][m]].find(hullPoints2[(k+1)%hullnum].t[faceEdges[l][m]]);
		      if(cmi != elementEdgeData[faceEdges[l][faceEdges[l][m]]].end()) {
			if(signofdot>0) (*cmi).second.addType(l,ENTRANCE_INTERSECTION);
			if(signofdot<0) (*cmi).second.addType(l,EXIT_INTERSECTION);
		      }
#ifdef DEBUG
		      cout << "setting point " << q2[0] << "," << q2[1] << "," << q2[2] << " to " << (signofdot<0?"exit":"entrance") << " for face " << l  << " edge " << faceEdges[l][m] << endl;
#endif
		    }
		  }

		}
	      }
	    }
	    (*result)[i] += 1.0/3.0 * dot * basearea;

	  } // end if hullPoints.size()>2)

	  // if the hull has only 2 points, it could still have a
	  // significant contribution to the facearea if the voxel
	  // group boundary face actually intersects the element face
	  // by a tiny amount.
	  else if(hullPoints.size()==2) {
	    double *q1 = hullPoints[0].x;
	    double *q2 = hullPoints[1].x;
	      

	    for(l=0;l<4;++l) {
#ifdef DEBUG
	      for(k=0;k<4;++k) cout << " " << eplane[l]->EvaluateFunction(polypoints->GetPoint(k));
	      cout << " (" << linetol << ")" << endl;
	      if(hullPoints[0].faces[l] && hullPoints[1].faces[l]) {
		cout << "'hull' points on face " << l << " " << q1[0] << "," << q1[1] << "," << q1[2];
		cout << " to " << q2[0] << "," << q2[1] << "," << q2[2] << endl;
	      }
#endif

	      if(hullPoints[0].faces[l] && hullPoints[1].faces[l] && 
		 // we don't want to consider points that are both on an edge
		 !((hullPoints[0].edges[faceEdges[l][0]] && hullPoints[1].edges[faceEdges[l][0]]) ||
		   (hullPoints[0].edges[faceEdges[l][1]] && hullPoints[1].edges[faceEdges[l][1]]) ||
		   (hullPoints[0].edges[faceEdges[l][2]] && hullPoints[1].edges[faceEdges[l][2]]))) {

		bool over=false, under=false;
		double dist;
		for(k=0;k<4;++k) {
		  dist = eplane[l]->EvaluateFunction(polypoints->GetPoint(k));
		  over |= dist>-linetol; //0;
		  under |= dist<-linetol; //0;
		}
		// the voxel group boundary plane must intersect the
		// element face plane in order to be considered.
		//cout << over << " " << under << endl;
		if(over && under) {

		  // We need to determine the direction of the edge
		  // formed by the two points, they are in the order in
		  // which they were found which has nothing to do with
		  // the ordering that encloses the category area on the
		  // face.  First we find the center of the voxel group
		  // boundary face.  Then we find the cross product of a
		  // vector from the center to the first point and the
		  // vector from the first point to the second point.
		  // This vector will be parallel or antiparallel to the
		  // voxel group boundary face normal -- the sign of the
		  // dot product between these two vectors will
		  // determine the order of the points.
		  center[0] = center[1] = center[2] = 0.0;
		  for(m=0; m<4; ++m) {
		    polypoints->GetPoint(m,x);
		    for(k=0; k<3; ++k) center[k] += x[k] * 0.25;
		  }
		  double cross[3];
		  cross[0] = (q1[1]-center[1])*(q2[2]-q1[2]) - (q1[2]-center[2])*(q2[1]-q1[1]); 
		  cross[1] = (q1[2]-center[2])*(q2[0]-q1[0]) - (q1[0]-center[0])*(q2[2]-q1[2]); 
		  cross[2] = (q1[0]-center[0])*(q2[1]-q1[1]) - (q1[1]-center[1])*(q2[0]-q1[0]);
		  dot = cross[0]*normal[0]+cross[1]*normal[1]+cross[2]*normal[2];
		  double signofdot = dot/fabs(dot);

		  facearea[l]+=(signofdot)*(1.0/2.0)*(q1[fpi[l][0]]*q2[fpi[l][1]]-q1[fpi[l][1]]*q2[fpi[l][0]]);
#ifdef DEBUG
		  cout << "'hull' points on face " << l << " " << q1[fpi[l][0]] << " " << q1[fpi[l][1]];
		  cout << " to " << q2[fpi[l][0]] << " " << q2[fpi[l][1]] << " contrib ";
		  cout <<  (signofdot)*(1.0/2.0)*(q1[fpi[l][0]]*q2[fpi[l][1]]-q1[fpi[l][1]]*q2[fpi[l][0]]);
		  cout << " " << dot << endl;
#endif
		  for(m=0; m<numFaceEdges; ++m) {
		    if(hullPoints[0].edges[faceEdges[l][m]]) {
		      cmi = elementEdgeData[faceEdges[l][m]].find(hullPoints[0].t[faceEdges[l][m]]);
		      if(cmi != elementEdgeData[faceEdges[l][m]].end()) {
			if(signofdot<0) (*cmi).second.addType(l,EXIT_INTERSECTION);
			if(signofdot>0) (*cmi).second.addType(l,ENTRANCE_INTERSECTION);
		      }
#ifdef DEBUG
		      cout << "setting point " << q1[0] << "," << q1[1] << "," << q1[2] << " to " << (signofdot<0?"exit":"entrance") << " for face " << l << " edge " << faceEdges[l][m] << endl;
#endif
		    }
		    if(hullPoints[1].edges[faceEdges[l][m]]) {
		      cmi = elementEdgeData[faceEdges[l][m]].find(hullPoints[1].t[faceEdges[l][m]]);
		      if(cmi != elementEdgeData[faceEdges[l][m]].end()) {
			if(signofdot<0) (*cmi).second.addType(l,ENTRANCE_INTERSECTION);
			if(signofdot>0) (*cmi).second.addType(l,EXIT_INTERSECTION);
		      }
#ifdef DEBUG
		      cout << "setting point " << q2[0] << "," << q2[1] << "," << q2[2] << " to " << (signofdot>0?"exit":"entrance") << " for face " << l  << " edge " << faceEdges[l][m] << endl;
#endif
		    }
		  }
		} // end if(over && under)
	      } // end if on same face but not same edge
	    } // end loop over faces
	  }  // end if(hullPoints.size()==2)
		  

	} // end vgb face not inside
	//} // end if (!onFace)
    } // end loop over faces (j)

// #ifdef DEBUG
//     cout << "end phase 1. vol so far: " << (*result)[i] << endl;
// #endif


    // Now find the contribution of the faces of the element.  We have
    // already found the contributions to the face category areas from
    // the voxel group boundaries.  Now we need to traverse the edges
    // of each element face and add up the contributions from them.
    // We use the intersections found earlier.  If numFaces=0, then
    // either the whole element is in this category, or it's out.
    if(numFaces) {
      boundedInput[i]->GetBounds(boundedInputBounds);
      for(j=0; j<numElementFaces; ++j) {

#ifdef DEBUG
	cout << "\n\nface " << j << " area so far " << facearea[j] << endl;
#endif

	numFaceEdges = 3; 
	ptIds = vtkTetra::GetFaceArray(j);
	double fpts[numFaceEdges][3];
	for(k=0;k<numFaceEdges;++k) {
	  for(l=0;l<3;++l)
	    fpts[k][l] = epts[ptIds[k]][l];
	}

#ifdef DEBUG
	cout << "points: " << endl;
	for(k=0;k<numFaceEdges;++k)
	  cout << epts[ptIds[k]][0] << " " << epts[ptIds[k]][1] << " " << epts[ptIds[k]][2] << endl;
	cout << "normal " << fnormal[j][0] << " " << fnormal[j][1] << " " << fnormal[j][2] << endl;
	cout << "total face area " << vtkTriangle::TriangleArea(epts[ptIds[0]],epts[ptIds[1]],epts[ptIds[2]]) << endl;
	cout << "linetol " << linetol << endl;
#endif

#ifdef DEBUG
	// polypoints is only used for debugging output
	polypoints = vtkPoints::New();
	for(k=0;k<numFaceEdges;++k)
	  polypoints->InsertNextPoint(epts[ptIds[k]]);
	polypoints->GetBounds(facebounds);	
#endif
	

#ifdef DEBUG
	cout << "intersections: " << endl;
#endif

	int edgeId2, edgeDir2;
	CellEdgeMap::iterator here2;
	// loop through in order and check for duplicate entries at corners
	for(k=0; k<numFaceEdges; k++) {
	  edgeId = faceEdges[j][k];
	  edgeDir = faceEdgeDirs[j][k];
	  for(CellEdgeMap::iterator here = elementEdgeData[edgeId].begin();
	      here!=elementEdgeData[edgeId].end(); ++here) {

#ifdef DEBUG
	    cout << edgeId << " " << (*here).first << " " << (*here).second.location << " " << (*here).second.getType(j) << endl;
#endif	    

	    // first check if we are on a corner and the next edge
	    // contains an identical intersection TODO 3D: this can
	    // probably be simplified taking advantage of the fact
	    // that the endpoints are either at begin() or one before
	    // end().
	    if( (edgeDir==1 && (*here).first>=(1-tol)) || (edgeDir==-1 && (*here).first<=tol) ) {
	      edgeId2 = faceEdges[j][(k+1)%numFaceEdges];
	      edgeDir2 = faceEdgeDirs[j][(k+1)%numFaceEdges];
	      for(CellEdgeMap::iterator here2 = elementEdgeData[edgeId2].begin();
		  here2!=elementEdgeData[edgeId2].end(); ++here2) {
		if( ((edgeDir2==-1 && (*here2).first>=(1-tol)) || (edgeDir2==1 && (*here2).first<=tol))
		    && (*here).second.getType(j) % GRAZE_INTERSECTION == (*here2).second.getType(j) % GRAZE_INTERSECTION ) {
		  (*here2).second.setType(j,UNCATEGORIZED_INTERSECTION);
		}
	      }
	    }
	  }
	}

	center[0] = center[1] = center[2] = 0.0;
	// For now, the center is just set to be the center of the
	// element face.  This is an approximation.  We should
	// calculate it using the average of the points defining the
	// portion of the face that is within the category in
	// question.
	for(k=0; k<3; k++) {
	  for(l=0; l<3; l++) center[l] += fpts[k][l]/3.0;
	}
	dot = 0.0;
	
#ifdef DEBUG		
	// Find where voxel group boundaries intersect the plane, useful for debugging
	if(boundsOverlap(facebounds,boundedInputBounds)) {
	  edgeExtractor->SetInput(boundedInput[i]); 
	  edgeExtractor->SetCutFunction(eplane[j]);
	}
#endif

	// Now loop over the edges in the face and add the
	// contributions between the exits and the entrances. 
	icount = 0;
	int numexits = 0;
	int numentrances = 0;

	for(k=0; k<numFaceEdges; k++) {
	  edgeId = faceEdges[j][k];
	  edgeDir = faceEdgeDirs[j][k];
	  for(CellEdgeMap::iterator here = elementEdgeData[edgeId].begin();
	      here!=elementEdgeData[edgeId].end(); ++here) {

	    if ( (*here).second.getType(j) % GRAZE_INTERSECTION == EXIT_INTERSECTION ) {
	      numexits++;
	      icount++;
	    }
	    if ( (*here).second.getType(j) % GRAZE_INTERSECTION == ENTRANCE_INTERSECTION ) {
	      numentrances++;
	      icount++;
	    }
	  }
	}

#ifdef DEBUG
 	cout << "icount " << icount << endl;
#endif
	if(icount >= 2) {

	  // This is kind of cheating, but it is sometimes unavoidable
	  // that we find intersections barely out of tolerance that
	  // have the same topological significance.  If there are two
	  // intersections on an edge that have the same type, we
	  // combine them.
// 	  if(abs(numexits-numentrances)==1) {
// 	    int last=UNCATEGORIZED_INTERSECTION;
// 	    CellEdgeMap::iterator here, end;
// 	    bool done = false;
// 	    for(k=0; k<numFaceEdges+1 && !done; k++) {
// 	      // get the element edgeId and dir
// 	      edgeId = faceEdges[j][k%numFaceEdges];
// 	      edgeDir = faceEdgeDirs[j][k%numFaceEdges];
// 	      if(edgeDir==1) {
// 		here = elementEdgeData[edgeId].begin();
// 		end = elementEdgeData[edgeId].end();
// 	      }
// 	      if(edgeDir==-1) {
// 		here = elementEdgeData[edgeId].end();
// 		end = elementEdgeData[edgeId].begin();
// 	      }
// 	      while(here!=end) {
// 		if(edgeDir==-1) --here;
// 		//for(CellEdgeMap::iterator here = elementEdgeData[edgeId].begin();
// 		//here!=elementEdgeData[edgeId].end(); ++here) {

// 		if ( (*here).second.getType(j) % GRAZE_INTERSECTION == last ) {
// 		  --icount;
// 		  if(last==EXIT_INTERSECTION) --numexits;
// 		  if(last==ENTRANCE_INTERSECTION) --numentrances;
// 		  (*here).second.setType(j,UNCATEGORIZED_INTERSECTION);
// #ifdef DEBUG
// 		  cout << "setting " << (*here).second.location << " to uncategorized" << endl;
// #endif
// 		  done = true;
// 		  break;
// 		}
		
// 		if((*here).second.getType(j) % GRAZE_INTERSECTION != UNCATEGORIZED_INTERSECTION)
// 		  last = (*here).second.getType(j) % GRAZE_INTERSECTION;

// 		if(edgeDir==1) ++here; 
// 	      }
// 	    }
// 	  }

	  if(numexits != numentrances) {
	    cerr << "inconsistent exits and entrances " << numexits << " " << numentrances << endl;
	    for(k=0; k<numFaceEdges; k++) {
	      edgeId = faceEdges[j][k];
	      for(CellEdgeMap::iterator here = elementEdgeData[edgeId].begin();
		  here!=elementEdgeData[edgeId].end(); ++here) {
		cerr << edgeId << " " << (*here).first << " " << (*here).second.location << " " << (*here).second.getType(j) << endl;
	      }
	    }
	    // set the result to zero so we get a low homogeneity so that this element will change
	    for(m=0; m<bdys.size(); ++m) {
	      (*result)[m] = 0;
	    }
	    return result;

#if DEBUG
	    writeFacets(polypoints, edgeExtractor, boundedInput[i]);
#endif DEBUG
	    //throw ErrProgrammingError("CSkeletonElement: inconsistent number of exits and entrances", __FILE__, __LINE__);
	  }

	  bool done = false;
	  bool started = false;
	  Coord starting_point;
	  CellEdgeMap::iterator here, end;
	  k = 0;
	  inside = 0;
	  while(!done) {
	    // set first point of first edge as the starting point regardless of category
	    for(l=0; l<3; ++l) {
	      x[l] = fpts[k][l];
	      p2[l] = fpts[(k+1)%numFaceEdges][l];
	    }
	    				
	    // get the element edgeId and dir
	    edgeId = faceEdges[j][k];
	    edgeDir = faceEdgeDirs[j][k];

	    if(edgeDir==1) {
	      here = elementEdgeData[edgeId].begin();
	      end = elementEdgeData[edgeId].end();
	    }
	    if(edgeDir==-1) {
	      here = elementEdgeData[edgeId].end();
	      end = elementEdgeData[edgeId].begin();
	    }
							
	    while(here!=end and !done) {

	      // decrement at the beginning of the loop if we are going backwards
	      if(edgeDir==-1) 
		--here;

#ifdef DEBUG
	      cout << "intersection info " << (*here).first << " " << (*here).second.location << " " << (*here).second.getType(j) << endl;
#endif

	      // We add up all the entrances and exits then mod by
	      // grazes so we get the net effect at a given point
	      switch ((*here).second.getType(j) % GRAZE_INTERSECTION) {
	      case EXIT_INTERSECTION:
		(*here).second.location.writePointer(x);
		inside = 1;
		if(!started) {
#ifdef DEBUG
		  cout << "started" << endl;
#endif
		  started = true;
		  starting_point = (*here).second.location;
		}
		else if(started && (*here).second.location == starting_point)
		  done = true;
		break;
	      case ENTRANCE_INTERSECTION:
		if (started && inside) {
		  (*here).second.location.writePointer(y);
#ifdef DEBUG
		  cout << "basearea contrib along el edge: " << x[fpi[j][0]] << " " << x[fpi[j][1]];
		  cout << " to " << (*here).second.location(fpi[j][0]) << " " << (*here).second.location(fpi[j][1]);
		  cout << " contrib: " << (1.0/2.0)*(x[fpi[j][0]]*(*here).second.location(fpi[j][1])-x[fpi[j][1]]*(*here).second.location(fpi[j][0])) << endl;
#endif

		  facearea[j]+=(1.0/2.0)*(x[fpi[j][0]]*(*here).second.location(fpi[j][1])-
				       x[fpi[j][1]]*(*here).second.location(fpi[j][0]));
		}		
		inside = 0;
		break;
	      case UNCATEGORIZED_INTERSECTION:
		break;
	      default:
		break;
	      }

	      // increment at the end of the loop if we are going forwards
	      if(edgeDir==1) 
		++here; 

	    }

	    if(inside && started && !done) {
#ifdef DEBUG
	      cout << "basearea contrib whole or rest of edge: " << x[fpi[j][0]] << " " << x[fpi[j][1]];
	      cout << " to " << p2[fpi[j][0]] << " " << p2[fpi[j][1]];
	      cout << " contrib: " << (1.0/2.0)*(x[fpi[j][0]]*p2[fpi[j][1]]-x[fpi[j][1]]*p2[fpi[j][0]]) << endl;
#endif
	      facearea[j]+=(1.0/2.0)*(x[fpi[j][0]]*p2[fpi[j][1]]-x[fpi[j][1]]*p2[fpi[j][0]]);
	    }
	    k = (k+1) % numFaceEdges;
	  } // end while(!done)     loop over face edges
	}
	else { //if icount == 0
	  cell_locator->SetDataSet(boundedInput[i]);
	  cell_locator->Update();
	  // All of the edges are either entirely in or entirely out
	  // (note that there can still be interior portions of the
	  // face that have a category that differs from that of the
	  // edges).  Traverse the edges until we find a point that is
	  // not near a voxel boundary.
	  bool foundpoint = false;
	  int cellId;
	  int edgeId = 0;
	  for(l=0; l<3; ++l) {
	    x[l] = p1[l] = fpts[0][l]; 
	    p2[l] = fpts[1][l]; 
	  }
	  int loops = 0;
	  while(!foundpoint && edgeId < numFaceEdges && loops < 100) {
	    ++loops;
	    cell_locator->FindClosestPoint(x,y,generic_cell,cellId,subId,dist2);
	    if(dist2>linetol2) {
	      cat = ms.category(Coord(x[0],x[1],x[2]));
	      foundpoint = true;
	    }
	    else {
	      if(vtkMath::Distance2BetweenPoints(x,p2) < linetol2) {
		++edgeId;
 		for(l=0; l<3; ++l) {
		  x[l] = p1[l] = fpts[edgeId][l]; 
		  p2[l] = fpts[(edgeId+1)%numFaceEdges][l]; 
		}
	      }
	      else
		for(l=0; l<3; ++l) x[l] += (p2[l]-p1[l])/10;
	    }
	  }
	  if(!foundpoint) { // the whole face is adjacent to a voxel group boundary
	    // perturb in a direction opposite the face normal
	    for(l=0; l<3; ++l) y[l] = center[l] - fnormal[j][l]*linetol;
	    cat = ms.category(Coord(y[0],y[1],y[2]));
	  }
	  if(cat==i) {
	    for(k=0; k<numFaceEdges; k++) {
	      facearea[j]+=(1.0/2.0)*(fpts[k][fpi[j][0]]*fpts[(k+1)%numFaceEdges][fpi[j][1]]-
				      fpts[k][fpi[j][1]]*fpts[(k+1)%numFaceEdges][fpi[j][0]]);
	    }
	  }
	}

	dot = fnormal[j][0]*center[0] + fnormal[j][1]*center[1] + fnormal[j][2]*center[2]; 
	facearea[j] /= fprojf[j];

#ifdef DEBUG
	cout << "facearea: " << facearea[j] << " fprojf[j]: " << fprojf[j] << " " << facearea[j]*fprojf[j] << endl;
#endif

	  if(facearea[j] > 0) {
	    (*result)[i] += 1.0/3.0 * dot * facearea[j];
	    if(facearea[j] > (1.05)*vtkTriangle::TriangleArea(epts[ptIds[0]], epts[ptIds[1]], epts[ptIds[2]]) ) {
	      cout << "\n\n\nPOINTS: " << endl;
	      for(m=0; m<4; m++){
		cout << epts[m][0] << " " << epts[m][1] << " " << epts[m][2] << endl;
	      }
	      cout << "strange area result " << facearea[j] << endl;
#ifdef DEBUG
	      writeFacets(polypoints, edgeExtractor, boundedInput[i]);
	      throw ErrProgrammingError("Invalid area.", __FILE__, __LINE__);
#endif
	      // set the result to zero so we get a low homogeneity so that this element will change
	      for(m=0; m<bdys.size(); ++m) {
		(*result)[m] = 0;
	      }
	      return result;
	    }
#ifdef DEBUG
	    cout << "center: " << center[0] << " " << center[1] << " " << center[2] << endl;
	    cout << "element face contrib. dot = " << dot << " vol: " << 	(*result)[i] << endl;
#endif
	  }

	  //}
			
      } // end loop over element faces (j)
    } // end if(numFaces)
    else {
      // the element is either entirely inside or entirely outside --
      // get the category at the center of the element.
      x[0] = x[1] = x[2] = 0;
      for(j=0; j<4; ++j) {
	for(k=0; k<3; ++k) x[k] += epts[j][k] / 4.0;
      }
      cat = ms.category(Coord(x[0],x[1],x[2]));
      if(cat==i)
	(*result)[i] = volume();
      else
	(*result)[i] = 0;
    }

  } // end loop over categories (i)

  // separate loop to delete boundedInput because they share objects
  for(i=0; i<bdys.size(); i++) {
    boundedInput[i]->Delete();
  }
  for(i=0; i<4; ++i) eplane[i]->Delete();
  insideinfo->Delete();
  tri->Delete();
  delete [] facearea;

  return result;


  // TODO 3D: cleanup!!!

  //#ifdef DEBUG

//   cerr << "\n\n\nPOINTS: " << endl;
//   double pt[3];
//   for(int i=0; i<4; i++){
//     this->getPoints()->GetPoint(i,pt);
//     cout << pt[0] << " " << pt[1] << " " << pt[2] << endl;
//   }


  //Redundant call to illegal() for debugging
//   if (illegal()) {
//     throw ErrProgrammingError("categoryVolumes called on illegal element.",
//    			      __FILE__, __LINE__);
//  }
  //cerr << "begin categoryVolumes " << volume() << endl;

  //#endif

//   int numElementFaces = this->getNumberOfFaces();
//   int numElementEdges = this->getNumberOfEdges();
//   vtkIdType numFaces, numBdyPoints, numNewPoints, ptId1, ptId2, numPts, *ptIds;
//   vtkIntArray *insideinfo = vtkIntArray::New();
//   double x[3], y[3], closestpt[3], weights[8], dist2, pcoords[3];
//   int nearestId, ptinside, j, k, l, m, inside, outside, onFace;
//   vtkPointsProjectedHull *newPoints = vtkPointsProjectedHull::New(); 
//   vtkPoints *polypoints, *thispoints = this->getPoints();
//   double normal[3], center[3], dot, area, p[4][3];
//   double t, p1[3], p2[3], projfactor, tol;
//   int subId, intersect, hullnum, proji[2], numEdges, numFaceEdges;
//   double *hullpts, basearea;
//   vtkCell *face;
//   vtkPlane *plane = vtkPlane::New();
//   double pcoords1[3], pcoords2[3], weights1[4], weights2[4];
//   int onEdge, p1inside, p2inside;
//   unsigned int i, cat = 0;
//   double facebounds[6], boundedInputBounds[6];
//   int edgeId, edgeDir, icount;
//   bool found;

//   // The tolerance should only be compared to pcoords or weights,
//   // which are in units of an element dimension (as calculated by vtk
//   // routines).  If we need to compare it to anything else, make the
//   // number dimensionless by dividing by the maxLength. We don't want 
//   // tol^2 to be within roundoff of zero.
//   tol = 1e-5;

//   double length, maxLength=0;
//   double bounds[6];
//   this->getBounds(bounds);

//   for(j=0; j<3; j++) {
//     length = bounds[2*j+1]-bounds[2*j];
//     if(length>maxLength)
//       maxLength=length;
//   }

//   // the vtkLine::IntersectWithLine routine is sensitive to the
//   // tolerance and compares a dimensionful number to it.
//   double linetol = tol*maxLength; 
//   double linetol2 = linetol*linetol;

//   int n = ms.nCategories();	// recomputes categories & boundaries if needed
//   std::vector<double> *result = new std::vector<double>(n);

//   // Get all the category boundaries.
//   const std::vector<vtkPolyData*>  &bdys = ms.getCategoryBdys();

//   // first loop over the categories and get portions of the boundaries 
//   // that overlap the bounds of the element.  We also set up one shared
//   // vtkPoints object and avoid using slow vtk objects such as CleanPolyData
//   // and the Locator.
//   std::vector<vtkPolyData*>  boundedInput;
//   // each entry in bdys has the same number of points
//   int numPoints = bdys[0]->GetPoints()->GetNumberOfPoints();
//   vtkPoints *points = vtkPoints::New();
//   points->Allocate(numPoints, numPoints); 
//   PointMap pointMap;
//   PointMap::iterator pmi;
//   int oldId, currId=0;
//   vtkQuad *quad;		
//   vtkDoubleArray *scalarData, *normalData;
//   quad = vtkQuad::New();
//   for(i=0; i<bdys.size(); i++) {
//     (*result)[i] = 0;
//     numFaces = bdys[i]->GetNumberOfPolys();
//     //bdys[i]->BuildCells();
//     boundedInput.push_back(vtkPolyData::New());
//     // This line leads to an Invalid Read in valgrind
//     boundedInput[i]->Allocate(numFaces,numFaces);
//     scalarData = vtkDoubleArray::New();
//     scalarData->SetNumberOfComponents(2);
//     boundedInput[i]->GetCellData()->SetScalars(scalarData);
//     normalData = vtkDoubleArray::New();
//     normalData->SetNumberOfComponents(3);
//     boundedInput[i]->GetCellData()->SetNormals(normalData);		
//     for(j=0; j<numFaces; j++) {
//       bdys[i]->GetCellBounds(j,facebounds);
//       if( boundsOverlap(facebounds,bounds) ) {
// 	// add the points to the point object, id map, and quad
// 	bdys[i]->GetCellPoints(j,numPts,ptIds);
// 	for(k=0; k<4; k++) {
// 	  oldId = ptIds[k];
// 	  pmi = pointMap.find(oldId);
// 	  if(pmi == pointMap.end()) {
// 	    pointMap[oldId] = currId;
// 	    bdys[i]->GetPoints()->GetPoint(oldId,x);
// 	    points->InsertNextPoint(x);
// 	    ++currId;
// 	  }
// 	  else
// 	    bdys[i]->GetPoints()->GetPoint(oldId, x);
// 	  quad->GetPointIds()->InsertId(k,pointMap[oldId]);
// 	  quad->GetPoints()->SetPoint(k,x);
// 	}
// 	// now add the cells
// 	boundedInput[i]->InsertNextCell(VTK_QUAD,quad->GetPointIds());
// 	scalarData->InsertNextTuple(bdys[i]->GetCellData()->GetScalars()->GetTuple(j));
// 	normalData->InsertNextTuple(bdys[i]->GetCellData()->GetNormals()->GetTuple(j));
//       }
//     }
//     boundedInput[i]->SetPoints(points);
//     scalarData->Delete();
//     normalData->Delete();
//   }
//   // this just deletes one reference to the shared points object
//   points->Delete();
//   quad->Delete();

//   // Start by looping over the points in the input and storing
//   // whether they are in, out. 
//   numBdyPoints = points->GetNumberOfPoints();
//   insideinfo->SetNumberOfValues(numBdyPoints);
//   // cout << "gathering insideinfo: " << endl;
//   for(j=0; j<numBdyPoints; j++) {
//     points->GetPoint(j,x);
//     ptinside = this->evaluatePosition(x, closestpt, nearestId, pcoords, dist2, weights);
//     insideinfo->SetValue(j,ptinside);
//     //cerr << j << " " << x[0] << " " << x[1] << " " << x[2] << " " << ptinside << endl;
//     // evaluatePosition can return a -1 if the element is very badly
//     // shaped, has a volume that is close to zero, but was not bad
//     // enough to be illegal.  In this case, we return category volumes
//     // of zero since any further calculations are unreliable.
//     if(ptinside == -1) 
//       return result;
//   }

//   // loop over categories
//   for(i=0; i<bdys.size(); i++) {
// // #ifdef DEBUG
// //     cerr << "\n\nexamining category " << i << endl;
// // #endif

//     CellEdge elementEdgeData(numElementEdges);
//     numFaces = boundedInput[i]->GetNumberOfPolys();
//     newPoints->Allocate(10*numBdyPoints,10*numBdyPoints);
//     normalData = vtkDoubleArray::SafeDownCast(boundedInput[i]->GetCellData()->GetNormals());
//     scalarData = vtkDoubleArray::SafeDownCast(boundedInput[i]->GetCellData()->GetScalars());

// // #ifdef DEBUG
// //     cerr << "phase 1. numFaces: " << numFaces << " numBdyPoints: " << numBdyPoints << endl;
// // #endif

//     for(j=0; j<numFaces; j++) {
//       face = boundedInput[i]->GetCell(j);
//       polypoints = face->GetPoints();
//       normalData->GetTuple(j,normal);

// //       // print polypoints
// // #ifdef DEBUG
// //       cerr << "face " << j << endl;
// //       for(k=0;k<4;k++) {
// // 	polypoints->GetPoint(k,x);
// // 	cout << k << " " << face->GetPointIds()->GetId(k) << " " << x[0] << " " << x[1] << " " << x[2] << endl;
// //       }
// // #endif

//       polypoints->GetPoint(0,x);
//       // We check if the face is actually coincident with
//       // one of the element faces.  If that's the case, the volume
//       // contribution will be added later, so we want to ignore it
//       // here.
//       onFace = this->faceOnElementFace(normal,x,maxLength,tol);

//       if(!onFace) {

// 	newPoints->Reset();
// 	locator->InitPointInsertion(newPoints, bounds);

// 	center[0] = center[1] = center[2] = 0.0;
// 	dot = 0.0;
// 	inside = 0;
// 	outside = 0;
// 	for(k=0; k<4; k++ ){
// 	  polypoints->GetPoint(k,p[k]);

// 	  ptinside = (int)insideinfo->GetValue(face->GetPointIds()->GetId(k));
// // 	  cerr << k << " " << face->GetPointIds()->GetId(k) << " " << insideinfo->GetNumberOfTuples() << " " << ptinside << endl;
// 	  inside |= ptinside;
// 	  outside |= !ptinside;
// 	  if(ptinside) {
// 	    if(locator->InsertUniquePoint(p[k], ptId1))
// 	      for(l=0; l<3; l++) center[l] += p[k][l];
// 	  }
// 	}
		
// 	if(inside && !outside) { // face totally inside
// 	  area = scalarData->GetComponent(j,0);
// 	  dot = scalarData->GetComponent(j,1);
// 	  (*result)[i] += 1.0/3.0 * dot * area;

// // #ifdef DEBUG
// // 	  cerr << "face totally inside. dot = " << dot << " area = " << area << endl;
// // #endif
					
// 	  // Loop over edges of element to find intersections - store points by edge for later.
// 	  for(k=0; k<numElementEdges; k++) {
// 	    this->getEdgePoints(k,ptIds);
// 	    thispoints->GetPoint(ptIds[0],p1);
// 	    thispoints->GetPoint(ptIds[1],p2);
// 	    intersect = face->IntersectWithLine(p1,p2,tol,t,x,pcoords,subId);
// 	    if(intersect) {
// 	      // carefully check that there isn't already an
// 	      // intersection within the tolerance
// 	      found = false;
// 	      for(CellEdgeMap::iterator here = elementEdgeData[k].begin();
// 		  here!=elementEdgeData[k].end(); ++here) 
// 		if( fabs( (*here).first-t ) < tol) found = true;
// 	      if(!found)
// 		elementEdgeData[k].insert(CellEdgeDatum(t,VoxelBdyIntersection(x[0],x[1],x[2])));
// 	    }
// 	  }
// 	} // end if(inside && !outside)

// 	else { 
// 	  // face partially inside - even if all of the points are
// 	  // outside, part of the face could be inside.

// 	  // Loop over edges of element, again, store points by edge for
// 	  // later, but this time we add to the locator and the center.
// 	  for(k=0; k<numElementEdges; k++) {
// 	    this->getEdgePoints(k,ptIds);
// 	    thispoints->GetPoint(ptIds[0],p1);
// 	    thispoints->GetPoint(ptIds[1],p2);
// 	    intersect = face->IntersectWithLine(p1,p2,tol,t,x,pcoords,subId);
// 	    if(intersect) {
// 	      // carefully check that there isn't already an
// 	      // intersection within the tolerance
// 	      found = false;
// 	      for(CellEdgeMap::iterator here = elementEdgeData[k].begin();
// 		  here!=elementEdgeData[k].end(); ++here) 
// 		if( fabs( (*here).first-t ) < tol) found = true;
// 	      if(!found)
// 		elementEdgeData[k].insert(CellEdgeDatum(t,VoxelBdyIntersection(x[0],x[1],x[2])));
// 	      if(locator->InsertUniquePoint(x, ptId1))
// 		for(l=0; l<3; l++) center[l] += x[l];
// 	    }
// 	  }

// 	  // Loop over edges of quad voxel group boundary face.
// 	  for(k=0; k<4; k++) {
// 	    polypoints->GetPoint(k,p1);
// 	    polypoints->GetPoint((k+1)%4,p2);
// 	    intersect = this->intersectWithLine(p1,p2,tol,t,x,pcoords,subId);
// 	    if(intersect) {
// 	      if(locator->InsertUniquePoint(x, ptId1))
// 		for(l=0; l<3; l++) center[l] += x[l];
// 	    }
// 	    // Go the other way because it's possible the line intersects the element in 2 places.
// 	    intersect = this->intersectWithLine(p2,p1,tol,t,x,pcoords,subId);
// 	    if(intersect) {
// 	      if(locator->InsertUniquePoint(x, ptId1))
// 		for(l=0; l<3; l++) center[l] += x[l];
// 	    }
// 	  }

// 	  // Since we are intersecting two convex shapes, the result must
// 	  // be convex.  Also, the normal is always along an axis.  This
// 	  // means we can use vtk's built in methods to find the 2D points
// 	  // that make up the face in CCW order.  Then we can calculate
// 	  // the base area of this pyramidal volume contribution the same
// 	  // way we calculate areas for the 2D homogeneity calculaton.
// 	  numNewPoints = newPoints->GetNumberOfPoints();

// 	  if(numNewPoints > 2) {

// 	    for(k=0; k<3; k++) center[k] /= numNewPoints;
// 	    dot = normal[0]*center[0] + normal[1]*center[1] + normal[2]*center[2]; 
// 	    basearea = 0.0;
// 	    hullnum = 0;
// 	    // hullpts must have TWICE the number of points in newPoints!
// 	    hullpts = new double[2*numNewPoints];
					
// 	    // Since the faces are axis aligned, only one component of the
// 	    // normal is non zero.
// 	    if(fabs(normal[0]))
// 	      hullnum = newPoints->GetCCWHullX(hullpts, numNewPoints);
// 	    if(fabs(normal[1]))
// 	      hullnum = newPoints->GetCCWHullY(hullpts, numNewPoints);
// 	    if(fabs(normal[2]))
// 	      hullnum = newPoints->GetCCWHullZ(hullpts, numNewPoints);
						
// 	    for(k=0; k<hullnum; k++) {
// 	      basearea += (1.0/2.0)*(hullpts[k*2]*hullpts[((k+1)%hullnum)*2+1]-hullpts[k*2+1]*hullpts[((k+1)%hullnum)*2]);
// 	    }

// 	    (*result)[i] += 1.0/3.0 * dot * basearea;

// // #ifdef DEBUG
// // 	    cerr << "face partially inside. dot = " << dot << " basearea = " << basearea << endl;
// // #endif

// 	    delete [] hullpts;

// 	  } // end if(numNewPoints)
// 	} // end else
//       } // end if (!onFace)
//     } // end loop over faces (j)

// // #ifdef DEBUG
// //     cerr << "end phase 1. vol so far: " << (*result)[i] << endl;
// // #endif

//     // Now find the contribution of the faces of the element.  For
//     // this, we find the intersection of the voxel group boundaries
//     // with each face of the element.  This is a polydata with a
//     // series of line cells, oriented counterclockwise about the voxel
//     // group in question.  We then traverse each part and look for
//     // intersections with the element face (stored earlier).  That is, 
//     // if we found any faces in the voxel group boundary that overlap 
//     // the bounds of the element.  If numFaces=0, then either the whole 
//     // element is in this category, or it's out.
//     projfactor = 0;
//     proji[0] = proji[1] = -1;
//     if(numFaces) {
//       boundedInput[i]->GetBounds(boundedInputBounds);
//       for(j=0; j<numElementFaces; ++j) {
// 	face = this->getFace(j);
// 	numFaceEdges = 3; // only tetra for now. face->GetNumberOfEdges();
// 	face->GetBounds(facebounds);
// 	polypoints = face->GetPoints();
// 	vtkPolygon::ComputeNormal(polypoints,normal);	

// 	// We find the axis direction which has the maximum contribution
// 	// to the normal.  We can save floating point operations later on
// 	// by projecting onto this direction, calculating the 2D cross
// 	// product, and then dividing by the cosine of the normal to find
// 	// the area contribution for each segment.  The sign of the
// 	// projfactor will correct for whether the bounding segments are
// 	// oriented CCW when projected onto the given axis oriented plane.
// 	if(fabs(normal[0]) >= fabs(normal[1]) && fabs(normal[0]) >= fabs(normal[2])) {
// 	  projfactor = normal[0];
// 	  proji[0]=1;
// 	  proji[1]=2;
// 	}
// 	else if(fabs(normal[1]) >= fabs(normal[0]) && fabs(normal[1]) >= fabs(normal[2])) {
// 	  projfactor = normal[1];
// 	  proji[0]=2;
// 	  proji[1]=0;
// 	}
// 	else if(fabs(normal[2]) >= fabs(normal[0]) && fabs(normal[2]) >= fabs(normal[1])) {
// 	  projfactor = normal[2];
// 	  proji[0]=0;
// 	  proji[1]=1;
// 	}
			
// 	// Find where voxel group boundaries intersect the plane
// 	if(boundsOverlap(facebounds,boundedInputBounds)) {
// 	  plane->SetNormal(normal);
// 	  polypoints->GetPoint(0,x);
// 	  plane->SetOrigin(x);
// 	  edgeExtractor->SetInputConnection(0, boundedInput[i]->GetProducerPort());
// 	  edgeExtractor->SetCutFunction(plane);
// 	  edgeExtractor->Update();
// 	  //edges = edgeExtractor->GetOutput();
// 	  //edges->Update();
// 	  edgeExtractor->GetOutput()->BuildCells();
// 	  numEdges = edgeExtractor->GetOutput()->GetNumberOfCells();
// 	}
// 	else
// 	  numEdges = 0;
// // #ifdef DEBUG
// // 	cout << "\n\n\nelement face: " << j << endl;
// // 	cout << "proj dir: " << (proji[1]+1)%3 << endl;
// // 	cout << "got edges in plane " << numEdges << endl;
// // 	for(k=0; k<numEdges; k++) {
// // 	  edgeExtractor->GetOutput()->GetCell(k)->GetPoints()->GetPoint(0,p1);
// // 	  edgeExtractor->GetOutput()->GetCell(k)->GetPoints()->GetPoint(1,p2);
// // 	  cout << "edge " << k << " " << p1[0] << " " << p1[1] << " " << p1[2] << " to ";
// // 	  cout << p2[0] << " " << p2[1] << " " << p2[2] << endl;
// // 	}
// // 	cout << "face points: " << endl;
// // 	for(k=0; k<numFaceEdges; k++) {
// // 	  polypoints->GetPoint(k,x);
// // 	  cout << x[0] << " " << x[1] << " " << x[2] << endl;
// // 	}
// // #endif 

// 	//cout << "before reset / init" << endl;
// 	newPoints->Reset();
// 	//cerr << "between reset and init" << endl;
// 	//cout << bounds[0] << " " << bounds[1] << " " << bounds[2] << " ";
// 	//cout << bounds[3] << " " << bounds[4] << " " << bounds[5] << endl;
// 	locator->InitPointInsertion(newPoints, bounds);
// 	//cerr << "after reset / init" << endl;
// 	CellEdgeMap::iterator cmi;
// 	// initialize this for each of the edges in the face such that
// 	// it has the same parametric coords as the CellEdgeData and
// 	// an intersection type of UNCATEGORIZED_INTERSECTION.  See
// 	// note above -- this is where we die if we don't have the
// 	// duplicates in the points object.
// 	for(k=0; k<numFaceEdges; k++) {
// 	  edgeId = faceEdgeToElementEdge(j,k);
// 	  for(CellEdgeMap::iterator here = elementEdgeData[edgeId].begin();
// 	      here!=elementEdgeData[edgeId].end(); ++here) {
// 	    (*here).second.type = UNCATEGORIZED_INTERSECTION;
// 	  }
// 	}

// 	basearea = 0.0;
// 	center[0] = center[1] = center[2] = 0.0;
// 	dot = 0.0;

// 	// Need to handle case where all points are inside, or all
// 	// points are outside (no intersections).  For each edge, we
// 	// want to handle cases of 0, 1, or 2 intersections.
			
// 	// Loop over the edges, testing for intersections and
// 	// whether they are inside or outside the element face.
// 	// Consider the following cases: 1. If both points are inside,
// 	// then there must be no intersections, and the whole segment
// 	// contributes to the area. 2. If one point is in, and one point
// 	// is out, there must be only one intersection, and the segment
// 	// between the intersection and the inside point contribute to
// 	// the area. 3.  If there are no intersections and both points
// 	// are outside, then there is no contribution.  4. If both
// 	// points are outside, and there are two intersections, then the
// 	// segment between the intersections contributes to the area.
			
// 	// Note, the segments are directed so that each segment is CCW
// 	// around the material in question, WRT the plane normal using the
// 	// RHR.  BUT the segments are NOT necessarily in order, which
// 	// actually doesn't matter.

// 	for(k=0; k<numEdges; k++) {
// 	  edgeExtractor->GetOutput()->GetCellPoints(k,numPts,ptIds);
// 	  edgeExtractor->GetOutput()->GetPoints()->GetPoint(ptIds[0],p1);
// 	  edgeExtractor->GetOutput()->GetPoints()->GetPoint(ptIds[1],p2);
// 	  // If the points are actually on the edge, we treat them as
// 	  // if they are outside because there are cases where one
// 	  // point is in and one point is on the edge and there is no
// 	  // external segment which will give us the same intersection
// 	  // such as when the element is on the edge of the
// 	  // microstructure.  We do this so that we definitely find
// 	  // the intersection.  The consequence is that we have to
// 	  // account for possibly finding and classifying the same
// 	  // intersection twice.
// 	  // TODO 3D: we should handle the case that evaluate position returns -1
// 	  p1inside = face->EvaluatePosition(p1, closestpt, nearestId, pcoords1, dist2, weights1);
// 	  onEdge = pointOnFaceEdge(polypoints, p1, linetol2);
// 	  if(onEdge) p1inside = 0;
// 	  p2inside = face->EvaluatePosition(p2, closestpt, nearestId, pcoords2, dist2, weights2);
// 	  onEdge = pointOnFaceEdge(polypoints, p2, linetol2);
// 	  if(onEdge) p2inside = 0;
// // #ifdef DEBUG
// // 	  cout << "p1: " << p1[0] << " " << p1[1] << " " << p1[2] << " " << p1inside << endl;
// // 	  cout << "p2: " << p2[0] << " " << p2[1] << " " << p2[2] << " " << p2inside << endl;
// // #endif 
				
// 	  // case 1 - both points inside
// 	  if(p1inside && p2inside){
					
// 	    //onEdge = edgeOnFaceEdge(numFaceEdges, weights1, weights2, tol);
// 	    onEdge = edgeOnFaceEdge(polypoints, p1, p2, linetol2);
					
// 	    if(!onEdge) {
// // #ifdef DEBUG
// // 	      cout << "basearea contrib both points inside: " << p1[proji[0]] << " " << p1[proji[1]];
// // 	      cout << " to " << p2[proji[0]] << " " << p2[proji[1]];
// // 	      cout << " contrib: " << (1.0/2.0)*(p1[proji[0]]*p2[proji[1]]-p1[proji[1]]*p2[proji[0]]) << endl;
// // #endif
// 	      basearea+=(1.0/2.0)*(p1[proji[0]]*p2[proji[1]]-p1[proji[1]]*p2[proji[0]]);
// 	      if(locator->InsertUniquePoint(p1, ptId1))
// 		for(l=0; l<3; l++) center[l] += p1[l];
// 	      if(locator->InsertUniquePoint(p2, ptId1))
// 		for(l=0; l<3; l++) center[l] += p2[l];
// 	    }
					
// 	  } // end (p1inside and p2inside)
				
// 	  // case 2 - one point inside and one point outside
// 	  if(p1inside != p2inside) {

// 	    if(p1inside) { // this is an exit
// 	      findOneIntersection(elementEdgeData, numFaceEdges, j, p1, p2, x, EXIT_INTERSECTION);
// 	      basearea+=(1.0/2.0)*(p1[proji[0]]*x[proji[1]]-p1[proji[1]]*x[proji[0]]);
// 	      if(locator->InsertUniquePoint(x, ptId1))
// 		for(m=0; m<3; m++) center[m] += x[m];
// 	      if(locator->InsertUniquePoint(p1, ptId1))
// 		for(l=0; l<3; l++) center[l] += p1[l];
// 	      //cout << "on edge " << edgeOnFaceEdge(polypoints, p1, x, linetol2) << " " << (1.0/2.0)*(p1[proji[0]]*x[proji[1]]-p1[proji[1]]*x[proji[0]]) << endl;
// 	    }
// 	    else if(p2inside) { // this is an entrance
// 	      findOneIntersection(elementEdgeData, numFaceEdges, j, p2, p1, x, ENTRANCE_INTERSECTION);
// 	      basearea+=(1.0/2.0)*(x[proji[0]]*p2[proji[1]]-x[proji[1]]*p2[proji[0]]);
// 	      if(locator->InsertUniquePoint(x, ptId1))
// 		for(m=0; m<3; m++) center[m] += x[m];
// 	      if(locator->InsertUniquePoint(p2, ptId2))
// 		for(l=0; l<3; l++) center[l] += p2[l];
// 	      //cout << "on edge " << edgeOnFaceEdge(polypoints, p2, x, linetol2) << " " << (1.0/2.0)*(x[proji[0]]*p2[proji[1]]-x[proji[1]]*p2[proji[0]]) << endl;
// 	    }

// 	  } // end case 2

// 	  // cases 3 and 4 - both points outside
// 	  if(!p1inside and !p2inside) {

// 	    if(findTwoIntersections(elementEdgeData, face, numFaceEdges, j, p1, p2, x, y, linetol2, tol)) {
// 	      basearea+=(1.0/2.0)*(x[proji[0]]*y[proji[1]]-x[proji[1]]*y[proji[0]]);
// 	      if(locator->InsertUniquePoint(x, ptId1))
// 		for(m=0; m<3; m++) center[m] += x[m];
// 	      if(locator->InsertUniquePoint(y, ptId2))
// 		for(m=0; m<3; m++) center[m] += y[m];
// 	    }

// 	    // There cannot be more than two intersections.  We ignore
// 	    // the case of 0 or 1 (corner graze) intersections.  So we
// 	    // need to find two intersections and order them by t (the
// 	    // parametric distance along p1 to p2).  The intersections
// 	    // should be along two different edges of the cell.  If
// 	    // they are on the same edge, we ignore it.  It's also
// 	    // possible that one of the intersections is on a corner
// 	    // and therefore appears as two coincident intersections.

// 	  } // end cases 3 and 4
				
// 	} // end loop over edges (k)


// 	// Now loop over the edges in the face and add the
// 	// contributions between the exits and the entrances. 
// 	icount = 0;
// 	int numexits = 0;
// 	int numentrances = 0;
// 	for(k=0; k<numFaceEdges; k++) {
// 	  edgeId = faceEdgeToElementEdge(j,k);
// 	  for(CellEdgeMap::iterator here = elementEdgeData[edgeId].begin();
// 	      here!=elementEdgeData[edgeId].end(); ++here) {
// 	    if ( (*here).second.type % GRAZE_INTERSECTION == EXIT_INTERSECTION ) {
// 	      numexits++;
// 	      icount++;
// 	    }
// 	    if ( (*here).second.type % GRAZE_INTERSECTION == ENTRANCE_INTERSECTION ) {
// 	      numentrances++;
// 	      icount++;
// 	    }
// 	  }
// 	}

// // #ifdef DEBUG
// //  	cout << "icount " << icount << endl;
// // #endif
// 	if(icount >= 2) {

// 	  if(numexits != numentrances) {
// 	    cout << "inconsistent exits and entrances " << numexits << " " << numentrances << endl;
// 	    for(k=0; k<numFaceEdges; k++) {
// 	      edgeId = faceEdgeToElementEdge(j,k);
// 	      for(CellEdgeMap::iterator here = elementEdgeData[edgeId].begin();
// 		  here!=elementEdgeData[edgeId].end(); ++here) {
// 		cout << edgeId << " " << (*here).first << " " << (*here).second.location << " " << (*here).second.type << endl;
// 	      }
// 	    }


// 	    int ids[3] = {0,1,2};
// 	    vtkPolyData *onetri = vtkPolyData::New();
// 	    onetri->SetPoints(face->GetPoints());
// 	    onetri->Allocate(1,1);
// 	    onetri->InsertNextCell(VTK_TRIANGLE,3,ids);

// 	    vtkFacetWriter *writer = vtkFacetWriter::New();
// 	    writer->SetInput(edgeExtractor->GetOutput());
// 	    writer->SetFileName("../../3DSandbox/testscripts/edges");
// 	    writer->Write();
	    
// 	    writer->SetInput(onetri);
// 	    writer->SetFileName("../../3DSandbox/testscripts/tri");
// 	    writer->Write();

// 	    writer->SetInput(boundedInput[i]);
// 	    writer->SetFileName("../../3DSandbox/testscripts/boundedinput");
// 	    writer->Write();

// 	    throw ErrProgrammingError("CSkeletonElement: inconsistent number of exits and entrances", __FILE__, __LINE__);
// 	  }


// 	  bool done = false;
// 	  bool started = false;
// 	  inside = 0;
// 	  Coord starting_point;
// 	  k = 0;
// 	  CellEdgeMap::iterator here, end;
// 	  while(!done) {
// 	    // set first point of first edge as the starting point regardless of category
// 	    //face->GetEdge(k)->GetPoints()->GetPoint(0,x);
// 	    //face->GetEdge(k)->GetPoints()->GetPoint(1,p2);
// 	    polypoints->GetPoint(k, x);
// 	    polypoints->GetPoint((k+1)%numFaceEdges, p2);
// 	    if(inside) {
// 	      if(locator->InsertUniquePoint(x, ptId1))
// 		for(l=0; l<3; l++) center[l] += x[l];
// 	    }
				
// 	    // Retrieve the intersections that we found earlier. There can
// 	    // be any number.

// 	    // get the element edgeId and dir
// 	    edgeId = faceEdgeToElementEdge(j,k);
// 	    edgeDir = faceEdgeToElementEdgeDir(j,k);

// // #ifdef DEBUG
// // 	    cout << "edge info " << edgeId << " " << edgeDir << endl;
// // 	    cout << "CellEdgeMap size = " << elementEdgeData[edgeId].size() << endl;
// // 	    cout << "inside " << inside << endl;
// // #endif

// 	    if(edgeDir==1) {
// 	      here = elementEdgeData[edgeId].begin();
// 	      end = elementEdgeData[edgeId].end();
// 	    }
// 	    if(edgeDir==-1) {
// 	      here = elementEdgeData[edgeId].end();
// 	      end = elementEdgeData[edgeId].begin();
// 	    }
							
// 	    while(here!=end and !done) {

// 	      // decrement at the beginning of the loop if we are going backwards
// 	      if(edgeDir==-1) 
// 		--here;

// // #ifdef DEBUG
// // 	      cout << "intersection info " << (*here).first << " " << (*here).second.location << " " << (*here).second.type << endl;
// // #endif

// 	      // We add up all the entrances and exits then mod by
// 	      // grazes so we get the net effect at a given point
// 	      switch ((*here).second.type % GRAZE_INTERSECTION) {
// 	      case EXIT_INTERSECTION:
// 		inside = 1;
// 		(*here).second.location.writePointer(x);
// 		if(locator->InsertUniquePoint(x,ptId1))
// 		  for(l=0; l<3; l++) center[l]+=x[l];
// 		if(!started) {
// // #ifdef DEBUG
// // 		  cout << "started" << endl;
// // #endif
// 		  started = true;
// 		  starting_point = (*here).second.location;
// 		}
// 		else if(started && (*here).second.location == starting_point)
// 		  done = true;
// 		break;
// 	      case ENTRANCE_INTERSECTION:
// 		if (started && inside) {
// 		  (*here).second.location.writePointer(y);
// 		  if(locator->InsertUniquePoint(y,ptId1))
// 		    for(l=0; l<3; l++) center[l]+=y[l];
// // #ifdef DEBUG
// // 		  cout << "basearea contrib along el edge: " << x[proji[0]] << " " << x[proji[1]];
// // 		  cout << " to " << (*here).second.location(proji[0]) << " " << (*here).second.location(proji[1]);
// // 		  cout << " contrib: " << (1.0/2.0)*(x[proji[0]]*(*here).second.location(proji[1])-x[proji[1]]*(*here).second.location(proji[0])) << endl;
// // #endif
// 		  basearea+=(1.0/2.0)*(x[proji[0]]*(*here).second.location(proji[1])-
// 				       x[proji[1]]*(*here).second.location(proji[0]));
// 		}		
// 		inside = 0;
// 		break;
// 	      case UNCATEGORIZED_INTERSECTION:
// 		break;
// 	      default:
// 		break;
// 	      }

// 	      // increment at the end of the loop if we are going forwards
// 	      if(edgeDir==1) 
// 		++here; 

// 	    }

// 	    if(inside && started && !done) {
// // #ifdef DEBUG
// // 	      cout << "basearea contrib whole or rest of edge: " << x[proji[0]] << " " << x[proji[1]];
// // 	      cout << " to " << p2[proji[0]] << " " << p2[proji[1]];
// // 	      cout << " contrib: " << (1.0/2.0)*(x[proji[0]]*p2[proji[1]]-x[proji[1]]*p2[proji[0]]) << endl;
// // #endif
// 	      basearea+=(1.0/2.0)*(x[proji[0]]*p2[proji[1]]-x[proji[1]]*p2[proji[0]]);
// 	    }
// 	    k = (k+1) % numFaceEdges;
// 	  } // end while(!done)     loop over face edges
// 	}
// 	else { //if icount == 0
// 	  cell_locator->SetDataSet(boundedInput[i]);
// 	  cell_locator->Update();

// 	  // All of the edges are either entirely in or entirely out
// 	  // (note that there can still be interior portions of the
// 	  // face that have a category that differs from that of the
// 	  // edges).  Traverse the edges until we find a point that is
// 	  // not near a voxel boundary.
// 	  bool foundpoint = false;
// 	  int cellId;
// 	  int edgeId = 0;
// 	  polypoints->GetPoint(0, p1);
// 	  polypoints->GetPoint(1, p2);
// 	  for(l=0; l<3; ++l) x[l] = p1[l];
// 	  int loops = 0;
// 	  while(!foundpoint && edgeId < numFaceEdges && loops < 100) {
// 	    ++loops;
// 	    cell_locator->FindClosestPoint(x,y,cellId,subId,dist2);
// // #ifdef DEBUG
// // 	    cout << "looking for safe point on edge" << endl;
// // 	    cout << dist2 << " " << x[0] << " " << x[1] << " " << x[2] << endl;
// // #endif
// 	    if(dist2>linetol2) {
// 	      cat = ms.category(Coord(x[0],x[1],x[2]));
// 	      foundpoint = true;
// 	    }
// 	    else {
// 	      if(vtkMath::Distance2BetweenPoints(x,p2) < linetol2) {
// 		++edgeId;
// 		polypoints->GetPoint(edgeId, p1);
// 		polypoints->GetPoint((edgeId+1)%numFaceEdges, p2);
// 		for(l=0; l<3; ++l) x[l] = p1[l];
// 	      }
// 	      else
// 		for(l=0; l<3; ++l) x[l] += (p2[l]-p1[l])/10;
// 	    }
// 	  }
// 	  if(!foundpoint) { // the whole face is adjacent to a voxel group boundary
// // #ifdef DEBUG
// // 	    cout << "whole face is on voxel group boundary" << endl;
// // 	    // cout << "double check polypoints:" << endl;
// // // 	    for(k=0; k<numFaceEdges; k++) {
// // // 	      polypoints->GetPoint(k,x);
// // // 	      cout << x[0] << " " << x[1] << " " << x[2] << endl;
// // // 	    }
// // #endif
// 	    // perturb inwards, this should be safe as long as the tol
// 	    // is small compared to pixel size
// 	    polypoints->GetPoint(0, x);
// 	    double pcenter[3];
// 	    // TODO 3D: we should handle the case that evaluate position returns -1
// 	    this->evaluatePosition(x, closestpt, nearestId, pcoords, dist2, weights);
// 	    this->getParametricCenter(pcenter);
// 	    for(k=0; k<3; k++)
// 	      pcoords[k] = pcenter[k]+(pcoords[k]-pcenter[k])*(1-tol);
// 	    this->evaluateLocation(subId,pcoords,y,weights);
// 	    //cout << "x: " << x[0] << " " << x[1] << " " << x[2] << endl;
// 	    //cout << "y: " << y[0] << " " << y[1] << " " << y[2] << endl;
// 	    cat = ms.category(Coord(y[0],y[1],y[2]));
// 	    //cout << cat << endl;
// 	  }
// 	  if(cat==i) {
// 	    vtkIndent indent;
// 	    //cout << face->GetClassName() << endl;
// 	    //face->PrintSelf(cout, indent);
// 	    for(k=0; k<numFaceEdges; k++) {
// 	      polypoints->GetPoint(k, p1);
// 	      polypoints->GetPoint((k+1)%numFaceEdges, p2);
// 	      //cout << p1[0] << " " << p1[1] << " " << p1[2] << endl;
// 	      //cout << p2[0] << " " << p2[1] << " " << p2[2] << endl;
// 	      if(locator->InsertUniquePoint(p1, ptId1))
// 		for(l=0; l<3; l++) center[l] += p1[l];
// // #ifdef DEBUG
// // 	      cout << "basearea contrib whole edge: " << p1[proji[0]] << " "<< p1[proji[1]];
// // 	      cout << " to " << p2[proji[0]] << " " << p2[proji[1]];
// // 	      cout << " contrib: " << (1.0/2.0)*(p1[proji[0]]*p2[proji[1]]-p1[proji[1]]*p2[proji[0]]) << endl;
// // #endif
// 	      basearea+=(1.0/2.0)*(p1[proji[0]]*p2[proji[1]]-p1[proji[1]]*p2[proji[0]]);
// 	    }
// 	  }
// 	}

// 	numNewPoints = newPoints->GetNumberOfPoints();
// 	// We add a contribution if there were more than two internal
// 	// and intersection points and the basearea is positive (clamp
// 	// to zero otherwise).
// 	if(numNewPoints > 2) { 
// 	  for(k=0; k<3; k++) center[k] /= numNewPoints;
// 	  dot = normal[0]*center[0] + normal[1]*center[1] + normal[2]*center[2]; 
// 	  basearea /= projfactor;

// 	  if(basearea > 0) {
// 	    (*result)[i] += 1.0/3.0 * dot * basearea;
// 	  }
// // #ifdef DEBUG
// // 	  cout << "basearea: " << basearea << " projfactor: " << projfactor << " " << basearea*projfactor << endl;
// // 	  cout << "center: " << center[0] << " " << center[1] << " " << center[2] << endl;
// // 	  cout << "element face contrib. dot = " << dot << " basearea = " << basearea;
// // 	  cout << " projfactor: " << projfactor << " vol: " << 	(*result)[i] << endl;
// // #endif

// 	}
// 	if(numEdges)
// 	  edgeExtractor->RemoveAllInputs();
			
//       } // end loop over element faces (j)
//     } // end if(numFaces)
//     else {
//       // the element is either entirely inside or entirely outside --
//       // get the category at the center of the element.
//       this->getParametricCenter(y); //gives parametric coordinates
//       this->evaluateLocation(subId,y,x,weights); // converts to physical coordinates
//       cat = ms.category(Coord(x[0],x[1],x[2]));
//       if(cat==i)
// 	(*result)[i] = this->volume();
//       else
// 	(*result)[i] = 0;
//     }

// // #ifdef DEBUG
// //     cerr << "end phase 2. " << endl;
// // #endif

//   } // end loop over categories (i)

//   // separate loop to delete boundedInput because they share objects
//   for(i=0; i<bdys.size(); i++) {
//     boundedInput[i]->Delete();
//   }

//   plane->Delete();
//   newPoints->Delete();
//   insideinfo->Delete();

//   return result;

}



#endif

// OBSOLETE.  Not yet removed, because it may be the only caller of
// some of the other routines, and they may need to be removed also.
// const DoubleVec *
// CSkeletonElement::old_categoryAreas(CMicrostructure &ms) const {
//   int ncategories = ms.nCategories();
//   //std::cerr << "before underlying_pixels(ms)" << std::endl;
//   const std::vector<ICoord> *pixels = underlying_pixels(ms);
//   //std::cerr << "after underlying_pixels(ms)" << std::endl;
//   // Find the categories of the underlying pixels, and determine if
//   // they're all the same.  If they are, we can skip the fractional
//   // area calculation.
//   if(pixels->size() == 0)
//     throw ErrProgrammingError("No pixels under element!", __FILE__, __LINE__);
//   std::vector<int> categories(pixels->size());
//   bool uniform = true;
//   categories[0] = ms.category((*pixels)[0]);
//   for(std::vector<int>::size_type i=1; i<categories.size(); i++) {
//     categories[i] = ms.category((*pixels)[i]);
//     uniform = uniform && (categories[i]==categories[i-1]);
//   }

//   DoubleVec *areas = new DoubleVec(ncategories, 0.0);

//   if(uniform) {
//     (*areas)[categories[0]] = area();
//   }
//   else {
//     // Fractional area calculation.
//     Coord pxlsize(ms.sizeOfPixels());
//     for(std::vector<ICoord>::size_type i=0; i<pixels->size(); i++) {
//       int category = categories[i];
//       (*areas)[category] += pixel_intersection(&(*pixels)[i], &pxlsize);
//     }
//   }
//   delete pixels;
  
//   return areas;
// } // end of old_categoryAreas

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

double CSkeletonElement::homogeneity(const CMicrostructure &ms) const {
  findHomogeneityAndDominantPixel(ms);
  return homogeneityData.value().get_homogeneity();
}

int CSkeletonElement::dominantPixel(const CMicrostructure &ms) const {
  findHomogeneityAndDominantPixel(ms);
  return homogeneityData.value().get_dominantpixel();
};

double CSkeletonElement::energyHomogeneity(const CMicrostructure &ms) const {
  findHomogeneityAndDominantPixel(ms);
  return homogeneityData.value().get_energy();
};

void CSkeletonElement::copyHomogeneity(const CSkeletonElement &other) {
  homogeneityData.copy(other.homogeneityData);
}

void CSkeletonElement::revertHomogeneity() {
  homogeneityData.revert();
}

HomogeneityData CSkeletonElement::getHomogeneityData() const {
  return homogeneityData.value();
}

void CSkeletonElement::setHomogeneityData(const HomogeneityData &hd) {
  homogeneityData.set_value(hd);
}

void CSkeletonElement::setHomogeneous(int dompxl) {
  HomogeneityData hd(1.0, dompxl);
  homogeneityData.set_value(hd);
}

void CSkeletonElement::findHomogeneityAndDominantPixel(
						       const CMicrostructure &ms)
  const
{
  // Only recompute if the microstructure has changed or nodes have moved
  if(homogeneityData.timestamp() > ms.getTimeStamp()) {
    bool uptodate = true;
    for(int i=0; i<nnodes() && uptodate; i++) {
      if(homogeneityData.timestamp() < nodes[i]->nodemoved) 
	uptodate = false;
    }
    if(uptodate)
      return;
  }
  homogeneityData.set_value(c_homogeneity(ms)); // recompute
}

// Find the homogeneity of the current element.  Element is assumed to
// be legal -- categoryAreas will throw an exception if it isn't.
HomogeneityData CSkeletonElement::c_homogeneity(const CMicrostructure &ms)
  const
{
#if DIM==2
  const DoubleVec *areas = categoryAreas(ms);
  int category = 0;
  double maxarea=0.0;
  for(DoubleVec::size_type i=0;i<areas->size();++i) {
    double area = (*areas)[i];
    if(area > maxarea) {
      maxarea = area;
      category = i;
    }
  }
  double homogeneity = maxarea/area();
  delete areas;
#elif DIM==3
  //TEMPORARY
  double maxvolume = 0, totalVolume = 0;
  int category = 0;

  // if the element is illegal, categoryVolumes is unreliable and not worth calling.
  if(illegal())
    return HomogeneityData(0, 0);

  const DoubleVec *volumes = categoryVolumes(ms);

  for(DoubleVec::size_type i=0;i<volumes->size();++i) {
    double volume = (*volumes)[i];
    totalVolume += volume;
    if(volume > maxvolume) {
      maxvolume = volume;
      category = i;
    }
    //if(volume < 0) {
    //  cerr << i << " volume: " << volume;
    //  throw ErrProgrammingError("Invalid volume.", __FILE__, __LINE__);
    //}
  }
#ifdef DEBUG
   if( fabs(totalVolume-volume())/volume() > 1e-3 && volume() > 1e-3 ) {  
    for(DoubleVec::size_type i=0;i<volumes->size();++i)
      cerr << i << " " << (*volumes)[i] << endl;
    cerr << totalVolume << " " << volume() << " " << fabs(totalVolume-volume())/volume() << endl;
    cerr << energyShape() << endl;
    cerr << "POINTS:" << endl;
    double pt[3];
    int id;
    for(int i=0; i<4; i++){
      this->getPoints()->GetPoint(i,pt);
      id = this->getPointIds()->GetId(i);
      cerr << id << " " << pt[0] << " " << pt[1] << " " << pt[2] << endl;
    }
    cerr << endl;
//     throw ErrProgrammingError("Invalid total volume.", __FILE__, __LINE__);
   }
#endif

  double homogeneity = maxvolume/volume();
  delete volumes;
#endif
  if (homogeneity>1.0) homogeneity=1.0;
  return HomogeneityData(homogeneity, category);
}

// Find the point along edge i where the pixel category in the
// Microstructure changes.  Return true iff there's just one such
// point.
bool CSkeletonElement::transitionPoint(CMicrostructure &microstructure, int i,
				       Coord *point)
  const
{
#if DIM==2
  return microstructure.transitionPoint(nodes[i]->position(),
					nodes[(i+1)%nnodes()]->position(),
					point);
#elif DIM==3
  double x0[3], x1[3];
  this->getEdge(i)->GetPoints()->GetPoint(0,x0);
  this->getEdge(i)->GetPoints()->GetPoint(1,x1);
  Coord c0(x0[0],x0[1],x0[2]);
  Coord c1(x1[0],x1[1],x1[2]);
  return microstructure.transitionPoint(c0,c1,point);
#endif
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
#if DIM==2

CSkeletonQuad::CSkeletonQuad(CSkeletonNode *n0, CSkeletonNode *n1,
			     CSkeletonNode *n2, CSkeletonNode *n3)
  : CSkeletonElement(4)
{
  nodes[0] = n0;
  nodes[1] = n1;
  nodes[2] = n2;
  nodes[3] = n3;
}

double CSkeletonQuad::shapefun(int i, const MasterCoord &mc) const {
  switch(i) {
  case 0:
    return  0.25*(mc(0) - 1.0)*(mc(1) - 1.0);
  case 1:
    return -0.25*(mc(0) + 1.0)*(mc(1) - 1.0);
  case 2:
    return  0.25*(mc(0) + 1.0)*(mc(1) + 1.0);
  case 3:
    return -0.25*(mc(0) - 1.0)*(mc(1) + 1.0);
  }
  throw ErrBadIndex(i, __FILE__, __LINE__);
}

bool CSkeletonQuad::interior(const Coord *point) const {
  //  3________2
  //   |     /|
  //   | L  / |     If a point is inside of either triangle,    
  //   |   /  |     then it is inside the quad.
  //   |  /   |
  //   | / R  |
  //   |/     |
  //   --------
  //   0      1
  Coord p0 = nodes[0]->position();
  Coord p1 = nodes[1]->position();
  Coord p2 = nodes[2]->position();
  Coord p3 = nodes[3]->position();
  double areaL = triangleArea(p0, p2, p3);
  double areaR = triangleArea(p0, p1, p2);
  if(areaR > 0.0 && areaL > 0.0)
    return insideTriangle(p0,p2,p3, *point) || insideTriangle(p0,p1,p2, *point);
  if(areaR < 0.0 && areaL > 0.0)
    // Boomerang shape.  Vertex 1 may lie inside L.  point must lie
    // within L, but not within the inverted R.
    return insideTriangle(p0,p2,p3, *point) && !insideTriangle(p0,p2,p1,*point);
  if(areaR > 0.0 && areaL < 0.0)
    return insideTriangle(p0,p1,p2, *point) && !insideTriangle(p0,p3,p2,*point);
  return false;
}

double CSkeletonQuad::area() const {
  return
    triangleArea(nodes[0]->position(), nodes[2]->position(),
		 nodes[3]->position()) +
    triangleArea(nodes[0]->position(), nodes[1]->position(),
		 nodes[2]->position());
}

bool CSkeletonQuad::illegal() const {
  Coord p0 = nodes[0]->position();
  Coord p1 = nodes[1]->position();
  Coord p2 = nodes[2]->position();
  Coord p3 = nodes[3]->position();

  //   // Using Area() in a thorough way.
  //   double a1 = (p1-p0)%(p3-p0);
  //   double a2 = (p2-p1)%(p0-p1);
  //   double a3 = (p3-p2)%(p1-p2);
  //   double a4 = (p0-p3)%(p2-p3);
  //   if ( (a1<=0) || (a2<=0) || (a3<=0) || (a4<=0) )
  //     return true;
  //   if ( (fabs((a1+a3)/(a2+a4))-1.0) > LEGAL_ELEMENT_AREA_TOLERANCE )
  //     return true;
  //   return false;

  //   // Using Area() only.
  //   return ((p1 - p0) % (p3 - p0) <= 0.0 ||
  // 	  (p2 - p1) % (p0 - p1) <= 0.0 ||
  // 	  (p3 - p2) % (p1 - p2) <= 0.0 ||
  // 	  (p0 - p3) % (p2 - p3) <= 0.0);

  // Using Area() & Cosine().
  return  ( (p1 - p0) % (p3 - p0) <= 0.0 ||
	    (p2 - p1) % (p0 - p1) <= 0.0 ||
	    (p3 - p2) % (p1 - p2) <= 0.0 ||
	    (p0 - p3) % (p2 - p3) <= 0.0 ) ||
    ( cosCornerAngleSquared(0) == 1 ||
      cosCornerAngleSquared(1) == 1 ||
      cosCornerAngleSquared(2) == 1 ||
      cosCornerAngleSquared(3) == 1 );
}

const std::vector<ICoord> *
CSkeletonQuad::underlying_pixels(const CMicrostructure &microstructure) const {
  // Divide the quad into two triangles with positive area, and mark
  // the pixels under them.  This avoids the possibly complicated
  // geometry of quadrilaterals.
  const Coord c0 = nodes[0]->position();
  const Coord c1 = nodes[1]->position();
  const Coord c2 = nodes[2]->position();
  const Coord c3 = nodes[3]->position();
  CRectangle bbox(c0, c2);
  bbox.swallow(c1);
  bbox.swallow(c3);
  MarkInfo *mm = microstructure.beginMarking(bbox);
  if(triangleArea(c0, c1, c2) >= 0.0 &&
     triangleArea(c0, c2, c3) >= 0.0) {
    microstructure.markTriangle(mm, c0, c1, c2);
    microstructure.markTriangle(mm, c0, c2, c3);
  }
  else if(triangleArea(c1, c2, c3) >= 0.0 &&
          triangleArea(c2, c3, c0) >= 0.0) {
    microstructure.markTriangle(mm, c1, c2, c3);
    microstructure.markTriangle(mm, c1, c3, c0);
  }
  std::vector<ICoord> *pxls = microstructure.markedPixels(mm); // new'd vector
  microstructure.endMarking(mm);
  return pxls;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

CSkeletonTriangle::CSkeletonTriangle(CSkeletonNode *n0, CSkeletonNode *n1,
				     CSkeletonNode *n2)
  : CSkeletonElement(3)
{
  nodes[0] = n0;
  nodes[1] = n1;
  nodes[2] = n2;
}

double CSkeletonTriangle::shapefun(int i, const MasterCoord &mc) const {
  switch(i) {
  case 0:
    return mc(0);
  case 1:
    return mc(1);
  case 2:
    return 1. - mc(0) - mc(1);
  }
  throw ErrBadIndex(i, __FILE__, __LINE__);
}

bool CSkeletonTriangle::interior(const Coord *point) const {
  return insideTriangle(nodes[0]->position(), nodes[1]->position(),
			nodes[2]->position(), *point);
};

double CSkeletonTriangle::area() const {
  return triangleArea(nodes[0]->position(), nodes[1]->position(),
		      nodes[2]->position());
}

bool CSkeletonTriangle::illegal() const {
  //   // Using Area() basically, but more thoroughly.
  //   const Coord &p1 = nodes[0]->position();
  //   const Coord &p2 = nodes[1]->position();
  //   const Coord &p3 = nodes[2]->position();

  //   double a1 = triangleArea(p1,p2,p3);
  //   if (a1<=0.0)
  //     return true;
  //   double a2 = triangleArea(p2,p3,p1);
  //   if (a2<=0.0)
  //     return true;
  //   double a3 = triangleArea(p3,p1,p2);
  //   if (a3<=0.0)
  //     return true;
    
  //   if (fabs((a1/a2)-1.0) > LEGAL_ELEMENT_AREA_TOLERANCE ||
  //       fabs((a2/a3)-1.0) > LEGAL_ELEMENT_AREA_TOLERANCE ||
  //       fabs((a3/a1)-1.0) > LEGAL_ELEMENT_AREA_TOLERANCE )
  //     return true;

  //   return false;

  // Using Area() & Cosine() only.
  return (area() <= 0.0 ||
	  cosCornerAngleSquared(0) == 1 ||
	  cosCornerAngleSquared(1) == 1 ||
	  cosCornerAngleSquared(2) == 1);
}

const std::vector<ICoord> *
CSkeletonTriangle::underlying_pixels(const CMicrostructure &microstructure)
  const
{
  Coord p0 = nodes[0]->position();
  Coord p1 = nodes[1]->position();
  Coord p2 = nodes[2]->position();
  CRectangle bbox(p0, p1);
  bbox.swallow(p2);
  MarkInfo *mm = microstructure.beginMarking(bbox);
  microstructure.markTriangle(mm, p0, p1, p2);
  std::vector<ICoord> *pxls = microstructure.markedPixels(mm); // new'd vector
  microstructure.endMarking(mm);
  return pxls;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// OLD VERSION USING MAP JACOBIAN

// Shape energy calculation.  The shape energy is (c-1)/c, where c 
// is the ratio of the condition number squared of the jacobian of 
// the ideal->physical space map, to the condition number squared of
// the jacobian of the identity map.  For quads, the maximum condition
// number over all corners of the element is used.
//   The energy is 0 for squares and equilateral triangles, and 1
// for degenerate shapes (ones with 180-degree vertices).

// double CSkeletonTriangle::energyShape() const {
//   Coord a_vec = nodes[1]->position()-nodes[0]->position();
//   Coord b_vec = nodes[2]->position()-nodes[0]->position();

//   double num = norm2(a_vec) + norm2(b_vec) - dot(a_vec, b_vec);
//   double den = cross(a_vec, b_vec);
//   double condtn = num*num/(3.0*den*den);
//   return (condtn-1.0)/condtn;
// }

// double CSkeletonQuad::energyShape() const {
//   double eshape=0.0;
//   for(int i=0;i<4;i++) {
//     Coord a_vec = nodes[(i+1)%4]->position() - nodes[i]->position();
//     Coord b_vec = nodes[(i+3)%4]->position() - nodes[i]->position();
//     double num = norm2(a_vec)+norm2(b_vec);
//     double den = cross(a_vec, b_vec);
//     double condtn = 0.25*num*num/(den*den);
//     double corner_eshape = (condtn-1.0)/condtn;
//     if (corner_eshape > eshape)
//       eshape = corner_eshape;
//   }
//   return pow(eshape, 1.5);
// }

// // Testing new shape energy functions using STANDARD DEVIATION
// double CSkeletonTriangle::energyShape() const {
//   const double pi = acos(-1.0);
//   const double a0 = acos(cosCornerAngle(0))*180.0/pi;
//   const double a1 = acos(cosCornerAngle(1))*180.0/pi;
//   const double a2 = 180.0 - (a0+a1);
//   const double meanA = 60.0;
//   const double normA = 1.4142135623731;
//   const double angleE = sqrt(((meanA-a0)*(meanA-a0)+
// 			      (meanA-a1)*(meanA-a1)+
// 			      (meanA-a2)*(meanA-a2))/3.0)/meanA/normA;

//   const double l0 = edgeLength(0);
//   const double l1 = edgeLength(1);
//   const double l2 = edgeLength(2);
//   const double meanL = (l0+l1+l2)/3.0;
//   const double normL = 0.7071067811866;
//   const double aspectE = sqrt(((meanL-l0)*(meanL-l0)+
// 			       (meanL-l1)*(meanL-l1)+
// 			       (meanL-l2)*(meanL-l2))/3.0)/meanL/normL;

//   if (angleE >= aspectE)
//     return angleE;
//   else
//     return aspectE;
// //   const double beta = 0.5;
// //   return beta*angleE + (1.0-beta)*aspectE;
// }

// double CSkeletonQuad::energyShape() const {
//   const double pi = acos(-1.0);
//   const double a0 = acos(cosCornerAngle(0))*180.0/pi;
//   const double a1 = acos(cosCornerAngle(1))*180.0/pi;
//   const double a2 = acos(cosCornerAngle(2))*180.0/pi;
//   const double a3 = 360.0 - (a0+a1+a2);
//   const double meanA = 90.0;
//   const double angleE = sqrt(((meanA-a0)*(meanA-a0)+
// 			      (meanA-a1)*(meanA-a1)+
// 			      (meanA-a2)*(meanA-a2)+
// 			      (meanA-a3)*(meanA-a3))/4.0)/meanA;

//   const double l0 = edgeLength(0);
//   const double l1 = edgeLength(1);
//   const double l2 = edgeLength(2);
//   const double l3 = edgeLength(3);
//   const double meanL = (l0+l1+l2+l3)/4.0;
//   const double aspectE = sqrt(((meanL-l0)*(meanL-l0)+
// 			      (meanL-l1)*(meanL-l1)+
// 			      (meanL-l2)*(meanL-l2)+
// 			      (meanL-l3)*(meanL-l3))/4.0)/meanL;

//   if (angleE >= aspectE)
//     return angleE;
//   else
//     return aspectE;
// }

// END OF OLD VERSION USING MAP JACOBIAN



// SHAPE ENERGY USING MIN. & MAX. ANGLE and MAX. EDGE RATIO.

// NB dependent max2, max3, max4 (and min) functions removed.

// double CSkeletonTriangle::energyShape() const
// {
//   const double a0 = getRealAngle(0);
//   const double a1 = getRealAngle(1);
//   const double a2 = 180.0 - (a0+a1);
//   const double l0 = edgeLength(0);
//   const double l1 = edgeLength(1);
//   const double l2 = edgeLength(2);

//   const double min_angle_energy = (60.0-min3(a0, a1, a2))/60.0;
//   const double max_angle_energy = (max3(a0, a1, a2)-60.0)/120.0;
//   const double angle_energy = max2(min_angle_energy, max_angle_energy);

//   const double min_length = min3(l0, l1, l2);
//   const double max_length = max3(l0, l1, l2);
//   const double aspect_energy = 1.0 - min_length/max_length;

//   return max2(angle_energy, aspect_energy*aspect_energy);
// }

// double CSkeletonQuad::energyShape() const
// {  
//   const double a0 = getRealAngle(0);
//   const double a1 = getRealAngle(1);
//   const double a2 = getRealAngle(2);
//   const double a3 = 360.0 - (a0+a1+a2);
//   const double l0 = edgeLength(0);
//   const double l1 = edgeLength(1);
//   const double l2 = edgeLength(2);
//   const double l3 = edgeLength(3);

//   const double min_angle_energy = (90.0-min4(a0, a1, a2, a3))/90.0;
//   const double max_angle_energy = (max4(a0, a1, a2, a3)-90.0)/90.0;
//   const double angle_energy = max2(min_angle_energy, max_angle_energy);

//   const double min_length = min4(l0, l1, l2, l3);
//   const double max_length = max4(l0, l1, l2, l3);
//   const double aspect_energy = 1.0 - min_length/max_length;

//   return max2(angle_energy, aspect_energy*aspect_energy);
// }

// SHAPE ENERGY USING TRIANGLE ALTITUDES
// These should be faster to compute.

// double triangleQ(const Coord &n0, const Coord &n1, const Coord &n2) {
//   // The Q of a triangle is the minimum altitude, squared, over the
//   // maximum side, squared.  The altitude from vertex n to side n is
//   // the length of one of the other sides times the sine of the angle
//   // between that side and side n.  That is,
//   //        h  =  side(n+1) cross side(n)/|side(n)|

//   // Sides
//   Coord s0 = n1 - n2;
//   Coord s1 = n2 - n0;
//   Coord s2 = n0 - n1;

//   double c = cross(s0, s1);
//   double ll0 = norm2(s0);
//   if(ll0 == 0.0) return 0.0;
//   double hh0 = c*c/ll0;

//   c = cross(s1, s2);
//   double ll1 = norm2(s1);
//   if(ll1 == 0.0) return 0.0;
//   double hh1 = c*c/ll1;

//   c = cross(s2, s0);
//   double ll2 = norm2(s2);
//   if(ll2 == 0.0) return 0.0;
//   double hh2 = c*c/ll2;

//   return min3(hh0, hh1, hh2)/max3(ll0, ll1, ll2);
// }

// double CSkeletonTriangle::energyShape() const {
//   The factor of 4/3 normalizes Q so that it's 1 for an equilateral
//   triangle.
//   static const double fourthirds = 4./3.;
//   double q = fourthirds*triangleQ(nodes[0]->position(), nodes[1]->position(),
// 				  nodes[2]->position());
//   double e = 1.-q;
//   return e*e*e;
// }

// double CSkeletonQuad::energyShape() const {
//   A quad can be divided into two triangles in two ways.  The Q of a
//   pair of triangles is the *geometric* mean of their individual Qs.
//   We use the geometric mean so that the Q of a pair is 0 if the Q
//   of either member of the pair is zero.  (The triangle Q is
//   normalized so that it's 1 for a right isosceles triangle.)  The Q
//   of the quad is Q of the pair of triangles with the smaller Q.
//   The energy of the quad is 1-Q.

//   double q0 = triangleQ(nodes[1]->position(), nodes[2]->position(),
// 			nodes[3]->position());
//   double q1 = triangleQ(nodes[2]->position(), nodes[3]->position(),
// 			nodes[0]->position());
//   double q2 = triangleQ(nodes[3]->position(), nodes[0]->position(),
// 			nodes[1]->position());
//   double q3 = triangleQ(nodes[0]->position(), nodes[1]->position(),
// 			nodes[2]->position());
//   double q02 = q0*q2;
//   double q13 = q1*q3;
//   double q = 4*sqrt(q02 < q13 ? q02 : q13); factor of 4 is normalization
//   double e = 1.-q;
//   if(e < 0.0) return 0.0;
//   return sqrt(e*e*e);
// }

double CSkeletonTriangle::energyShape() const
{
  // Old OOF way: 4sqrt(3)*AREA/(L0^2 + L1^2 + L2^2),
  // where L0, L1, L2 represent three edges of the triangle.
  // Equilateral triangle: 1
  static const double four_sqrt_three = 4.*sqrt(3.);
  double q = four_sqrt_three*area()/perimeterSquared();
  return 1. - q;
}

double paralleloQ(const Coord &n0, const Coord &n1, const Coord &n2)
{
  // For the case of a square, l0=l1, and area2 = 2*l0*l1 = 2*l0*l0.
  // In the same (square) case, the denominator is l0*l0+l1*l1 =
  // 2*l0*l0, so a square of any size returns 1 from this function.  A
  // bit of examination reveals that any deviation from squareness
  // (either because l1 != l2, or because the angle is not 90 degrees)
  // reduces this value, so this is maximal for squares.
  double area2 = 4.*triangleArea(n0, n1, n2);
  double l0 = norm2(Coord(n0 - n1));
  double l1 = norm2(Coord(n1 - n2));
  return area2/(l0 + l1);
}

#define OPPOSITE_NODE_WEIGHT 1.0e-5

double CSkeletonQuad::energyShape() const
{
  // Copied from GiD (http://gid.cimne.upc.es) manual.
  // GiD is a pre- & post process tool developed in Spain.
  // Take the worst of the values computed at four corners.
  // q = 2*AREA/(L0^2 + L1^2),
  // where AREA is area of parallelogram formed with two edges
  // of the corner node and L0 and L2 represent the two edges.
  double q0 = paralleloQ(nodes[3]->position(), nodes[0]->position(),
			 nodes[1]->position());
  double q1 = paralleloQ(nodes[0]->position(), nodes[1]->position(),
			 nodes[2]->position());
  double q2 = paralleloQ(nodes[1]->position(), nodes[2]->position(),
			 nodes[3]->position());
  double q3 = paralleloQ(nodes[2]->position(), nodes[3]->position(),
			 nodes[0]->position());

  double minq = q0;  // Minimum (most-bad) paralleloQ result
  double oppq = q2;  // paralleloQ from node opposite to minimum.
  if (q1<minq) {
    minq = q1; 
    oppq = q3;
  }
  if (q2<minq) {
    minq = q2; 
    oppq = q0;
  }
  if (q3<minq) {
    minq = q3;
    oppq = q1;
  }

  // Return the weighted sum of the worst case seen and the opposite
  // node.  The reason for this is to make sure that any node movement
  // changes the energy, even if it doesn't shift the minimum.  The
  // reason for using a lopsided weight and not just averaging is that
  // we would still like degenerate sub-triangles of quads to give
  // (nearly) the worst energy possible.  Straight averaging would
  // dilute that.
  return 1.0-((1.0-OPPOSITE_NODE_WEIGHT)*minq + OPPOSITE_NODE_WEIGHT*oppq);
  //return 1. - minq;
}

#elif DIM==3

// CSkeletonBrick::CSkeletonBrick(CSkeletonNode *n0, CSkeletonNode *n1,
// 			       CSkeletonNode *n2, CSkeletonNode *n3,
// 			       CSkeletonNode *n4, CSkeletonNode *n5,
// 			       CSkeletonNode *n6, CSkeletonNode *n7)
//   : CSkeletonElement(8)
// {
//   nodes[0] = n0;
//   nodes[1] = n1;
//   nodes[2] = n2;
//   nodes[3] = n3;
//   nodes[4] = n4;
//   nodes[5] = n5;
//   nodes[6] = n6;
//   nodes[7] = n7;

// }

// // TODO 3D: FILL THESE FUNCTIONS IN
// double CSkeletonBrick::shapefun(int i, const MasterCoord &mc) const {
//   return 1.0;
// }

// bool CSkeletonBrick::interior(const Coord *point) const {
//   // use EvaluatePosition function here, if we can figure out what the
//   // hell all the inputs are.  is this used only by selection?  in
//   // that case, we'll just use the picker and don't need this
//   // function.
//   return 1;
// }

// double CSkeletonBrick::area() const {
//   return 1.0;
// }

double tetrahedraVolume(const Coord  &v0, const Coord &v1, const Coord &v2, const Coord &v3) {
  return (1.0/6.0) *  dot( (v3-v0) , ((v3-v1)%(v3-v2)) ) ;
}

// double CSkeletonBrick::volume() const {
//   // Find the total volume of 5 sub-tetrahedra
//   return 
//     tetrahedraVolume(nodes[0]->position(), nodes[1]->position(), 
// 		     nodes[2]->position(), nodes[5]->position()) +
//     tetrahedraVolume(nodes[2]->position(), nodes[5]->position(), 
// 		     nodes[6]->position(), nodes[7]->position()) +
//     tetrahedraVolume(nodes[0]->position(), nodes[2]->position(), 
// 		     nodes[3]->position(), nodes[7]->position()) +
//     tetrahedraVolume(nodes[0]->position(), nodes[4]->position(), 
// 		     nodes[5]->position(), nodes[7]->position()) +
//     tetrahedraVolume(nodes[0]->position(), nodes[2]->position(), 
// 		     nodes[5]->position(), nodes[7]->position());
// }

// bool CSkeletonBrick::illegal() const {
//   // the four vertices of each face must lie in one plane
//   // and form a legal quad ... or do they?
//   return 0;
// }

// const std::vector<ICoord> *
// CSkeletonBrick::underlying_pixels(const CMicrostructure &microstructure) const {
//   // copied from triangle for now, just to return correct type
//   Coord p0 = nodes[0]->position();
//   Coord p1 = nodes[1]->position();
//   Coord p2 = nodes[2]->position();
//   CRectangle bbox(p0, p1);
//   MarkInfo *mm = microstructure.beginMarking(bbox);
//   std::vector<ICoord> *pxls = microstructure.markedPixels(mm); // new'd vector
//   microstructure.endMarking(mm);
//   return pxls;
// }

// double CSkeletonBrick::energyShape() const
// {
//   return 0.0;
// }


// Tetra


CSkeletonTetra::CSkeletonTetra(CSkeletonNode *n0, CSkeletonNode *n1,
			       CSkeletonNode *n2, CSkeletonNode *n3)
  : CSkeletonElement(4)
{

  // TODO 3D: Get rid of this redundancy...
  nodes[0] = n0;
  nodes[1] = n1;
  nodes[2] = n2;
  nodes[3] = n3;

  //cout << "in CSkeletonTetra" << endl;

  cell = vtkTetra::New();
  vtkPoints *p = nodes[0]->getPoints();
  for(int i = 0; i < 4; ++i) {
    cell->GetPointIds()->SetId(i,nodes[i]->index());
    cell->GetPoints()->InsertPoint(i,p->GetPoint(nodes[i]->index()));
  //  cout << nodes[i]->index() << endl;
  }
}


// TODO 3D: FILL THESE FUNCTIONS IN
double CSkeletonTetra::shapefun(int i, const MasterCoord &mc) const {
  return 1.0;
}

bool CSkeletonTetra::interior(const Coord *point) const {
  // use EvaluatePosition function here, if we can figure out what the
  // hell all the inputs are.  is this used only by selection?  in
  // that case, we'll just use the picker and don't need this
  // function.
  return 1;
}

double CSkeletonTetra::area() const {
  return 1.0;
}



double CSkeletonTetra::volume() const {
  return 
    tetrahedraVolume(nodes[0]->position(), nodes[1]->position(), 
  		     nodes[2]->position(), nodes[3]->position());

  //vtkPoints* points = cell->GetPoints();
  //double p[4][3];
  //for(int i=0; i<4; i++)
  //  points->GetPoint(i,p[i]);
  //return vtkTetra::ComputeVolume(p[0],p[1],p[2],p[3]);
}

bool CSkeletonTetra::illegal() const {
  // vtkTetras are defined such that the normal of the fourth face
  // (defined by points (2,1,0)) is directed away from the fourth
  // point.
//   if(volume() <= 0)
//     return true;

//   double n[3],x[3],pts[4][3];
//   int i;

//   for(i=0; i<4; i++) 
//     this->getPoints()->GetPoint(i,pts[i]);
//   vtkTriangle::ComputeNormal(pts[2],pts[1],pts[0],n);
//   for(i=0;i<3;i++)
//     x[i]=pts[3][i]-pts[0][i];
//   vtkMath::Normalize(x);

  //#ifdef DEBUG
//   float dot = vtkMath::Dot(n,x);
//   if(dot >= 0.0 || volume() <=0 ) {
//     cout << "in illegal" << endl;
//     for(i=0; i<4; i++) 
//       cout << pts[i][0] << " " << pts[i][1] << " " << pts[i][2] << endl;
//     cout << n[0] <<  " " << n[1] << " " << n[2] << endl;  
//     cout << x[0] <<  " " << x[1] << " " << x[2] << endl; 
//     cout << dot << " " << volume() << endl;
//   }
  //#endif

//   return ( vtkMath::Dot(n,x) >= 0.0); // || volume() <= 0 );

  if(volume() <= 0)
    return true;

  for(int i = 0; i<4; ++i) {
    for(int j = 0; j<3; ++j) {
      if(cosCornerAngleSquared(i,j) == 1)
	return true;
    }
  }
  return false;

}

const std::vector<ICoord> *
CSkeletonTetra::underlying_pixels(const CMicrostructure &microstructure) const {
  // copied from triangle for now, just to return correct type
  Coord p0 = nodes[0]->position();
  Coord p1 = nodes[1]->position();
  Coord p2 = nodes[2]->position();
  CRectangle bbox(p0, p1);
  MarkInfo *mm = microstructure.beginMarking(bbox);
  std::vector<ICoord> *pxls = microstructure.markedPixels(mm); // new'd vector
  microstructure.endMarking(mm);
  return pxls;
}

const double regularTetrahedraFactor = (1.5)*(pow(27,0.25)/sqrt(2));

double CSkeletonTetra::energyShape() const
{
  Coord p0 = nodes[0]->position();
  Coord p1 = nodes[1]->position();
  Coord p2 = nodes[2]->position();
  Coord p3 = nodes[3]->position();
  double denominator = pow( 0.25*(triangleArea2(p0,p2,p1) +
				  triangleArea2(p2,p3,p1) + 
				  triangleArea2(p0,p1,p3) + 
				  triangleArea2(p3,p2,p0)),
			    0.75);
	
  return 1 - regularTetrahedraFactor*tetrahedraVolume(p0,p1,p2,p3)/denominator;
}




#endif

// Homogeneity computations.
// The HomogeneityData constructor is the One True Place where the
// homogeneity energy is computed.
HomogeneityData::HomogeneityData(double hom, int cat) : homogeneity(hom),
							dominantpixel(cat) {
  homog_energy = 1.0-homogeneity;
							}



