# -*- python -*-
# $RCSfile: nonlinearsolver.py,v $
# $Revision: 1.40 $
# $Author: langer $
# $Date: 2010-11-15 22:02:46 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@nist.gov. 

# Nonlinear solvers for use by subproblems.  These are derived from
# the classes in nonlinearsolvercore.py, which do the actual work, but
# are agnostic about subproblems.

import sys

from ooflib.SWIG.engine import ooferror2
from ooflib.common import debug
from ooflib.common import registeredclass
from ooflib.common.IO import parameter
from ooflib.engine import nonlinearsolvercore


class NonlinearSolverBase(registeredclass.RegisteredClass):
    registry = []

#=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=#

# NoNonlinearSolver is actually a linear solver.  The order of the
# base classes is important to prevent swig 1.1 from complaining
# about the type of the "self" arg to CNonlinearSolver::nonlinear.

class NoNonlinearSolver(nonlinearsolvercore.NoNLSolver, NonlinearSolverBase):
    def step(self, subproblem, *args, **kwargs):
        nlprops = not subproblem.getObject().properties_provide_K()
        explicitstep = subproblem.time_stepper.explicit()
        if nlprops:
            if explicitstep:
                self.requireResidual(True) # Sets flags in CNonlinearSolver.
                self.requireJacobian(False) # Ditto.
                ## In this case "nonlinearstep" really means
                ## "linearstep by evaluating the static residual
                ## instead of K*u+f."  No nonlinear solver will be
                ## used.
                v = subproblem.time_stepper.nonlinearstep(subproblem,
                                                          nonlinearMethod=None,
                                                          *args, **kwargs)
            else:
                raise ooferror2.ErrSetupError(
                    "Tried to use a linear solver on a nonlinear "
                    "problem with an implicit stepper.")
        else:                           # properties provide K
            self.requireResidual(False) # Sets flags in CNonlinearSolver.
            self.requireJacobian(False) # Ditto.
            v = subproblem.time_stepper.linearstep(subproblem, *args, **kwargs)
        return v
            
    def computeStaticFields(self, subprobctxt, linsys, unknowns):
        # Called by SubProblemContext.initializeStaticFields.
        return subprobctxt.computeStaticFieldsL(linsys, unknowns)
    def shortrepr(self):
        return "None"
    def __repr__(self):
        return registeredclass.RegisteredClass.__repr__(self)

registeredclass.Registration(
    'None',
    NonlinearSolverBase,
    NoNonlinearSolver,
    ordering = 0,
    tip = "Use linear equation solvers.")

#=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=#

class NonlinearSolver(NonlinearSolverBase):
    def step(self, subprob, *args, **kwargs):
        # if (subproblem.getObject().properties_provide_K()
        #     and subprob.time_stepper.explicit()):
        #     return subprob.time_stepper.linearstep(subprob, *args, **kwargs)
        return subprob.time_stepper.nonlinearstep(
            subprob, nonlinearMethod=self, *args, **kwargs)
    def computeStaticFields(self, subprobctxt, linsys, unknowns):
        # Called by SubProblemContext.initializeStaticFields.
        return subprobctxt.computeStaticFieldsNL(linsys, unknowns)

# Shared parameters for the nontrivial solvers.

nonlin_params = [
    parameter.FloatParameter(
        'relative_tolerance', 1.e-8,
        tip="Relative tolerance for convergence to a nonlinear solution."),
    parameter.FloatParameter(
        'absolute_tolerance', 1.e-13,
        tip="Absolute tolerance for convergence to a nonlinear solution."),
    parameter.IntParameter(
        'maximum_iterations', 200,
        tip="Maximum number of iterations for convergence to a nonlinear solution.")
    ]


#=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=#

class Newton(nonlinearsolvercore.Newton, NonlinearSolver):
    def __init__(self, *args, **kwargs):
        nonlinearsolvercore.Newton.__init__(self, *args, **kwargs)
    def solve(self, *args, **kwargs):
        niters, residual = nonlinearsolvercore.Newton.solve(
            self, *args, **kwargs)
        self.subproblem.solverStats.nonlinearSolution(niters, residual)
    def __repr__(self):
        return registeredclass.RegisteredClass.__repr__(self)

registeredclass.Registration(
    'Newton',
    NonlinearSolverBase,
    Newton,
    ordering = 1,
    params   = nonlin_params,
    tip      = "Solve nonlinear equations with Newton's method."
    )

#=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=#

class Picard(nonlinearsolvercore.Picard, NonlinearSolver):
    def __init__(self, *args, **kwargs):
        nonlinearsolvercore.Picard.__init__(self, *args, **kwargs)
    def solve(self, *args, **kwargs):
        niters, residual = nonlinearsolvercore.Picard.solve(
            self, *args, **kwargs)
        self.subproblem.solverStats.nonlinearSolution(niters, residual)
    def __repr__(self):
        return registeredclass.RegisteredClass.__repr__(self)

registeredclass.Registration(
    'Picard',
    NonlinearSolverBase,
    Picard,
    ordering = 2,
    params   = nonlin_params,
    tip      = "Solve nonlinear equations with Picard iteration."
    )


#=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=#

# The nonlinear equations can have the following forms, where "derivs"
# means M*d2u/dt2 + C*du/dt:
#
# (A) derivs + grad (K grad u) + f(u) = 0
# (B) derivs + grad (sigma(u)) + g = 0
# (C) derivs + grad (sigma(u)) + f(u) = 0
# (L) derivs + grad (K grad u) + g = 0
#
# That is,
# (A) <==> FluxProperties are linear, EqnProperties are nonlinear
# (B) <==> FluxProperties are nonlinear, EqnProperties are linear
# (C) <==> FluxProperties and EqnProperties are nonlinear
# (L) <==> FluxProperties and EqnProperties are linear.
#
# f and g may be time dependent in all cases, but don't depend on u in
# case B.
#
# To solve with Newton, we need the following quantities for each type
# of equation:
# (A) K, f, Df
# (B) sigma, Dsigma, f 
# (C) sigma, Dsigma, f, Df
#
# To solve with Picard, we need:
# (A) K, f
# (B) sigma, Dsigma, f
# (C) sigma, Dsigma, f
#
# [To solve with the (currently nonexistent) super Picard solver, we'd
# need:
# (A) K, f
# (B) sigma, f
# (C) sigma, f
# ie, nothing except the terms that appear explicitly in the
# equations.]
#
# sigma and f can always be computed from the Properties.  It's
# possible that some nonlinear Properties can't compute K, Dsigma, or
# Df.

def allowedSolvers(subproblem):
    properties = []
    for mat in subproblem.all_materials():
        properties.extend(mat.properties())
    fields = subproblem.all_active_fields()
    nlregs = [p.registration() for p in properties if p.is_nonlinear(fields)]
    fluxproperties = [p for p in nlregs if p.propertyClass == "FluxProperty"]
    eqnproperties = [p for p in nlregs if p.propertyClass == "EqnProperty"]

    explicitstep = subproblem.time_stepper.explicit()

    # flux is nonlinear if len(fluxproperties) > 0
    # direct eqn terms are nonlinear if len(eqnproperties) > 0

    # Case L.  Everything is linear, but we can still use nonlinear
    # solvers if the user insists.  On the other hand, even if the
    # problem is nonlinear, we don't need a nonlinear solver if the
    # stepper is explicit.  On the other other hand, that's not true
    # if there are nonlinear static components to the problem, but we
    # can't test for that until after the matrices have been
    # constructed.
    if (not fluxproperties and not eqnproperties) or explicitstep:
        return (NoNonlinearSolver, Newton, Picard)

    # Case A.
    elif eqnproperties and not fluxproperties:
        if all(reg.redefines("force_deriv_matrix") for reg in eqnproperties):
            return (Newton, Picard)
        return (Picard,)

    # Case B.
    elif fluxproperties and not eqnproperties:
        if all(reg.redefines("flux_matrix") for reg in fluxproperties):
            return (Newton, Picard)
        return ()

    # Case C.
    elif fluxproperties and eqnproperties:
        if all(reg.redefines("flux_matrix") for reg in fluxproperties):
            if all(reg.redefines("force_deriv_matrix") for reg in eqnproperties):
                return (Newton, Picard)
            return (Picard,)

    return ()

        
