# -*- python -*-
# $RCSfile: problem.py,v $
# $Revision: 1.70 $
# $Author: reida $
# $Date: 2010-11-03 19:25:34 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# TODO 3D: we can get rid of the separation between in plane and out of plane.

from ooflib.SWIG.common import config
from ooflib.SWIG.common import switchboard
from ooflib.SWIG.common import ooferror
from ooflib.SWIG.common import config
from ooflib.engine import conjugate
from ooflib.SWIG.engine import equation
from ooflib.SWIG.engine import field
from ooflib.SWIG.engine import fieldindex
from ooflib.SWIG.engine import flux
from ooflib.common import debug
from ooflib.common import utils
from ooflib.common.IO import xmlmenudump
import sys
import types

def _advertise(obj):
    utils.OOFdefine(obj.name(), obj)
    return obj

def advertiseField(fld):
    field.newCompoundField(fld)
    _advertise(fld)
    td = _advertise(fld.time_derivative())
    field.newField(td)
    if config.dimension() == 2:
        field.newField(_advertise(fld.out_of_plane()))
        field.newField(_advertise(fld.out_of_plane_time_derivative()))
    # "new field" is sent here, instead of from the Field constructor,
    # because it must be called *after* the field is defined in the
    # OOF namespace.
    switchboard.notify("new field")
    return fld

def advertiseFlux(flx):
    _advertise(flx)
    switchboard.notify("new flux")
    return flx

def advertiseEquation(eqn):
    _advertise(eqn)
    switchboard.notify("new equation")
    return eqn

def advertise(obj):
    # This code is ugly, but at least it's compact.  It's not much
    # uglier than the previous version, which required other code to
    # call advertiseField, et al, directly.
    if isinstance(obj, field.FieldPtr):
        return advertiseField(obj)
    if isinstance(obj, flux.FluxPtr):
        return advertiseFlux(obj)
    if isinstance(obj, equation.EquationPtr):
        return advertiseEquation(obj)
    raise ooferror.ErrPyProgrammingError("Don't know what to do with %s!"% obj)

#=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=#

# Define a field.  This creates an object named 'Temperature' in the
# OOF namespace.
Temperature = advertise(field.ScalarField('Temperature'))
# Define a flux
Heat_Flux = advertise(flux.VectorFlux('Heat_Flux'))
# And equations
HeatBalanceEquation = advertise(equation.DivergenceEquation(
    'Heat_Eqn',
    Heat_Flux,
    1
    ))

if config.dimension() == 2:
    HeatOutOfPlane = advertise(equation.PlaneFluxEquation(
            'Plane_Heat_Flux', Heat_Flux, 1))

## this creates the Displacement, Stress, and Mechanical Equilibrium equations
if config.dimension() == 2:
    Displacement = advertise(field.TwoVectorField('Displacement'))
elif config.dimension() == 3:
    Displacement = advertise(field.ThreeVectorField('Displacement'))
Stress = advertise(flux.SymmetricTensorFlux('Stress'))

ForceBalanceEquation = advertise(equation.DivergenceEquation(
    'Force_Balance',
    Stress,
    config.dimension()
    ))

if config.dimension() == 2:
    ForcesOutOfPlane = \
        advertise(equation.PlaneFluxEquation('Plane_Stress',
                                             Stress, 3))


## Define electrostatic potential
Voltage = advertise(field.ScalarField('Voltage'))
## Define total polarization vector
Total_Polarization = advertise(flux.VectorFlux('Total_Polarization'))
## Differential form of Coulomb's Law
CoulombEquation = advertise(equation.DivergenceEquation(
    'Coulomb_Eqn', Total_Polarization, 1))
if config.dimension() == 2:
    PolarizationOutOfPlane = advertise(equation.PlaneFluxEquation(
            'InPlanePolarization', Total_Polarization, 1))


# Plasticity -- start with the yield equation.
# TODO: Make hidden equations.
YieldEquation = advertise(equation.NaturalEquation(
    'Yield_Eqn',1))


#=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=#

## Define conjugate quantities for all fields and equations

## Force balance equation
##
## In-plane components (u,v)
##
## The 3D displacement vector is (u, v, w),
## where
## u lies along the x-direction,
## v, along the y-direction,
## and w, out-of-the-plane, or along
## the z-direction
u = fieldindex.VectorFieldIndex(0)
v = fieldindex.VectorFieldIndex(1)

fx = fieldindex.VectorFieldIndex(0)
fy = fieldindex.VectorFieldIndex(1)

if config.dimension() == 2:
    # fx is conjugate to u and fy is conjugate to v for Elasticity
    conjugate.conjugatePair("Elasticity", ForceBalanceEquation, [fx, fy],
                        Displacement, [u, v]) 

    ## out-of-plane compoments

    ## The available out-of-plane components of stress are $\sigma_{zz},
    ## \sigma_{zy}, \sigma_{zx}$, in *that* order, (0, 1, 2).  The
    ## out-of-plane displacement is (\frac{\partial u}{\partial z},
    ## \frac{\partial v}{partial z}, \frac{\partial w}{\partial z}.

    u_xz = fieldindex.VectorFieldIndex(0)
    u_yz = fieldindex.VectorFieldIndex(1)
    u_zz = fieldindex.VectorFieldIndex(2)
    sigma_xz = fieldindex.OutOfPlaneSymTensorIndex(2,0)
    sigma_yz = fieldindex.OutOfPlaneSymTensorIndex(2,1)
    sigma_zz = fieldindex.OutOfPlaneSymTensorIndex(2,2)

    ## \sigma_{zz} is conjugate to \frac{\partial w}{\partial z}
    ## \sigma_{xz} is conjugate to \frac{\partial u}{\partial z}
    ## \sigma_{yz} is conjugate to \frac{\partial v}{\partial z}

    conjugate.conjugatePair("Elasticity",
                            ForcesOutOfPlane, [sigma_zz, sigma_xz, sigma_yz],
                            Displacement.out_of_plane(), [u_zz, u_xz, u_yz])

elif config.dimension() == 3:
    w = fieldindex.VectorFieldIndex(2)
    fz = fieldindex.VectorFieldIndex(2)
    conjugate.conjugatePair("Elasticity", ForceBalanceEquation, [fx, fy, fz],
                            Displacement, [u, v, w])

###############################################################
##
## Heat flux equation
##
## In-plane components, T
##
T = fieldindex.ScalarFieldIndex()
DivJ = fieldindex.ScalarFieldIndex()

conjugate.conjugatePair("ThermalConductivity", HeatBalanceEquation, DivJ,
                        Temperature, T)
 ## $\nabla \cdot \vec{J}$ is conjugate to T

## out-of-plane components, $\frac{\partial T}{\partial z}$
if config.dimension() == 2:
    T_z = fieldindex.OutOfPlaneVectorFieldIndex(2)
    J_z = fieldindex.OutOfPlaneVectorFieldIndex(2)

    conjugate.conjugatePair("ThermalConductivity", HeatOutOfPlane, J_z,
                            Temperature.out_of_plane(), T_z)
 ##  $J_{z}$ is conjugate to $\frac{\partial T}{\partial z}$

###############################################################
##
## Coulomb equation
##
## In-plane components, phi
##
phi = fieldindex.ScalarFieldIndex()
DivD = fieldindex.ScalarFieldIndex()

conjugate.conjugatePair("DielectricPermittivity",
                        CoulombEquation, DivD,
                        Voltage, phi)
 ## $\nabla \cdot \vec{D}$ is conjugate to D

## out-of-plane components, $\frac{\partial D}{\partial z}$
if config.dimension() == 2:
    phi_z = fieldindex.OutOfPlaneVectorFieldIndex(2)
    D_z = fieldindex.OutOfPlaneVectorFieldIndex(2)

    conjugate.conjugatePair("DielectricPermittivity",
                            PolarizationOutOfPlane, D_z,
                            Voltage.out_of_plane(), phi_z)
 ##  $D_{z}$ is conjugate to $\frac{\partial phi}{\partial z}$



#=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=##=--=#

# xmldump generates the manual page for the built-in physics.

def xmldump(file):
    print >> file, "<section id='Section-Builtin'>"
    print >> file, " <title>Built-In Physics</title>"
    print >> file, """<para>
    &oof2; is designed to be extendible, so it is easy to add new
    &fields;, &fluxes;, and &equations;.  That means that the
    following tables of <emphasis>built-in</emphasis> objects may not
    be complete.  The listed dimension is the number of independent
    components in each object.
    </para>"""

    print >> file, "<table id='Table-BuiltInFields'>"
    print >> file, " <title>Built-in Fields</title>"
    print >> file, " <tgroup cols='3'>"
    print >> file, "  <thead><row>"
    print >> file, "   <entry><para>Name</para></entry>"
    print >> file, "   <entry><para>Type</para></entry>"
    print >> file, "   <entry><para>Dimension</para></entry>"
    print >> file, "  </row></thead>"
    print >> file, "  <tbody>"
    for fld in field.allFields():
        print >> file, "<row id='Field-%s'>" % fld.name()
        print >> file, " <entry><para><userinput>%s</userinput></para></entry>"\
              % fld.name()
        print >> file, " <entry><para><classname>%s</classname></para></entry>"\
              % fld.__class__.__name__
        print >> file, " <entry><para>%d</para></entry>" % fld.ndof()
        print >> file, "</row>"
    print >> file, "  </tbody>"
    print >> file, " </tgroup>"
    print >> file, "</table>"

    print >> file, "<table id='Table-BuiltInFluxes'>"
    print >> file, " <title>Built-in Fluxes</title>"
    print >> file, " <tgroup cols='3'>"
    print >> file, "  <thead><row>"
    print >> file, "   <entry><para>Name</para></entry>"
    print >> file, "   <entry><para>Type</para></entry>"
    print >> file, "   <entry><para>Dimension</para></entry>"
    print >> file, "  </row></thead>"
    print >> file, "  <tbody>"
    for flx in flux.allFluxes:
        print >> file, "<row id='Flux-%s'>" % flx.name()
        print >> file, " <entry><para><userinput>%s</userinput></para></entry>"\
              % flx.name()
        print >> file, " <entry><para><classname>%s</classname></para></entry>"\
              % flx.__class__.__name__[:-3] # strip 'Ptr'
        print >> file, " <entry><para>%d</para></entry>" % flx.ndof()
        print >> file, "</row>"
    print >> file, "  </tbody>"
    print >> file, " </tgroup>"
    print >> file, "</table>"

    print >> file, "<table id='Table-BuiltInEqns'>"
    print >> file, " <title>Built-in Equations</title>"
    print >> file, " <tgroup cols='4'>"
    print >> file, "  <thead><row>"
    print >> file, "   <entry><para>Name</para></entry>"
    print >> file, "   <entry><para>Type</para></entry>"
    print >> file, "   <entry><para>Flux</para></entry>"
    print >> file, "   <entry><para>Dimension</para></entry>"
    print >> file, "  </row></thead>"
    print >> file, "  <tbody>"
    links = {'PlaneFluxEquation' : 'Section-Concepts-Mesh-Equation-PlaneFlux',
             'DivergenceEquation' : 'Section-Concepts-Mesh-Equation-Divergence'}
    for eqn in equation.allEquations:
        classname = eqn.__class__.__name__[:-3] # strip off 'Ptr'
        if classname in links:
            print >> file, "<row id='Equation-%s'>" % eqn.name()
            print >> file, \
                " <entry><para><userinput>%s</userinput></para></entry>" \
                  % eqn.name()
            print >> file, \
                " <entry><para><link linkend='%s'><classname>%s</classname></link></para></entry>"  % (links[classname], classname)
            print >> file, \
                " <entry><para><userinput>%s</userinput></para></entry>" \
                % eqn.fluxname()
            print >> file, " <entry><para>%d</para></entry>" % eqn.ndof()
            print >> file, "</row>"        
        else:
            print >> sys.stderr, \
                "*** problem.py:xmldump: Not generating entry for", \
                classname, "***"
    print >> file, "  </tbody>"
    print >> file, " </tgroup>"
    print >> file, "</table>"

    print >> file, "</section>"
    

xmlmenudump.addSection(xmldump, 5)


